### $Id: analyzer.R,v 1.20 2011/08/19 00:49:18 srevilak Exp $

my.cex <- 1.5

my.dir <- "~/Documents/schoolwork/phd/si/mysql-test-results/par-cooked"

## Prefer fixed to scientific notation.
options(scipen=5)

# write the current graphics device to a file
#
# This routine uses the file extension to select the
# output routine.  
write.device <- function(outfile) {

  if (regexpr("\\.png$", outfile) >= 0) {
    dev.copy(png, file=outfile, width=800, height=600);
  }

  else if (regexpr("\\.pdf$", outfile) >= 0) {
    dev.copy(pdf, file=outfile, # width=8, height=6,
             pointsize=12,
             paper="USr", width=10, height=7.5,
             useDingbats=FALSE)
  }
  else {
    stop("image extension not handled in ", outfile)
  }
  dev.off();
}


### Plots data frame where the first column is named 'MPL'
### and columns 2..n are data series.
###
###
mpl.plot <- function(data, columns, title, ylab, outfile,
                     ylim.scale="fixed",
                     xlab="MPL",
                     legend.pos="topleft",
                     legend.labels=columns) {
  my.col <- "black"
  my.lty <- 1
  my.pch <- 1:length(columns)

  ylim <- NA
  if (ylim.scale == "fixed") {
    ylim <- c(0, max(data[2:length(data)], na.rm=TRUE))
  }
  else if (ylim.scale == "scaled") {
    ylim <- c(0, max(data[columns], na.rm=TRUE))
  }
  else {
    stop("ylim.scale must be 'fixed' or 'scaled'")
  }

  matplot(data['MPL'], data[columns],
          ylim=ylim,
          xlab=xlab,
          ylab=ylab,
          main=title,
          cex.main=my.cex,
          cex.lab=0.9 * my.cex,
          cex.axis=my.cex,
          type="n");
  
  grid(col="gray25");
  box();

  # better for lines to overplot legend, than for
  # legend to overplot lines.
  legend(legend.pos, lty=my.lty, pch=my.pch, col=my.col,
         legend=legend.labels, bg="white", cex=0.85 * my.cex,
         pt.cex = my.cex);
  
  matlines(data['MPL'], data[columns],
           lty=my.lty,
           col=my.col,
           pch=my.pch,
           cex=my.cex*1.3,
           type="b");
          
  if (! is.na(outfile)) {
    write.device(outfile);
  }
}




### INPUT: two column matrix or data frame, first column contains x values,
### second column contains y values.
###
### OUTPUT: data frame with three columns: x, y, and m.  x, y come
### from the input data frame.  m is the slope at point (x, y).
### The slope is computed from the secant line to points immediately
### surrounding (x, y)
###
mpl.slope <- function(d) {
  d.matrix <- as.matrix(d);
  mpl <- d[,1];
  y <- d[,2];

  ## throw out the first point (aka, MPL 1)
  mpl.out <- mpl[3:(length(mpl)-1)];
  slope <- array(data=rep(0, length(mpl.out)));
  point <- y[3:(length(y)-1)];

  for (i in 1:length(slope)) {
    rise <- y[i+3] - y[i+1];
    run <- mpl[i+3] - mpl[i+1];
    slope[i] <- (rise/run);
    point[i] <- y[i+2];
  }
  
  rv <- data.frame(mpl.out, point, slope);
  names(rv) <- c('x', 'y', 'm');
  rv
}



### INPUT: data frame produced by mpl.slopes.  This data frame has
### columns (x, y, m), where (x, y) is a point, and m is the slope at
### (x, y).
###
### This function adds tangent lines to an existing plot of (x, y).
###
mpl.plot.tangent <- function(xym, len=10, col="red", lty=1) {
  x.arr <- unlist(xym['x']);
  y.arr <- unlist(xym['y']);
  m.arr <- unlist(xym['m']);

  for (i in 1:length(x.arr)) {
    x0 <- x.arr[i] - (len/2);
    x1 <- x.arr[i] + (len/2);
    y0 <- y.arr[i] - ((len/2) * m.arr[i]);
    y1 <- y.arr[i] + ((len/2) * m.arr[i]);
    segments(x0=x0, y0=y0, x1=x1, y1=y1, col=col, lty=lty);
  }
}


### performs analysis on tx execution by second per thread
### 
### assumes one thread per DATA row, with time series in columns.
###
threads.by.second <- function(data) {
  d <- t(as.matrix(data))

  rv <- list();
  rv$table <- d;
  rv$cor <- cor(d);
  rv$cor.mean <- mean(rv$cor);
  rv$cor.min <- min(rv$cor);
  rv$row.mean <- apply(d, 1, mean)
  rv$row.sd <- apply(d, 1, sd)
  rv$row.mean.sd <- cbind(rv$row.mean, rv$row.sd);
  
  rv
}


### INPUT: a two-column matrix where column one contains
###        means and column two contains standard deviations
###
### generates a graph of means with error bars at +/- 2 standard
### deviations
### 
mean.sd.plot <- function(m, title="", outfile=NA) {
  m <- as.matrix(m);

  high.y <- m[,1] + 3 * m[,2];
  max.y <- ceiling(max(high.y))
  num.points <- length(m[,1]);

  matplot(m[,1], type="n",
          xlab="time (seconds)",
          ylab="avg. transactions per thread",
          main=title,
          ylim=c(0, max.y));

  grid(col="gray25");
  box();

  x <- 1:length(m[,1]);
  y.low <- m[,1] - 2 * m[,2];
  y.high <- m[,1] + 2 * m[,2];
  segments(x0=x, y0=y.low, x1=x, y1=y.high, col="gray5", lty=1, lend=0);

  matlines(m[,1], col="black", lty=1, type="l");

  if (! is.na(outfile)) {
    dev.copy(png, outfile, width=800, height=600);
    dev.off();
  }
}


### Plots data frame where each column is a time series of CPU 
### measurements
###
cpu.plot <- function(data,
                     title="",
                     outfile=NA,
                     legend.pos="topleft",
                     ylim=c(0,100),
                     use.col=TRUE) {

  columns <- names(data);
  if (use.col) {
    my.col <- rainbow(length(columns));
  }
  else {
    use.col <- "black"
  }

  my.lty <- 1;
  my.pch <- 1:length(columns)

  matplot(data,
          ylim=ylim,
          xlab="time",
          ylab="%cpu (user + system)",
          main=title,
          type="n");
  
  grid(col="gray25");
  box();
  
  matlines(data,
           lty=my.lty,
           col=my.col,
           pch=my.pch,
           type="l");
          
  legend(legend.pos, lty=my.lty, col=my.col, legend=columns, bg="white");

  if (! is.na(outfile)) {
    write.device(outfile);
  }
}


# plot a series of uptime data.  Series are contained in colums
# 
uptime.plot <- function(data,
                        title="",
                        outfile=NA,
                        legend.pos="topleft",
                        legend.names=names(data),
                        xlab="time",
                        ylab="load average") {

  columns <- legend.names;

  my.lty <- 1:length(columns);
  my.pch <- NULL
  my.col <- "black";

  my.type <- "b"
  if (is.null(my.pch)) {
    my.type <- "l"
  }
  
  matplot(data,
          xlab=xlab,
          ylab=ylab,
          main=title,
          type="n");
  
  grid(col="gray25");
  box();

  legend(legend.pos, pch=my.pch, col=my.col,
         lty=my.lty, legend=columns, bg="white");
  
  matlines(data,
           lty=my.lty,
           col=my.col,
           pch=my.pch,
           type=my.type);
          
  if (! is.na(outfile)) {
    write.device(outfile);
  }
}


### Plots data frame where each column is a time series of CPU 
### measurements
###
cycle.hist.plot <- function(data,
                            title="",
                            outfile=NA,
                            legend.pos="topright",
                            use.col=TRUE) {

  columns <- names(data)[2:length(data)];
  if (use.col) {
    my.col <- rainbow(length(columns));
  }
  else {
    my.col <- "black"
  }

  my.lty <- 1;
  my.pch <- 1:length(columns)

  matplot(data[1],
          data[columns],
          xlab="cycle length",
          ylab="number of cycles",
          main=title,
          type="n");
  
  grid(col="gray25");
  box();
  
  matlines(data[1],
           data[columns],
           lty=my.lty,
           col=my.col,
           pch=my.pch,
           type="b");
  
  legend(legend.pos, lty=my.lty, pch=my.pch, col=my.col, legend=columns, bg="white");

  if (! is.na(outfile)) {
    write.device(outfile);
  }
}


##################################################################
# Types
##################################################################

## Simple S4 object to hold information about a data set ("dset")
## 
setClass("dset",
         representation(id="character",
                        sequence="integer",
                        file="character",
                        data="data.frame",
                        notes="character"));


####################################################################
## Standard IO Stuff
####################################################################

my.read <- function(filename, is.numeric=TRUE) {
  if (is.numeric) {
    d <- read.table(filename, header=TRUE, sep="\t", colClasses="numeric");
  }
  else {
    d <- read.table(filename, header=TRUE, sep="\t");
  }
  d;
}



# reads a set of files and return a list of
# "dset" objects.
#
# Filenames are assumed to have the form "SEQUENCE.EXT".  For example,
# "325.txt".  These should be unqualified names under "dir".
#
# Returns a list of dset objects
read.files <- function(dir=".", filenames, fileids=filenames) {
  if (length(filenames) != length(fileids)) {
    stop("filenames and fileids have different lengths")
  }

  outfiles <- list();
  counter = 1;
  for (f in filenames) {
    seq <- as.integer(gsub("\\.[^.]*$", "", f))
    d <- my.read(file.path(dir, f), is.numeric=FALSE);
    rownames(d) <- d[['MPL']]
    dset <- new("dset")
    dset@data <- d;
    dset@file <- f;
    dset@id <- fileids[counter]
    dset@sequence <- seq
    outfiles[counter] <- dset
    counter <- counter + 1
  }

  outfiles
}


# Given a list of dset object, write them out
# to a file
#
write.dsets <- function(file, data) {
  fd <- file(description=file, open="w+")

  # iterate over the data elements.  Write each one.
  for (d in data) {
    seq <- d@sequence
    desc <- d@id

    write(sprintf("Sequence\t%d", seq), file=fd)
    write(sprintf("Description\t%s", desc), file=fd)
    write.table(d@data, file=fd, quote=FALSE, sep="\t", row.names=FALSE, na="")
    write("", file=fd)
  }

  # finally, write some provenance information, about computed
  # values
  first <- data[[1]]
  write("", file=fd)

  for (note in first@notes) {
    write(sprintf("# Note: %s", note), file=fd)
  }
  close(fd)  
}

# Write a data frame (or a list of data frames) to the file named
# by "file"
#
write.derived <- function(file, data) {
  fd <- file(description=file, open="w+");
  if (is.data.frame(data)) {
    write.derived.sub(fd, data)
  }
  else {
    for (d in data) {
      write.derived.sub(fd, d)
      write("", file=fd)
    }
  }
  close(fd)
}


# writes one data frame to the connection "fd".  The caller must close
# fd.
#
write.derived.sub <- function(fd, data) {
  attr <- attributes(data)
  if (! is.null(attr['measure'])) {
    write(paste("# Measure:\t", paste(attr[['measure']], collapse=" ")), file=fd)
  }

  if (! is.null(attr['sequences'])) {
    write(paste("# Sequence numbers:\t",
                paste(attr[['sequences']], collapse=" ")), file=fd);
  }

  write.table(data, file=fd, quote=FALSE, sep="\t", row.names=FALSE, na="")
}



####################################################################
## Computation of Derived Data
####################################################################


add.ctps <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "ctps=tx.committed/avg.thread.duration.sec");
    df <- d@data
    ctps <- df['tx.committed']/df['avg.thread.duration.sec']
    d@data['ctps'] <- ctps
    d
  })
}

add.fuw.sec <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "fuw_sec=errors.fcw/avg.thread.duration.sec");
    df <- d@data
    fuw <- df['errors.fcw']/df['avg.thread.duration.sec']
    d@data['fuw_sec'] <- fuw
    d
  })
}


add.sr.err.sec <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "sr_error_sec=errors.cycle/avg.thread.duration.sec");
    df <- d@data
    sr.error <- df['errors.cycle']/df['avg.thread.duration.sec']
    d@data['sr_error_sec'] <- sr.error
    d
  })
}


add.log.flush.rate <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "log_flush_sec=Innodb_os_log_fsyncs/(adjusted avg.thread.duration.sec)");
    df <- d@data

    warmup <- 70
    cooldown <- 5
    if ("warmupPeriod" %in% names(df)) {
      warmup <- df['warmupPeriod']
      cooldown <- warmup;
      if (cooldown > 5) {
        cooldown <- 5;
      }
    }
    syncs <- df['Innodb_os_log_fsyncs'] /
      (df['avg.thread.duration.sec'] + warmup + cooldown)
    d@data['log_flush_sec'] <- syncs
    d;
  })
}


add.commits.per.flush <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d,
                   paste("commits_per_flush=",
                         "(tx-committed)/",
                         "(adjusted Innodb_os_log_fsyncs)",
                         sep=""));
    df <- d@data

    warmup <- 70
    cooldown <- 5
    if ("warmupPeriod" %in% names(df)) {
      warmup <- df['warmupPeriod']
      cooldown <- warmup;
      if (cooldown > 5) {
        cooldown <- 5;
      }
    }

    scale =  (df['avg.thread.duration.sec'] /
              (df['avg.thread.duration.sec'] + warmup + cooldown));

    commits <- df['tx.committed'];
    adj.flush <- df['Innodb_os_log_fsyncs'] * scale;
                  

    d@data['commits_per_flush'] <- commits / adj.flush
    d;
  })
}


add.pct.failed <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_failed=100*(tx.rollback/total.transactions)");
    df <- d@data
    pct.fail <- 100.0 * (df['tx.rollback']/df['total.transactions'])
    d@data['pct_failed'] <- pct.fail
    d
  })
}

add.pct.fuw <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_fuw=100*(errors.fcw/total.transactions)");
    df <- d@data
    pct.fuw <- 100.0 * (df['errors.fcw']/df['total.transactions'])
    d@data['pct_fuw'] <- pct.fuw
    d
  })
}

add.pct.sr.err <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "pct_sr_err=100*(errors.cycle/total.transactions)");
    df <- d@data
    pct.sr.err <- 100.0 * (df['errors.cycle']/df['total.transactions'])
    d@data['pct_sr_err'] <- pct.sr.err
    d
  })
}

add.avg.commit.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, "avg.commit.duration=sum.commit.duration.ms/tx.committed");
    df <- d@data;
    sum.duration <- df['sum.commit.duration.ms']
    tx.committed <- df['tx.committed']
    d@data['avg.commit.duration'] <- (sum.duration / tx.committed)
    d
  })
}


# Average duration of committed transactions, including the commit
add.avg.commited.tx.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.commit.duration=",
                            "(sum.commit.duration.ms+sum.exec.duration.ms)",
                            "/tx.committed", sep=""));
    df <- d@data;
    total.duration <- df['sum.commit.duration.ms'] + df['sum.exec.duration.ms']
    tx.committed <- df['tx.committed']
    d@data['avg.commited.tx.duration'] <- (total.duration / tx.committed)
    d
  })
}

add.avg.exec.duration <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.exec.duration=",
                            "sum.exec.duration.ms",
                            "/tx.committed", sep=""));
    df <- d@data;
    exec.duration <- df['sum.exec.duration.ms'] / df['tx.committed']
    d@data['avg.exec.duration'] <- exec.duration
    d
  })
}

add.pct.time.in.commit <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("pct.time.in.commit=",
                            "100 * (sum.commit.duration.ms)/",
                            "(sum.commit.duration.ms + sum.exec.duration.ms)",
                            sep=""));
    df <- d@data;
    total.duration <- df['sum.commit.duration.ms'] + df['sum.exec.duration.ms']
    commit.duration <- df['sum.commit.duration.ms']
    d@data['pct.time.in.commit'] <- 100 * (commit.duration / total.duration)
    d
  })
}


add.avg.ctg.size <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.ctg.size=",
                            "(Innodb_sum_ctg_size/Innodb_cycle_tests_done)",
                            sep=""));
    df <- d@data;
    d@data['avg.ctg.size'] <- df['Innodb_sum_ctg_size'] /
      df['Innodb_cycle_tests_done'] 
    d
  })
}


add.avg.prune.nodes.examined <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.prune.nodes.examined=",
                            "(Innodb_prune_nodes_examined/Innodb_prune_count)",
                            sep=""));
    df <- d@data;
    d@data['avg.prune.nodes.examined'] <- df['Innodb_prune_nodes_examined'] /
      df['Innodb_prune_count'] 
    d
  })
}


# number of edges traversed during cycle testing.
add.cycle.edges.traversed <- function(data) {
  lapply(data, function(d) {
    d <- dset.note(d, paste("avg.cycle.edges.traversed=",
                            "(Innodb_ctg_edges_traversed/Innodb_cycle_tests_done)",
                            sep=""));
    df <- d@data;
    d@data['avg.cycle.edges.traversed'] <- df['Innodb_ctg_edges_traversed'] /
      df['Innodb_cycle_tests_done'] 
    d
  })
}



# Adds note to a dset.  This is typically used to keep track
# of provenance.
#
dset.note <- function(d, note) {
  idx <- 1+ length(slot(d, 'notes'));
  slot(d, 'notes')[idx] <- note
  d
}



####################################################################
## Data Summarization
####################################################################


collect.measure <- function(data, measure) {
  first <- data[[1]]
  if (class(first)[1] != "dset") {
    stop("data is not a list of dset objects")
  }
  df <- first@data;
  dnames <- names(df)

  if (! (measure %in% dnames)) {
    namelist <- paste(dnames, sep=",")
    stop("no ", measure, " in data frame")
  }
  
  outf <- data.frame(MPL=df['MPL'])
  sequences <- numeric()
  counter <- 1
  
  for (d in data) {
    id <- d@id
    datum <- d@data[measure]
    outf[id] <- datum
    sequences[counter] <- d@sequence
    counter <- counter + 1
  }
  attr(outf, "sequences") <- sequences
  attr(outf, "measure") <- measure
  
  outf
}


