/*
 * $Id: BaseForeman.java,v 1.4 2011/06/04 23:40:51 srevilak Exp $
 */

package edu.umb.bench.framework;
import java.sql.SQLException;
import java.sql.Connection;
import java.util.concurrent.atomic.AtomicInteger;
import java.text.DecimalFormat;
import java.text.FieldPosition;

/**
 * Base class for a Foreman.
 *
 * @author srevilak
 * @since 6/2/2011
 */
public abstract class BaseForeman implements IForeman {

    private ConnectionManager mConnectionManager;
    private BenchConfig mBenchConfig;
    private ILogger mLogger;
    private IErrorInterpreter mErrorInterpreter;
    private AtomicInteger mWorkers = new AtomicInteger();
    private DecimalFormat mFormat;

    /**
     * @inheritDoc
     */
    public void setConnectionManager(ConnectionManager cm) {
        mConnectionManager = cm;
    }

    protected ConnectionManager getConnectionManager() {
        return mConnectionManager;
    }


    /**
     * @inheritDoc
     */
    public void setConfig(BenchConfig config) {
        mBenchConfig = config;
    }

    protected BenchConfig getConfig() {
        return mBenchConfig;
    }


    public void setLogger(ILogger logger) {
        mLogger = logger;
    }


    protected ILogger getLogger() {
        return mLogger;
    }


    public void init() {
        if (mBenchConfig == null) {
            throw new IllegalStateException("Missing Config object");
        }

        if (mLogger == null) {
            throw new IllegalStateException("missing logger");
        }

        if (mConnectionManager == null) {
            throw new IllegalStateException("missing connection manager");
        }

        mErrorInterpreter = 
            Util.makeInstance(mBenchConfig.getErrorInterpreterClass());

        mFormat = new DecimalFormat("00000");

        childInit();
    }


    protected abstract void childInit();


    /**
     * Creates and configures a single worker.
     */
    @SuppressWarnings("unchecked")
    protected IWorker createWorker() throws SQLException {

        if (mErrorInterpreter == null) {
            throw new IllegalStateException("no error interpreter");
        }
        
        IWorker worker = Util.makeInstance(mBenchConfig.getWorkerClass());
        ITransactionFactory factory
            = Util.makeInstance(mBenchConfig.getFactoryClass());
            
        Connection conn = mConnectionManager.getConnection();
        worker.setConnection(conn);
        factory.setConnection(conn);

        worker.setLogger(mLogger);
        factory.setLogger(mLogger);

        worker.setConfig(mBenchConfig.getTestProperties());
        factory.setConfig(mBenchConfig.getTestProperties());
        
        factory.setScale(mBenchConfig.getScale());
        factory.init();

        worker.setTransactionFactory(factory);
        worker.setErrorInterpreter(mErrorInterpreter);

        int count = mWorkers.addAndGet(1);
        StringBuffer buffer = new StringBuffer();
        buffer.append("thread-");
        mFormat.format(count, buffer, new FieldPosition(0));

        worker.setName(buffer.toString());
        worker.setID(count);

        worker.setThinkTime(mBenchConfig.getThinkTime());

        return worker;
    }
}
