/*
 * $Id: BaseWorker.java,v 1.16 2011/06/04 19:57:51 srevilak Exp $
 */

package edu.umb.bench.framework;
import edu.umb.bench.framework.IErrorInterpreter.ErrorType;
import java.sql.SQLException;
import java.util.Map;
import java.util.HashMap;

/**
 * Base class for workers.
 *
 * @version $Revision: 1.16 $ $Date: 2011/06/04 19:57:51 $
 */
public abstract class BaseWorker<T> extends BaseObject implements IWorker<T> {

    private ITransactionFactory<T> mFactory;
    private Exception mError;
    private String mName;
    private int mID;
    private IErrorInterpreter mErrorInterpreter;
    private Object mTrigger;
    private IBeforeCommitListener mCommitListener;
    private IEndListener mEndListener;

    private int mCommit;
    private int mRollback;
    private int mFCW;
    private int mCycle;
    private int mDeadlock;
    private int mTimeout;
    private int mPathTooLong;
    private long mThreadDuration;
    private double mThinkTime;
    private int mPeriodicTransactionCount;

    /**
     * Sets the transaction object for this workload.
     */
    public void setTransactionFactory(ITransactionFactory<T> factory) {
        mFactory = factory;
    }


    /**
     * Sets the execution trigger (if triggers are used)
     */
    public void setTrigger(Object trigger) {
        mTrigger = trigger;
    }

    /**
     * @inheritDoc
     */
    public void setBeforeCommitListener(IBeforeCommitListener listener) {
        mCommitListener = listener;
    }

    /**
     * Sets an atEnd Listener
     *
     * @see IEndListener
     */
    public void setEndListener(IEndListener listener) {
        mEndListener = listener;
    }

    /**
     * Sets the error interpreter for this worker.
     */
    public void setErrorInterpreter(IErrorInterpreter interpreter) {
        mErrorInterpreter = interpreter;
    }


    /**
     * Returns the transaction factory for this workload.
     */
    public ITransactionFactory<T> getTransactionFactory() {
        return mFactory;
    }


    /**
     * Returns the error for this worker, or null if there was no error.
     */
    public Exception getError() {
        return mError;
    }


    /**
     * Sets the error for this workload.
     */
    public void setError(Exception e) {
        mError = e;
    }

    /**
     * @inheritDoc
     */
    public void setName(String name) {
        mName = name;
    }

    
    /**
     * @inheritDoc
     */
    public String getName() {
        return mName;
    }


    /**
     * @inheritDoc
     */
    public void setID(int id) {
        mID = id;
    }


    /**
     * @inheritDoc
     */
    public int getID() {
        return mID;
    }

    /**
     * @inheritDoc
     */
    public void setThinkTime(double d) {
        if (d >= 0.0) {
            mThinkTime = d;
        }
    }

    /**
     * @inheritDoc
     */
    public double getThinkTime() {
        return mThinkTime;
    }

    /**
     * @inheritDoc
     */
    public int getPeriodicTransactionCount() {
        int oldvalue = 0;
        synchronized (this) {
            oldvalue = mPeriodicTransactionCount;
            mPeriodicTransactionCount = 0;
        }
        return oldvalue;
    }

    protected void incrementPeriodicTransactionCount() {
        synchronized (this) {
            mPeriodicTransactionCount++;
        }
    }


    /**
     * Runs the worker thread.
     */
    public void run() {
        if (mName != null) {
            Thread.currentThread().setName(getName());
        }
        else {
            mName = Thread.currentThread().getName();
        }

        try {
            setError(null);
            assert getTransactionFactory() != null;
            runInternal();
        }
        catch (InterruptedException e) {
            getLogger().info("worker " + getName() + " interrupted");
        }
        catch (Exception e) {
            getLogger().error("worker " + Thread.currentThread().getName()
                              + " failed", e);
            setError(e);
        }
        runEndListener();
    }

    
    /**
     * Internal runloop.
     *
     * Note that we do very little error handling here.
     * The framework creates connections for us, and the
     * framework releases them for us.
     */
    private void runInternal() throws SQLException, InterruptedException {
        long t0 = 0;
        long t1 = 0;

        boolean ranOnce = false;

        while (true) {

            if (mTrigger != null && ranOnce) {
                synchronized(mTrigger) {
                    mTrigger.wait();
                }
            }
            ranOnce = true;

            if (TEST_STATE == TestState.DONE) {
                if (t1 == 0 && t0 > 0) {
                    /* in case thread missed the cool down signal */
                    t1 = System.currentTimeMillis();
                }
                break;
            }

            TestState state = TEST_STATE;

            /* grab start of measurement period */
            if (state == TestState.MEASUREMENT && t0 == 0) {
                t0 = System.currentTimeMillis();
            }

            boolean warmup = (state == TestState.WARMUP 
                              || state == TestState.COOLDOWN);

            ErrorType errorType = ErrorType.NONE;
            
            SQLException err = null;
            ITransaction<T> tx = getTransactionFactory().createTransaction();

            boolean listener_notified = false;

            try {
                tx.execute(warmup);
                if (mCommitListener != null) {
                    mCommitListener.beforeCommit(this);
                    listener_notified = true;
                }
                
                tx.commit();
            }
            catch (SQLException e) {
                err = e;
                if (mCommitListener != null && ! listener_notified) {
                    mCommitListener.beforeCommit(this);
                }

                tx.rollback();
            }

            /*
             * If we got an error, decide whether it's transient and
             * we can continue, or if it's non-transient and we should
             * rethrow.
             */
            if (err != null) {
                errorType = mErrorInterpreter.classifyError(err);

                switch (errorType) {
                case TERMINAL:
                case UNKNOWN:                    
                    throw err;

                case FCW:
                case CYCLE:
                case DEADLOCK:
                case TIMEOUT:
                case PATH_TOO_LONG:
                case NONE:
                    break;

                default:
                    throw new RuntimeException("unhandled error type "
                                               + errorType
                                               + ", for error "
                                               + err, err);
                }
            }


            /* grab end time of execution */
            if (state == TestState.COOLDOWN && t1 == 0) {
                t1 = System.currentTimeMillis();
            }

            /*
             * No measurements during warmup/cooldown.
             */
            if (warmup) {
                randomSleep();
                runEndListener();
                continue;
            }

            /*
             * Note tranaction-specific results
             */ 
            T result = tx.getResults();
            addResult(result);
            
            /*
             * Note commit vs rollback
             */
            if (err == null) {
                mCommit++;
            }
            else {
                mRollback++;
            }
            incrementPeriodicTransactionCount();

            /*
             * Record error categorization.
             *
             * UMB's MySQL has counters for these, but those counters
             * would include measurements during the warmup period.
             * We don't want to count the warmup period.
             */
            switch (errorType) {
            case FCW:
                mFCW++;
                break;

            case CYCLE:
                mCycle++;
                break;

            case DEADLOCK:
                mDeadlock++;
                break;

            case TIMEOUT:
                mTimeout++;
                break;

            case PATH_TOO_LONG:
                mPathTooLong++;
                break;

            default:
                break;
            }

            randomSleep();      

            runEndListener();
        }

        /* duration of this thread */
        mThreadDuration = t1 - t0;
    }


    private void runEndListener() {
        if (mEndListener != null) {
            mEndListener.atEnd(this);
        }
    }
    
    public final Map<String,Object> getBaseResults() {

        Map<String,Object> child_results = getResults();

        if (child_results == null) {
            child_results = new HashMap<String,Object>();
        }

        int total_tx = mCommit + mRollback;

        /* results for global summation */
        child_results.put(TX_COMMITTED, Integer.valueOf(mCommit));
        child_results.put(TX_ROLLBACK, Integer.valueOf(mRollback));
        child_results.put(ERR_FCW, Integer.valueOf(mFCW));
        child_results.put(ERR_DEADLOCK, Integer.valueOf(mDeadlock));
        child_results.put(ERR_CYCLE, Integer.valueOf(mCycle));
        child_results.put(ERR_PATH_TOO_LONG, Integer.valueOf(mPathTooLong));
        child_results.put(ERR_TIMEOUT, Integer.valueOf(mTimeout));
        child_results.put(TOTAL_TRANSACTIONS, Integer.valueOf(total_tx));
        child_results.put(SUM_THREAD_DURATION, Long.valueOf(mThreadDuration));
        
        /* per-thread results */
        String name = getName();

        child_results.put(name + "_duration_ms", Long.valueOf(mThreadDuration));
        child_results.put(name + "_tx_run", Integer.valueOf(total_tx));
        
        double avg_duration = ((double) mThreadDuration) / ((double) total_tx);
        
        child_results.put(name + "_avg_tx_duration_ms", Double.valueOf(avg_duration));

        double tps = ((double) total_tx) / 
            (((double) mThreadDuration) / 1000.0);

        child_results.put(name + "_tps", Double.valueOf(tps));

        return child_results;
    }


    public Map<String,Object> getResults() {
        return new HashMap<String,Object>();
    }


    private void randomSleep() throws InterruptedException {
        Util.randomSleep(mThinkTime);
    }

}
