/*
 * $Id: BenchConfig.java,v 1.12 2011/06/04 00:40:27 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.io.File;
import java.util.Properties;
import java.util.Map;
import java.util.Arrays;
import java.util.HashMap;

/**
 * Container for benchmark configuration parameters.
 * This is basically everything that needs to be set up in order
 * to run a benchmark.
 *
 * <p>
 * General idea
 * <ul>
 * <li>call setters to configure different classes
 * <li>call {@link #init}.  This performs two functions: (a) checking the 
 * configuration parameters, and (b) loading the classes needed to run
 * a specific benchmark.  After the classes are loaded, you can use 
 * the different getters to retrieve them.
 * </ul>
 *
 * Once you've called <tt>init()</tt>, you shouldn't try to change
 * the configuration.
 *
 * @version $Revision: 1.12 $ $Date: 2011/06/04 00:40:27 $
 */
public class BenchConfig {

    /**
     * Setup step ({@value}).
     */
    public static final int STEP_SETUP = 1;

    /**
     * Audit step ({@value}).
     */
    public static final int STEP_AUDIT = 2;

    /**
     * Test execution step ({@value}).
     */
    public static final int STEP_TEST = 4;

    /**
     * All steps.  By default, all steps are enabled, but they
     * can be selectively togged via {@link #enableStep} and
     * {@link #disableStep}.
     */
    public static final int STEP_ALL = 
        (STEP_SETUP | STEP_AUDIT | STEP_TEST);

    private int mMPL;
    private int mDuration;
    private double mScale;
    private int mWarmupPeriod;
    private File mTestConfig;
    private File mDBConfig;
    private File mResultFile;
    private String mAuditorClassname;
    private Class <? extends IAuditor> mAuditorClass;
    private String mAggregatorClassname;
    private Class <? extends IAggregator> mAggregatorClass;
    private String mSetupClassname;
    private Class <? extends ISetup> mSetupClass;
    private String  mFactoryClassname;
    private Class <? extends ITransactionFactory> mFactoryClass;
    private String mWorkerClassname;
    private Class <? extends IWorker> mWorkerClass;
    private String mSystemAuditorClassname;
    private Class <? extends ISystemAuditor> mSystemAuditorClass;
    private String mInterpreterClassname;
    private Class <? extends IErrorInterpreter> mInterpreterClass;
    private String mForemanClassname;
    private Class <? extends IForeman> mForemanClass;
    private Properties mTestProperties;
    private Properties mDBProperties;
    private boolean inited;
    private int mSteps;
    private double mThinkTime;

    public BenchConfig() {
        inited = false;
        mSteps = STEP_ALL;
    }


    /**
     * Sets the multiprogramming level for this benchmark.
     */
    public void setMPL(int mpl) {
        assert(! inited);
        mMPL = mpl;
    }


    /**
     * Returns the MPL for this benchmark.
     */
    public int getMPL() {
        return mMPL;
    }


    /**
     * Sets the duration of this benchmark.  The duration is the total number
     * of transactions that must be run after the warmup period.
     */
    public void setDuration(int duration) {
        assert(! inited);
        mDuration = duration;
    }


    /**
     * Returns the duration for this benchmark.
     */
    public int getDuration() {
        return mDuration;
    }


    /**
     * Sets the think time, in milliseconds.
     * This controls how long each thread waits after committing
     * or aborting a transaction.  This value is randomized to +/- 50%
     */
    public void setThinkTime(double ttime) {
        assert(! inited);
        mThinkTime = ttime;
    }

    public double getThinkTime() {
        return mThinkTime;
    }


    /**
     * Sets the scale for this benchmark (data scaling).
     */
    public void setScale(double scale) {
        assert(! inited);
        mScale = scale;
    }

    /**
     * Returns the scale for this benchmark.
     */
    public double getScale() {
        return mScale;
    }


    /**
     * Sets the warmup period, in seconds.
     */
    public void setWarmupPeriod(int warmup) {
        assert(! inited);
        mWarmupPeriod = warmup;
    }

    
    /**
     * Returns the warmup period.
     */
    public int getWarmupPeriod() {
        return mWarmupPeriod;
    }

    
    /**
     * Returns the cooldown period.  The cooldown period is
     * the greater of {warmupPeriod, 5 seconds}
     */
    public int getCooldownPeriod() {
        if (mWarmupPeriod < 5) {
            return mWarmupPeriod;
        }
        return 5;
    }



    /**
     * Sets the test configuration file.  This is a properties
     * file that holds benchmark-specific configurations.
     */
    public void setTestConfig(File config) {
        assert(! inited);
        mTestConfig = config;
    }

    /**
     * Returns the test config file path.
     */
    public File getTestConfig() {
        return mTestConfig;
    }


    /**
     * Sets the configuration file that contains database
     * connection parameters.
     */
    public void setDatabaseConfig(File config) {
        assert(! inited);
        mDBConfig = config;
    }

    /**
     * Returns the database configuration file.
     */
    public File getDatabaseConfig() {
        return mDBConfig;
    }
        
    
    /**
     * Sets the location for the result file.
     */
    public void setResultFile(File result) {
        assert(! inited);
        mResultFile = result;
    }


    /**
     * Returns the result file path.
     */
    public File getResultFile() {
        return mResultFile;
    }
    

    /**
     * Sets the fully qualified name of the auditor class
     *
     * @see IAuditor
     * @see BaseAuditor
     */
    public void setAuditorClassname(String name) {
        assert(! inited);
        mAuditorClassname = name;
    }


    /**
     * Returns the auditor classname. 
     */
    public String getAuditorClassname() {
        return mAuditorClassname;
    }


    /**
     * Sets the aggregator classname.
     *
     * @see IAggregator
     * @see BaseAggregator
     */
    public void setAggregatorClassname(String name) {
        assert(! inited);
        mAggregatorClassname = name;
    }


    /**
     * Returns the aggregator classname
     */
    public String getAggregatorClassname() {
        return mAggregatorClassname;
    }


    /**
     * Sets the classname for the setup object.
     *
     * @see ISetup
     * @see BaseSetup
     */
    public void setSetupClassname(String name) {
        assert (! inited);
        mSetupClassname = name;
    }


    /**
     * Returns the setup object classname.
     */
    public String getSetupClassname() {
        return mSetupClassname;
    }


    /**
     * Specifies the name of the transaction factory class.
     *
     * @see ITransactionFactory
     * @see BaseTransactionFactory
     */
    public void setFactoryClassname(String name) {
        assert (! inited);
        mFactoryClassname = name;
    }


    /**
     * Returns the transaction factory classname.
     */
    public String getFactoryClassname() {
        return mFactoryClassname;
    }


    /**
     * Specifies the classname of client workers.
     *
     * @see IWorker
     * @see BaseWorker
     */
    public void setWorkerClassname(String name) {
        assert (! inited);
        mWorkerClassname = name;
    }


    /**
     * Returns the worker classname.
     */
    public String getWorkerClassname() {
        return mWorkerClassname;
    }


    /**
     * Sets the system auditor classname.
     *
     * @see ISystemAuditor
     * @see MySQLAuditor
     */
    public void setSystemAuditorClassname(String name) {
        assert (! inited);
        mSystemAuditorClassname = name;
    }


    /**
     * Returns the system auditor classname
     */
    public String getSystemAuditorClassname() {
        return mSystemAuditorClassname;
    }


    /**
     * Sets the name of the error interpreter
     *
     * @see IErrorInterpreter
     * @see MySQLErrorInterpreter
     */
    public void setErrorInterpreterClassname(String name) {
        assert (! inited);
        mInterpreterClassname = name;
    }

    /**
     * Returns the classname of the error interpreter
     */
    public String getErrorInterpreterClassname() {
        return mInterpreterClassname;
    }

    
    /**
     * Sets the name of the Foreman class
     * @see IForeman
     */
    public void setForemanClassname(String name) {
        assert(! inited);
        mForemanClassname = name;
    }

    
    public String getForemanClassname() {
        return mForemanClassname;
    }


    /**
     * Returns the Aggregator class object.
     */
    public Class <? extends IAggregator> getAggregatorClass() {
        assert(inited);
        return mAggregatorClass;
    }


    /**
     * Returns the setup class.
     */
    public Class <? extends ISetup> getSetupClass() {
        assert(inited);
        return mSetupClass;
    }


    /**
     * Returns the Transaction factory class.
     */
    public Class <? extends ITransactionFactory> getFactoryClass() {
        assert(inited);
        return mFactoryClass;
    }


    /**
     * Returns the class object for workers.
     */
    public Class <? extends IWorker> getWorkerClass() {
        assert(inited);
        return mWorkerClass;
    }


    /**
     * Returns the auditor class object
     */
    public Class <? extends IAuditor> getAuditorClass() {
        assert(inited);
        return mAuditorClass;
    }


    /**
     * Returns the system auditor class object.
     */
    public Class <? extends ISystemAuditor> getSystemAuditorClass() {
        assert(inited);
        return mSystemAuditorClass;
    }


    /**
     * Returns the error interpreter class object.
     */
    public Class <? extends IErrorInterpreter> getErrorInterpreterClass() {
        assert (inited);
        return mInterpreterClass;
    }

    /**
     * Returns the foreman class object
     */
    public Class<? extends IForeman> getForemanClass() {
        assert (inited);
        return mForemanClass;
    }


    /**
     * Returns test properties, as read from the 
     * test config properties file
     *
     * @see #getTestConfig
     */
    public Properties getTestProperties() {
        assert(inited);
        return mTestProperties;
    }

    /**
     * Returns the set of database configuration properties.
     */
    public Properties getDatabaseProperties() {
        assert(inited);
        return mDBProperties;
    }


    /**
     * Initializes the benchmark configuration.  This method
     * (a) validates each of the parameters, and (b) loads
     * each of the class objects.
     *
     * <p>
     * Once init has been called successfully, subsequent calls
     * will not change the state of this object.
     *
     * @exception IllegalArgumentException if a parameter is missing or invalid.
     *
     * @exception TypeNotPresentException if a class object was not loadable.
     */
    public void init() {
        if (inited) {
            return;
        }
        validate();
        loadClasses();
        loadProperties();

        inited = true;
    }


    /**
     * Validates the configuration.
     */
    private void validate() {
        if (getMPL() < 1) {
            throw new IllegalArgumentException("MPL must be positive, not"
                                               + getMPL());
        }

        if (getDuration() < 0) {
            throw new IllegalArgumentException("Duration must be non-negative, "
                                               + "not " + getDuration());
        }

        if (getScale() < 0.0) {
            throw new IllegalArgumentException("scale must be non-negative, "
                                              + "not " + getScale());
        }

        if (getWarmupPeriod() < 0) {
            throw new IllegalArgumentException("warmup period must be non-negative, "
                                              + "not " + getWarmupPeriod());
        }

        /* we'll allow a test to have a null configuration */

        if (getDatabaseConfig() == null) {
            throw new IllegalArgumentException("missing database connection "
                                              + "configuration file");
        }

        if (getResultFile() == null) {
            throw new IllegalArgumentException("missing result (output) file");
        }

        if (getAuditorClassname() == null) {
            throw new IllegalArgumentException("missing auditor classname");
        }

        if (getAggregatorClassname() == null) {
            throw new IllegalArgumentException("missing aggregator classname");
        }

        if (getSetupClassname() == null) {
            throw new IllegalArgumentException("missing setup classname");
        }

        if (getFactoryClassname() == null) {
            throw new IllegalArgumentException("missing transaction factory "
                                              + "classname");
        }

        if (getWorkerClassname() == null) {
            throw new IllegalArgumentException("missing worker classname");
        }

        if (getSystemAuditorClassname() == null) {
            throw new IllegalArgumentException("missing system auditor "
                                               + "classname");
        }

        if (getErrorInterpreterClassname() == null) {
            throw new IllegalArgumentException("missing error interpreter "
                                               + "classname");
        }

    }


    /**
     * Loads the class objects specified by the configuration.
     */
    private void loadClasses() {
        mAuditorClass = Util.loadClass(getAuditorClassname(), 
                                       IAuditor.class);

        mAggregatorClass = Util.loadClass(getAggregatorClassname(),
                                          IAggregator.class);

        mSetupClass = Util.loadClass(getSetupClassname(),
                                     ISetup.class);

        mFactoryClass = Util.loadClass(getFactoryClassname(),
                                       ITransactionFactory.class);

        mWorkerClass = Util.loadClass(getWorkerClassname(),
                                      IWorker.class);

        mAuditorClass = Util.loadClass(getAuditorClassname(),
                                       IAuditor.class);

        mSystemAuditorClass = Util.loadClass(getSystemAuditorClassname(),
                                             ISystemAuditor.class);

        mInterpreterClass = Util.loadClass(getErrorInterpreterClassname(),
                                           IErrorInterpreter.class);

        mForemanClass = Util.loadClass(getForemanClassname(),
                                       IForeman.class);
    }


    private void loadProperties() {
        if (getTestConfig() != null) {
            mTestProperties = Util.readProperties(getTestConfig());
        }
        else {
            mTestProperties = new Properties();
        }

        mDBProperties = Util.readProperties(getDatabaseConfig());
    }


    /**
     * Enables a step.  <tt>step</tt> should be one of the
     * <tt>STEP_*</tt> constants defined above.
     *
     * <p> By default all steps are enabled.
     *
     * @return the current step configuration.
     */
    public int enableStep(int step) {
        mSteps = mSteps | step;
        mSteps = mSteps & STEP_ALL;
        return mSteps;
    }


    /**
     * Disables a particular step.
     * 
     * @see #enableStep
     */
    public int disableStep(int step) {
        mSteps = mSteps & (~step);
        return mSteps;
    }


    /**
     * Tests whether a particular step
     * is enabled.
     *
     * @see #enableStep
     */
    public boolean isStepEnabled(int step) {
        return ((mSteps & step) == step);
    }

    /**
     * Returns the set of enabled steps.
     */
    public int getEnabledSteps() {
        return mSteps;
    }


    /**
     * Returns the Benchmark configuration as a set of 
     * key/value pairs.  This is similar to the information provided
     * my {@link #toString}, but in the form of a Map.
     */
    public Map<String,Object> getBenchConfig() {
        Map<String,Object> map = new HashMap<String,Object>();

        map.put("mpl", Integer.valueOf(getMPL()));

        map.put("duration", Integer.valueOf(getDuration()));

        map.put("scale", Double.valueOf(getScale()));

        map.put("warmupPeriod", Integer.valueOf(getWarmupPeriod()));

        map.put("testConfig", getTestConfig());

        map.put("databaseConfig", getDatabaseConfig());

        map.put("resultFile", getResultFile());

        map.put("setup", getSetupClassname());

        map.put("auditor", getAuditorClassname());

        map.put("factory", getFactoryClassname());

        map.put("worker", getWorkerClassname());

        map.put("aggregator", getAggregatorClassname());

        map.put("systemAuditor", getSystemAuditorClassname());

        map.put("errorInterpreter", getErrorInterpreterClassname());

        map.put("foreman", getForemanClassname());

        map.put("steps", Integer.valueOf(mSteps));

        map.put("thinkTime", Double.valueOf(getThinkTime()));

        return map;
    }

    
    public String toString() {
        final String NL = System.getProperty("line.separator");
        StringBuilder sb = new StringBuilder(4096);

        Map<String,Object> map = getBenchConfig();
        
        String keys[] = map.keySet().toArray(new String[map.size()]);
        Arrays.sort(keys);

        sb.append("BenchConfig:");
        sb.append(NL);

        for (int iter = 0; iter < keys.length; iter++) {
            Object value = map.get(keys[iter]);

            sb.append("  ");
            sb.append(keys[iter]);
            sb.append('=');

            if (value != null) {
                sb.append(value.toString());
            }
            sb.append(NL);
        }

        return sb.toString();
    }
    
}
