/*
 * $Id: BenchRunner.java,v 1.18 2011/06/04 19:57:51 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Benchmark runner
 */
public class BenchRunner {
    
    private BenchConfig mBenchConfig;
    private ConnectionManager mManager;
    private ILogger mLogger;
    private IAuditor mTestAuditor;
    private ISystemAuditor mSystemAuditor;
    private IAggregator mAggregator;
    private ResultContainer mResultContainer;
    private BenchTimers mTimers;
    private PeriodicCollector mPeriodicCollector;
    private IForeman mForeman;
    
    private static class BenchTimers {
        /** very beginning of test run */
        long start;

        /** start of warmup */
        long startWarmup;

        /** beginning of measurement period */
        long startMeasure;

        /** beginning of cooldown period */
        long startCooldown;

        /** end of test */
        long stopTest;

        /** number of threads that failed */
        int failedWorkers;
    }


    /**
     * Constructs a new benchmark runner, using the specified bench
     * configuration
     */
    public BenchRunner(BenchConfig config) {
        if (config == null) {
            throw new NullPointerException("BenchConfig config required");
        }
        mBenchConfig = config;
        mResultContainer = new ResultContainer();
        mTimers = new BenchTimers();
    }

    
    /**
     * Runs a benchmark.
     */
    public void runBenchmark() {
        try {
            init();
            setup();
            preAudit();
            runWorkers();
            postAudit();
            collectResults();
            saveResults();
            
        }
        catch (Exception e) {
            String err = "unhandled exception running benchmark";
            mLogger.error(err, e);
            throw new RuntimeException(err, e);
        }
        finally {
            if (mManager != null) {
                mManager.shutdown();
            }
        }
    }


    private void init() {
        mBenchConfig.init();
        mTimers.start = System.currentTimeMillis();

        mLogger = new SimpleLogger();
        mManager = new ConnectionManager(mBenchConfig.getDatabaseProperties());
        mManager.setLogger(mLogger);

        Thread.currentThread().setName("BenchRunner");
        mAggregator = Util.makeInstance(mBenchConfig.getAggregatorClass());
    }


    /**
     * Setup portion of a test
     */
    private void setup() throws SQLException {
        mLogger.info("creating setup object "
                     + mBenchConfig.getSetupClass().getName());
        ISetup setup = Util.makeInstance(mBenchConfig.getSetupClass());
        baseObjectSetup(setup, true);

        setup.setScale(mBenchConfig.getScale());

        if (mBenchConfig.isStepEnabled(BenchConfig.STEP_SETUP)) {

            mLogger.info("creating schema");
            setup.createSchema();

            mLogger.info("filling tables");
            setup.fillTables();
        }
        else {
            mLogger.info("skipping table setup");
        }

        mLogger.info("executing warmup");
        setup.warmup();

        mLogger.info("setup complete");
    }



    /**
     * Configures an IConnectLog, with properties, a logger object,
     * and optinally a database connection.
     */
    private void baseObjectSetup(IConnectLog object,
                                 boolean needsConnection) 
        throws SQLException {

        assert object != null;
        if (object == null) {
            return;
        }

        if (needsConnection) {
            object.setConnection(mManager.getConnection());
        }

        object.setLogger(mLogger);
        object.setConfig(mBenchConfig.getTestProperties());
    }



    /**
     * Sets up a pair of base objects.  Both base objects
     * will share the same connection.
     */
    private void baseObjectSetup(IConnectLog a, IConnectLog b) 
        throws SQLException {

        assert a != null;
        assert b != null;

        baseObjectSetup(a, false);
        baseObjectSetup(b, false);

        Connection c = mManager.getConnection();
        a.setConnection(c);
        b.setConnection(c);
    }



    /**
     * Creates auditor objects, and runs the post-setup audit
     */
    private void preAudit() throws SQLException {
        if (! mBenchConfig.isStepEnabled(BenchConfig.STEP_AUDIT)) {
            mLogger.info("skipping post-setup audit");
            return;
        }
        
        mLogger.info("creating auditor object "
                     + mBenchConfig.getAuditorClass().getName());
        mTestAuditor = Util.makeInstance(mBenchConfig.getAuditorClass());
        
        mLogger.info("creating system auditor object "
                     + mBenchConfig.getSystemAuditorClass().getName());
        mSystemAuditor = Util.makeInstance(mBenchConfig.getSystemAuditorClass());
        baseObjectSetup(mTestAuditor, mSystemAuditor);

        mLogger.info("running post-setup audit");
        mTestAuditor.postSetup();
        mSystemAuditor.postSetup();
    }


    /**
     * Runs post-test audit
     */
    private void postAudit() throws SQLException {
        if (! mBenchConfig.isStepEnabled(BenchConfig.STEP_AUDIT)) {
            mLogger.info("skipping post-test audit");
            return;
        }
        assert mTestAuditor != null;
        assert mSystemAuditor != null;

        mSystemAuditor.postTest();
        mTestAuditor.postTest();

        mAggregator.addAuditorResults(mTestAuditor.getResults());
    }


    /**
     * Runs the workers.  Returns true if all workers completed
     * successfully, false otherwise 
     */
    private boolean runWorkers() throws SQLException {
        if (! mBenchConfig.isStepEnabled(BenchConfig.STEP_TEST)) {
            mLogger.info("skipping tests");
            return true;
        }

        mLogger.info("initializing foreman");
        mForeman = Util.makeInstance(mBenchConfig.getForemanClass());
        mForeman.setConfig(mBenchConfig);
        mForeman.setConnectionManager(mManager);
        mForeman.setLogger(mLogger);
        mForeman.init();

        runWorkers(mForeman);
        return collectWorkerResults(mForeman.getWorkers());

    }

    
    /**
     * Creates and starts worker threads
     */
    private void runWorkers(IForeman foreman) throws SQLException {
        mLogger.info("starting workers");

        mPeriodicCollector = new PeriodicCollector(foreman.getWorkers(),
                                                   foreman);
        mPeriodicCollector.setInterval(1);

        /* start workers */
        mTimers.startWarmup = System.currentTimeMillis();
        foreman.start();

        /* wait for warmup period to finish */
        try {
            Thread.sleep(mBenchConfig.getWarmupPeriod() * 1000);
        }
        catch (InterruptedException e) {
            /* not relevant here */
        }

        /* switch to measurement state */
        Thread periodicCollectorThread = new Thread(mPeriodicCollector,
                                                    "PeriodicCollector");
        mLogger.info("warmup done");
        periodicCollectorThread.start();

        mTimers.startMeasure = System.currentTimeMillis();
        BaseObject.TEST_STATE = BaseObject.TestState.MEASUREMENT;

        /* wait for measurement period to finish */
        try {
            Thread.sleep(mBenchConfig.getDuration() * 1000);
        }
        catch (InterruptedException e) {
            /* skip */
        }

        /* cooldown phase */
        mLogger.info("measurement done");
        mTimers.startCooldown = System.currentTimeMillis();
        BaseObject.TEST_STATE = BaseObject.TestState.COOLDOWN;

        /* wait for cooldown period to finish. */
        try {
            Thread.sleep(mBenchConfig.getCooldownPeriod() * 1000);
        }
        catch (InterruptedException e) {
            /* not relevant here */
        }

        /* switch to done state */
        mLogger.info("stopping threads");
        mTimers.stopTest = System.currentTimeMillis();
        BaseObject.TEST_STATE = BaseObject.TestState.DONE;

        /* wait for everyone to finish */
        foreman.shutdown();

        try {
            periodicCollectorThread.interrupt();
            periodicCollectorThread.join();
        }
        catch (InterruptedException e) {
            /* okay, stop waiting */
        }

        mLogger.info("workers finished");
    }


    /**
     * Collects the set of worker results and presents them
     * to the aggregator.
     *
     * Prints an error message if any worker encountered an
     * exception 
     *
     * @return true if all workers completed without error, false otherwise
     */
    @SuppressWarnings("unchecked")
    private boolean collectWorkerResults(List<IWorker> workers) {
        assert workers != null;

        int okay_workers = 0;
        int failed_workers = 0;
        List<Map<String,Object>> results = new ArrayList<Map<String,Object>>();
        
        for (IWorker worker : workers) {
            Exception e = worker.getError();
            if (e == null) {
                okay_workers++;
                /* not sure why the next line creates an unchecked warning.
                 * getBaseResults returns Map<String,Object>, which is the 
                 * generic type for results */
                results.add(worker.getBaseResults());
            }
            else {
                failed_workers++;
            }
        }
        mAggregator.addWorkerResults(results);
        mTimers.failedWorkers = failed_workers;

        List<Map<String,Object>> periodicResults = 
            new ArrayList<Map<String,Object>>();
        periodicResults.add(mPeriodicCollector.getResults());

        mAggregator.addWorkerResults(periodicResults);
        
        return (failed_workers == 0);
    }


    /**
     * Assemble all results in a ResultContainer.
     *
     * @see ResultContainer;
     */
    private void collectResults() {
        assert mResultContainer != null;

        mTimers.stopTest = System.currentTimeMillis();

        mResultContainer.addDatabaseConfig(mBenchConfig.getDatabaseProperties());
        mResultContainer.addTestConfig(mBenchConfig.getTestProperties());
        mResultContainer.addTestResults(mAggregator.getResults());

        if (mSystemAuditor != null) {
            mResultContainer.addServerConfig(mSystemAuditor.getConfiguration());
            mResultContainer.addServerCounters(mSystemAuditor.getCounters());
        }
        mResultContainer.addGlobals(getGlobals());

        calcDerivedResults();
    }


    private Map<String,Object> getGlobals() {
        Map<String,Object> globals = mBenchConfig.getBenchConfig();

        globals.put("z10_test_begin", new Timestamp(mTimers.start));
        globals.put("z20_warmup_begin", new Timestamp(mTimers.startWarmup));
        globals.put("z30_measurement_begin", new Timestamp(mTimers.startMeasure));
        globals.put("z40_cooldown_begin", new Timestamp(mTimers.startCooldown));
        globals.put("z50_test_end", new Timestamp(mTimers.stopTest));
        globals.put("z60_FAILED_threads", Integer.valueOf(mTimers.failedWorkers));

        double duration = (mTimers.startCooldown - mTimers.startMeasure)/1000.0;

        globals.put("test-duration", Double.valueOf(duration));
        return globals;
    }


    private void calcDerivedResults() {
        Map<String,Object> globals = mResultContainer.getGlobals();
        Map<String,Object> testResults = mResultContainer.getTestResults();
        int mpl = mBenchConfig.getMPL();
        
        Number total_transactions = 
            (Number) testResults.get(IWorker.TOTAL_TRANSACTIONS);

        Number test_duration = (Number) globals.get("test-duration");

        double overall_tps = 0;

        if (total_transactions != null && test_duration != null) {
            /* overall tps */
            overall_tps = total_transactions.doubleValue()
                / test_duration.doubleValue();
            
        }

        testResults.put("overall-tps", overall_tps);
        
        /* for backwards compatibility with programs that consume
         * bench results */
        testResults.put("avg-thread-duration-sec", globals.get("test-duration"));
    }

    private void saveResults() throws IOException {
        mResultContainer.writeResults(mBenchConfig.getResultFile());
    }

}


