/*
 * $Id: BenchTask.java,v 1.8 2011/06/04 00:40:27 srevilak Exp $
 */

package edu.umb.bench.framework;

import org.apache.tools.ant.Task;
import org.apache.tools.ant.BuildException;
import java.io.File;

/**
 * Ant interface to the benchmarking framework.
 *
 * <p>For the most part, the Task creates a {@link BenchConfig}, and 
 * passes that configuration on to a {@link BenchRunner}.
 *
 * @version $Revision: 1.8 $ $Date: 2011/06/04 00:40:27 $
 */
public class BenchTask extends Task {


    private BenchConfig benchConfig;
    private boolean mDryRun;

    public void init() throws BuildException {
        benchConfig = new BenchConfig();
        mDryRun = false;
    }


    public void execute() throws BuildException {

        if (mDryRun) {
            System.out.println(benchConfig);
            return;
        }

        benchConfig.init();
        BenchRunner runner = new BenchRunner(benchConfig);
        runner.runBenchmark();
    }


    /**
     * Sets the MPL (multi-programming level) for this benchmark run
     */
    public void setMPL(int mpl) {
        benchConfig.setMPL(mpl);
    }

    /**
     * Sets the duration of this test run, in number of transactions
     * per client thread.
     */
    public void setDuration(int duration) {
        benchConfig.setDuration(duration);
    }


    /**
     * Sets the scale for this test.  Scale is used to vary data
     * size.  (i.e., the scale of a TPC benchmark test).
     */
    public void setScale(double scale) {
        benchConfig.setScale(scale);
    }


    /**
     * Sets the warmup period for this benchmark run (in seconds).
     */
    public void setWarmupPeriod(int warmup) {
        benchConfig.setWarmupPeriod(warmup);
    }

    
    /**
     * Specifies the configuration file for a particular benchmark.
     */
    public void setTestConfig(File file) {
        benchConfig.setTestConfig(file);
    }

    
    /**
     * Specifies the database configuration for a benchmark run
     */
    public void setDatabaseConfig(File file) {
        benchConfig.setDatabaseConfig(file);
    }


    /**
     * Specifies the result file for the benchmark run
     */
    public void setResultFile(File file) {
        benchConfig.setResultFile(file);
    }


    /**
     * Sets the classname of the setup class.
     *
     * @see ISetup
     */
    public void setSetup(String name) {
        benchConfig.setSetupClassname(name);
    }

    /**
     * Sets the classname of the (test) auditor
     *
     * @see IAuditor
     */
    public void setAuditor(String name) {
        benchConfig.setAuditorClassname(name);
    }


    /**
     * Sets the classname for the (test) aggregator
     *
     * @see IAggregator
     */
    public void setAggregator(String name) {
        benchConfig.setAggregatorClassname(name);
    }

    
    /**
     * Sets the classname of the transaction factory.
     *
     * @see ITransactionFactory
     */
    public void setFactory(String name) {
        benchConfig.setFactoryClassname(name);
    }


    /**
     * Sets the name of the client worker class.
     *
     * @see IWorker
     */
    public void setWorker(String name) {
        benchConfig.setWorkerClassname(name);
    }


    /**
     * Sets the classname of the system auditor
     *
     * @see IAuditor
     * @see MySQLAuditor
     */
    public void setSystemAuditor(String name) {
        benchConfig.setSystemAuditorClassname(name);
    }


    /**
     * Sets the classname of the error interpreter.
     *
     * @see IErrorInterpreter
     * @see MySQLErrorInterpreter
     */
    public void setErrorInterpreter(String name) {
        benchConfig.setErrorInterpreterClassname(name);
    }


    /**
     * Sets the classname of the Foreman
     *
     * @see IForeman
     * @see FixedClientForeman
     * @see BaseForeman
     */
    public void setForeman(String name) {
        benchConfig.setForemanClassname(name);
    }

    
    /**
     * Sets the transaction 'thinktime' (in milliseconds).
     *
     * If <i>t</i> is the think time, then each thread will
     * sleep between [0.5<i>t</i>, 1.5<i>t</i>] after each commit or 
     * rollback.  Values &le; 0 disable think time.
     */
    public void setThinkTime(double ttime) {
        if (ttime > 0) {
            benchConfig.setThinkTime(ttime);
        }
        else {
            benchConfig.setThinkTime(0.0);
        }
    }


    /**
     * Set the list of steps to perform for this task.
     * 
     * <p>If unset, or set to the empty string, then all steps are performed.
     *
     * <p>Otherwise, this should be a string consisting of one
     * or more of the following characters:
     *
     * <ul>
     * <li><b>s</b> - to run the setup steps
     * <li><b>a</b> - to run the audit steps
     * <li><b>t</b> - to run the test steps
     * </ul>
     *
     * The order in which these characters appear does not matter.
     */
    public void setSteps(String steps) {
        if (steps == null || steps.length() == 0) {
            benchConfig.enableStep(BenchConfig.STEP_ALL);
            return;
        }

        
        benchConfig.disableStep(BenchConfig.STEP_ALL);
        String lc_steps = steps.toLowerCase();

        for (int iter = 0; iter < lc_steps.length(); iter++) {
            char s = lc_steps.charAt(iter);

            switch (s) {

            case 's':
                benchConfig.enableStep(BenchConfig.STEP_SETUP);
                break;

            case 'a':
                benchConfig.enableStep(BenchConfig.STEP_AUDIT);
                break;

            case 't':
                benchConfig.enableStep(BenchConfig.STEP_TEST);
                break;
                
            default:
                throw new BuildException("invalid step designation "
                                         + "'" + s + "' in"
                                         + " '" + lc_steps + "'");
            }
                
        }

    }


    /**
     * Do a dry run of the configuration?  If set to true,
     * the configuration is printed to standard output, but not executed.
     * This can be useful check when writing new test targets.
     */
    public void setDryrun(boolean yesno) {
        mDryRun = yesno;
    }
}
