/*
 * $Id: ConnectionManager.java,v 1.3 2010/02/09 02:42:19 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.sql.Connection;
import java.util.Properties;
import java.util.Set;
import java.util.HashSet;
import java.io.PrintWriter;
import java.sql.SQLException;
import java.sql.DriverManager;

/**
 * Sort of a connection pool for the benchmark testing framework; it's
 * not really a pool, so I've called it a manager.
 *
 * <p>
 * Connection manager serves two purposes.
 *
 * <ul>
 * <li>It creates connections.
 * <li>Removing a connection from the manager closes the connection.
 * </ul>
 *
 * Connection manager is a resource manger.  Each worker thread has a
 * single database connection.  The manager keeps a copy of each
 * active connection, so that all threads can be closed down at the
 * end of a test.
 */
public class ConnectionManager {

    /**
     * Property containing the driver class ({@value}).
     */
    public static final String DRIVER_CLASS = "bench.driver.class";

    /**
     * Property containing the jdbc url ({@value})
     */
    public static final String JDBC_URL = "bench.jdbc.url";

    private Properties mProperties;
    private String mDriverClass;
    private String mURL;
    private Set<Connection> mConnections;
    private boolean mInit;
    private PrintWriter mWriter;
    private ILogger mLogger;

    /**
     * Creates a new connection manager, using the driver properties
     * in props.
     *
     * <p>
     * props should contain all of the configuration information that
     * you'd like to pass to the driver, plus two additional properties:
     *
     * <ul>
     * <li>{@link #DRIVER_CLASS}.  The classname of the driver.
     * <li>{@link #JDBC_URL}.  The connection url
     * </ul>
     * The username and password must be contained in the properties,
     * or in the jdbc url.
     */
    public ConnectionManager(Properties props) {
        assert props != null;
        if (props == null) {
            throw new NullPointerException("`Properties props' cannot be null");
        }

        mProperties = props;

        String driver = mProperties.getProperty(DRIVER_CLASS);
        if (driver == null ||
            driver.length() == 0) {
            throw new IllegalArgumentException
                ("missing " + DRIVER_CLASS + " property");
        }

        String url = mProperties.getProperty(JDBC_URL);
        if (url == null || url.length() == 0) {
            throw new IllegalArgumentException
                ("missing " + JDBC_URL + " property");
        }

        mDriverClass = driver.trim();
        mURL = url.trim();

        mConnections = new HashSet<Connection>();
        mWriter = new PrintWriter(System.err);
    }


    public void setLogger(ILogger log) {
        mLogger = log;
    }


    /**
     * Initialize the manager, aka load the driver class.
     */
    private synchronized void init() {
        if (mInit) {
            return;
        }
        
        try {
            Class.forName(mDriverClass);
        }
        catch (ClassNotFoundException e) {
            throw new TypeNotPresentException(mDriverClass, e);
        }

        mInit = true;
    }
    


    /**
     * Returns a new connection.  The connection manager keeps a
     * reference to this connection, so that it can be closed later.
     *
     * <p> The returned connection is not read only, and has
     * autocommit turned off.
     */
    public Connection getConnection() throws SQLException {
        if (! mInit) {
            init();
        }

        int size = mConnections.size();

        Connection conn = DriverManager.getConnection(mURL, mProperties);
        assert conn != null;

        mConnections.add(conn);
        assert(size + 1 == mConnections.size());

        conn.setAutoCommit(false);
        conn.setReadOnly(false);

        /* note: connections are created with the `default' isolation
         * level.  We plan to set the default isolation via my.cnf */

        return conn;
    }


    /**
     * Same as {@link #getConnection()}.  Username and password are
     * ignored.
     */
    public Connection getConnection(String user, String pass) 
        throws SQLException {
        return getConnection();
    }


    /**
     * Returns zero.  This class does not use a login timeout
     */
    public int getLoginTimeout() {
        return 0;
    }

    
    /**
     * Does nothing. This class does not use login timeouts
     */
    public void setLoginTimeout(int timeout) {
        /* does nothing */
    }


    /**
     * Returns a pointer to the default logwriter (which is
     * a wrapped System.err)
     */
    public PrintWriter getLogWriter() {
        return mWriter;
    }

    
    /**
     * Sets the log writer for this Connection manager.
     */
    public void setLogWriter(PrintWriter writer) {
        if (writer != null) {
            mWriter = writer;
        }
    }

    
    /**
     * Closes a connection, removing it from the list of 
     * outstanding connections.
     */
    public void closeConnection(Connection conn) {
        assert conn != null;
        if (conn == null) {
            return;
        }

        boolean ok = mConnections.remove(conn);

        if (! ok) {
            mLogger.warn("attempt to remove unregistered connection " + conn);
        }

        closeIt(conn);
    }


    /**
     * Closes all connections.
     */
    public void shutdown() {

        for (Connection conn : mConnections) {

            if (conn == null) {
                continue;
            }
            closeIt(conn);
        }

        /* we're shutting down, so pretend they were all closed */
        mConnections.clear();
    }


    private void closeIt(Connection conn) {
        assert conn != null;

        try {
            if (conn.isClosed()) {
                mLogger.warn("found closed connection in pool " + conn);
            }
            else {
                conn.close();
            }
        }
        catch (SQLException e) {
            mLogger.error("error closing connection " + conn, e);
        }
    }
}
