/*
 * $Id: FixedClientForeman.java,v 1.1 2011/06/04 00:40:27 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.List;
import java.util.ArrayList;
import java.sql.SQLException;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;


/**
 * A foreman that creates a fixed number of client threads, and
 * let's them run on their own.  This was the original execution 
 * model for this benchmark program.
 *
 * This forman does not use triggers, before Commit, or at end hooks.
 *
 * @author srevilak
 * @since 6/3/2011
 */
public class FixedClientForeman extends BaseForeman {

    private List<IWorker> mWorkers;
    private ThreadPoolExecutor mPool;

    protected void childInit() {
        try {
            BenchConfig config = getConfig();
            mWorkers = new ArrayList<IWorker>(config.getMPL());
            
            for (int iter = 0; iter < config.getMPL(); iter++) {
                IWorker worker = createWorker();
                mWorkers.add(worker);
            }
        }
        catch (SQLException e) {
            getLogger().error("worker setup failed", e);
            throw new RuntimeException(e);
        }
    }




    /**
     * Starts the workers
     */
    public void start() {
        if (mWorkers == null) {
            throw new IllegalStateException("not initialized; no workers");
        }

        BenchConfig config = getConfig();
        ArrayBlockingQueue<Runnable> queue
            = new ArrayBlockingQueue<Runnable>(config.getMPL());

        mPool = 
            new ThreadPoolExecutor(config.getMPL(),  /* core size */
                                   config.getMPL(),  /* max size */
                                   0L, /* don't keep threads around after finish */
                                   TimeUnit.MILLISECONDS,
                                   queue);
        
        for (IWorker worker : mWorkers) {
            mPool.execute(worker);
        }
    }


    /**
     * This method waits until all workers are finished.
     */
    public void shutdown() {
        mPool.shutdown();
         
        while (true) {
            try {
                if (mPool.awaitTermination(1L, TimeUnit.SECONDS)) {
                    break;
                }
            }
            catch (InterruptedException e) {
                break;
            }
        }
    }


    /**
     * Returns the list of worker objects.
     */
    public List<IWorker> getWorkers() {
        return mWorkers;
    }


    /**
     * Returns the list of client threads connected to the database
     */
    public int getClientCount() {
        if (mWorkers != null) {
            return mWorkers.size();
        }
        return 0;
    }

}
