/*
 * $Id: GenericHotspot.java,v 1.3 2010/05/02 01:36:10 srevilak Exp $
 */

package edu.umb.bench.framework;

/**
 * A generic hotspot.
 *
 * <p>On configures the hotspot by providing three values:
 * <ul>
 * <li>The number of values.  For <i>n</i> values, we assume a
 * range of <i>[1, n]</i>
 *
 * <li>The number of values <i>h</i> that are in the hotspot.
 * Given <i>h</i>, <i>[1, h]</i> are 'hot rows', and <i>[h+1, n]</i>
 * are 'not hot'.
 *
 * <li>
 * The probability of using the hotspot, [0.0, 1.0].
 * </ul>
 *
 * If <i>h==0</i>, then all values are chosen using a uniform
 * random distribution from <i>[1, n]</i>.
 */
public class GenericHotspot implements IHotspot {

    /**
     * Number of rows in the table, [1, mTableSize].
     */
    private int mTableSize;

    /**
     * Hotspot.  [1, mHotRows]
     */
    private int mHotRows;

    private INumberGenerator mRandom;

    /**
     * Probability of choosing a row within hotspot
     */
    private double mProbability;

    public GenericHotspot(int nRows, int hotRows, double probability) {
        if (hotRows > nRows) {
            throw new IllegalArgumentException("must have hotRows <= nRows");
        }

        if (probability < 0.0 || probability > 1.0) {
            throw new IllegalArgumentException("invalid probability: " 
                                                + probability);
        }

        if (nRows < 0) {
            throw new IllegalArgumentException("invalid nRows: " + nRows);
        }
        
        if (hotRows < 0) {
            throw new IllegalArgumentException("invalid hotRows: " + hotRows);
        }

        mRandom = new GeneratorUnif();
        mTableSize = nRows;
        mHotRows = hotRows;
        mProbability = probability;
    }


    public int getValue() {
        double r1 = mRandom.getNumber();

        if (mHotRows == 0) {
            /* no hotspot.  Uniform access to table */
            return 1 + ((int) (r1 * mTableSize));
        }

        boolean hot = (r1 < mProbability);
        
        double r2 = mRandom.getNumber();

        if (hot) {
            return 1 + ((int) (r2 * mHotRows));
        }
        else {
            return 1 + mHotRows + (int) ((mTableSize - mHotRows) * r2);
        }
    }


    /**
     * @inheritDoc
     */
    public int getTableSize() {
        return mTableSize;
    }

    /**
     * @inheritDoc
     */
    public int getHotspotSize() {
        return mHotRows;
    }


    /**
     * @inheritDoc
     */
    public double getHotspotProbability() {
        return mProbability;
    }

    @Override public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("tableSize=");
        sb.append(getTableSize());
        sb.append(", hotspotSize=");
        sb.append(getHotspotSize());
        sb.append(", hotspotProbability=");
        sb.append(getHotspotProbability());
        return sb.toString();
    }

}
