/*
 * $Id: IWorker.java,v 1.11 2011/06/03 01:16:05 srevilak Exp $
 */

package edu.umb.bench.framework;
import java.util.Map;

/**
 * Interface for client worker threads
 *
 * @version $Revision: 1.11 $ $Date: 2011/06/03 01:16:05 $
 * @see BaseWorker
 */
public interface IWorker<T> extends IConnectLog, Runnable {

    public static final String TX_COMMITTED = "tx-committed";

    public static final String TX_ROLLBACK = "tx-rollback";

    public static final String TOTAL_TRANSACTIONS = "total-transactions";

    public static final String ERR_FCW = "errors-fcw";

    public static final String ERR_DEADLOCK = "errors-deadlock";

    public static final String ERR_CYCLE = "errors-cycle";

    public static final String ERR_PATH_TOO_LONG = "errors-path-too-long";

    public static final String ERR_TIMEOUT = "errors-timeout";

    public static final String SUM_THREAD_DURATION = "sum-thread-duration-ms";

    /**
     * Sets the transaction factory for this Worker.  
     */
    public void setTransactionFactory(ITransactionFactory<T> factory);

    /**
     * Sets the trigger for this worker.  If the trigger is not null,
     * then the worker should call trigger.wait() before executing
     * each transaction.
     */
    public void setTrigger(Object trigger);


    /**
     * Accumulates the results of executing a single transaction.
     */
    public void addResult(T result);


    /**
     * Sets the name of this worker
     */
    public void setName(String name);

    
    /**
     * Returns the name of this worker
     */
    public String getName();


    /**
     * Sets the think time for this worker thread, in milliseconds.
     *
     * @see BenchConfig#setThinkTime
     * @see BenchTask#setThinkTime
     */
    public void setThinkTime(double d);

    /**
     * Returns the think time for this worker thread.
     */
    public double getThinkTime();


    /**
     * Sets the id of this worker.  Every worker is given
     * a unique numeric id starting with 1.
     */
    public void setID(int id);


    /**
     * Returns the id of this worker.
     */
    public int getID();


    /**
     * Sets the error interpreter for this worker.
     */
    public void setErrorInterpreter(IErrorInterpreter interpreter);


    /**
     * This method allows the framework to (periodically) ask each
     * worker how many threads it has executed.  
     *
     * <p>This method should return the number of transactions since
     * the last time <tt>getPeriodicTransactionCount</tt> was called.
     * In other words, suppose each worker has a member variable named
     * <tt>periodicTransactionCount</tt>; this method should return
     * the value of the counter, and reset the counter to zero.
     *
     * <p><tt>periodicTransactionCount</tt> should include committed
     * and aborted transactions during the {@link
     * edu.umb.bench.framework.BaseObject.TestState#MEASUREMENT} period.
     */
    public int getPeriodicTransactionCount();


    /**
     * Gets the aggregated results of all transactions run by 
     * this worker.
     *
     * @return a map of Key/value pairs.  In general, the values will be simple
     * scalars: Strings, or subclasses of Number.
     */
    public Map<String,Object> getResults();


    /**
     * Get results from a worker base class.
     *
     * <p>The framework calls <tt>getBaseResults()</tt> (implemented
     * in a base worker class), and <tt>getBaseResults()</tt>
     * calls <tt>getResults()</tt>.
     *
     * <p>The base class adds two Integer fields:
     * {@link #TX_COMMITTED} and {@link #TX_ROLLBACK}
     */
    public Map<String,Object> getBaseResults();


    /**
     * If the worker failed, (i.e. encountered a fatal exception), 
     * then this method returns the exception.
     * Returns null if no fatal errors were encountered.
     *
     * @see IErrorInterpreter
     */
    public Exception getError();

    /**
     * Sets a Before Commit listener.
     *
     * @see IBeforeCommitListener
     */
    public void setBeforeCommitListener(IBeforeCommitListener listener);

    /**
     * Sets an atEnd Listener
     *
     * @see IEndListener
     */
    public void setEndListener(IEndListener listener);

}
