/*
 * $Id: MySQLAuditor.java,v 1.5 2010/02/10 02:49:53 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.Map;
import java.sql.SQLException;
import java.util.HashMap;
import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;

/**
 * MySQL system auditor.
 *
 * @version $Revision: 1.5 $ $Date: 2010/02/10 02:49:53 $
 */
public class MySQLAuditor extends BaseObject implements ISystemAuditor {

    private Map<String,Object> mStatusBefore;
    private Map<String,Object> mStatusAfter;
    private Map<String,Object> mConfiguration;

    public MySQLAuditor() {
        super();
        mStatusBefore = new HashMap<String,Object>();
        mStatusAfter = new HashMap<String,Object>();
        mConfiguration = new HashMap<String,Object>();
    }

    public void postSetup() throws SQLException {
        collectStatus(mStatusBefore);
    }

    public void postTest() throws SQLException {
        collectStatus(mStatusAfter);
        collectVariables(mConfiguration);
    }

    public Map<String,Object> getConfiguration() {
        return mConfiguration;
    }
    

    /**
     * Returns the value of mysql's "show status" counter.
     *
     * <p>Specifically, we run "show status" before and after 
     * tests, and report (after - before) for each counter value.
     *
     * <p>In other words, we return a map of counter deltas.
     */
    public Map<String,Object> getCounters() {
        Map<String,Object> delta = new HashMap<String,Object>();

        delta.putAll(mStatusAfter);

        Util.addMaps(delta, mStatusBefore, new Util.SubtractOperator());

        return delta;
    }


    /**
     * Runs "show status", and puts the results in map
     */
    private void collectStatus(Map<String,Object> map) throws SQLException {
        assert map != null;
        if (map == null) {
            return;
        }

        Connection conn = getConnection();
        Statement st = null;
        ResultSet rs = null;
        try {
            st = conn.createStatement();
            st.executeQuery("show status");

            rs = st.getResultSet();
            while (rs.next()) {
                String param = rs.getString(1);
                String value = rs.getString(2);

                if (ignoreStatusVariable(param)) {
                    continue;
                }

                Object ovalue = dwimCoerce(value);

                if (value != null) {
                    map.put(param, ovalue);
                }
            }
        }

        finally {
            if (rs != null) {
                rs.close();
            }

            if (st != null) {
                st.close();
            }

            conn.commit();
        }

    }


    /**
     * Returns true if we should ignore the status variable <tt>var</tt>.
     */
    private boolean ignoreStatusVariable(String var) {
        if (var == null
            || "Compression".equals(var)
            || "Rpl_status".equals(var)
            || "Slave_running".equals(var)
            || var.startsWith("Com_")
            || var.startsWith("Handler_")
            || var.startsWith("Ssl_")) {
            return true;
        }
        return false;
    }


    /**
     * Do What I Mean type conversion for strings.
     *
     * <p>In short, if <tt>value</tt> can obviously be coerced to a
     * number, then we perform that coersion and return a Long or Double.
     *
     * <p>Otherwise, we leave value as a dash.
     */
    private Object dwimCoerce(String value) {
        if (value == null) {
            return null;
        }

        int digits = 0;
        int dots = 0;
        int other = 0;
        int dashPos = 0;

        for (int iter = 0; iter < value.length(); iter++) {
            char c = value.charAt(iter);
            
            if (c == '.') {
                dots++;
            }
            else if (Character.isDigit(c)) {
                digits++;
            }
            else if (c == '-') {
                /* if we end up with dashPos != 0, then it's not a number */
                dashPos = iter;
            }
            else {
                other++;
            }
        }

        if (other > 0) {
            return value;
        }

        if (dashPos == 0 && dots == 0 && digits > 0) {
            /* a number */
            return Long.valueOf(value);
        }
        else if (dashPos == 0 && dots == 1 && digits > 0) {
            /* A floating point number.  Note: we don't try to interpret
             * scientific notation, because MySQL does not use it.
             */
            return Double.valueOf(value);
        }
        else {
            return value;
        }
    }


    /**
     * Collects MySQL system variables (aka "show variables")
     */
    private void collectVariables(Map<String,Object> variables) 
        throws SQLException {

        if (variables == null) {
            return;
        }

        Connection conn = getConnection();
        Statement st = null;
        ResultSet rs = null;
        try {
            st = conn.createStatement();
            st.executeQuery("show variables");
            rs = st.getResultSet();

            while (rs.next()) {
                String k = rs.getString(1);
                String v = rs.getString(2);
                if (v != null) {
                    variables.put(k, v);
                }
            }
        }
        finally {
            if (rs != null) {
                rs.close();
            }

            if (st != null) {
                st.close();
            }

            conn.commit();
        }

    }
}
