/*
 * $Id: PeriodicCollector.java,v 1.5 2011/06/04 19:57:51 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.util.Collections;

/**
 * A runnable that collects periodic per-worker statistics during
 * a benchmark run.
 *
 * @see IWorker#getPeriodicTransactionCount
 * @author srevilak
 */
class PeriodicCollector implements Runnable {

    private List<IWorker> mWorkers;
    private List<List<Integer>> mCommitCounts;
    private List<Integer>mClientCounts;
    private int mInterval;
    private int mCollections;
    private IForeman mForeman;

    public PeriodicCollector(List<IWorker> workers, IForeman foreman) {
        mWorkers = workers;
        mForeman = foreman;
        mInterval = 1;  /* one second */

        if (mWorkers == null) {
            throw new NullPointerException("List<IWorker> workers required");
        }
        if (foreman == null) {
            throw new NullPointerException("IForeman required");
        }

        mCollections = 0;
    }
    
    
    /**
     * Sets the collection interval, in seconds.
     */
    public void setInterval(int interval) {
        if (interval > 0) {
            mInterval = interval;
        }
    }
    

    /**
     * Returns the collection interval
     */
    public int getInterval() {
        return mInterval;
    }

    
    public void run() {
        setup();
        
        while (true) {
            collect();

            try {
                Thread.currentThread().sleep(mInterval * 1000L);
            }
            catch (InterruptedException e) {
                /* stop sleeping */
            }
        
            if (BaseObject.TEST_STATE == BaseObject.TestState.DONE) {
                return;
            }
        }
    }


    private void setup() {
        mCommitCounts = new ArrayList<List<Integer>>(mWorkers.size());

        for (int iter = 0; iter < mWorkers.size(); iter++) {
            /* 512 should be an adequate starting capactiy */
            List<Integer> alist = new ArrayList<Integer>(512);
            mCommitCounts.add(alist);
        }

        mClientCounts = new ArrayList<Integer>(mWorkers.size());
    }


    /**
     * Collects information from each worker
     */
    private void collect() {
        if (BaseObject.TEST_STATE != BaseObject.TestState.MEASUREMENT) {
            return;
        }

        mCollections++;
        
        for (int iter = 0; iter < mWorkers.size(); iter++) {
            collectOneWorker(iter);
        }

        int clientCount = mForeman.getClientCount();
        mClientCounts.add(Integer.valueOf(clientCount));
    }


    /**
     * Collects information from a single worker.
     */
    private void collectOneWorker(int iter) {

        IWorker worker = mWorkers.get(iter);

        /* see if new workers were added to the system */
        int size = mCommitCounts.size();
        int need = mWorkers.size() - size;

        for (int adding = 0; adding < need; adding++) {
            List<Integer> template = mCommitCounts.get(0);
            List<Integer> newList = new ArrayList<Integer>(2 * template.size());

            /* left pad the new list */
            for (int z = 0; z < template.size(); z++) {
                newList.add(0);
            }
            mCommitCounts.add(newList);
        }

        List<Integer> counts = mCommitCounts.get(iter);
        
        int txcount = worker.getPeriodicTransactionCount();

        if (mCollections == 1) {
            /* The first measurement is likely to be a partial
             * second of activity.  Throw it out */
            return;
        }

        counts.add(Integer.valueOf(txcount));
    }


    public Map<String,Object> getResults() {
        Map<String,Object> result = new HashMap<String,Object>();

        /* make sure we report uniform series for each transaction */
        int max = 0;
        for (int iter = 0; iter < mCommitCounts.size(); iter++) {
            List<Integer> counts = mCommitCounts.get(iter);
            if (counts.size() > max) {
                max = counts.size();
            }
        }

        for (int iter = 0; iter < mWorkers.size(); iter++) {

            IWorker worker = mWorkers.get(iter);
            List<Integer> counts = Collections.emptyList();

            if (iter < mCommitCounts.size()) {
                counts = mCommitCounts.get(iter);
            }

            String key = worker.getName() + "_tx_run_each_second";
            String value = formatCounts(max, counts);
            result.put(key, value);

        }

        /* compute tx run each second over all transactions */
        int sum_tps[] = new int[max];
        for (int iter = 0; iter < mWorkers.size(); iter++) {
            IWorker worker = mWorkers.get(iter);
            List<Integer> counts = Collections.emptyList();

            if (iter < mCommitCounts.size()) {
                counts = mCommitCounts.get(iter);
            }

            for (int sum_idx = 0; sum_idx < counts.size(); sum_idx++) {
                if (sum_idx < sum_tps.length) {
                    sum_tps[sum_idx] += counts.get(sum_idx).intValue();
                }
            }
        }

        List<Integer> sum_list = new ArrayList<Integer>(sum_tps.length);
        for (int iter = 0; iter < sum_tps.length; iter++) {
            sum_list.add(Integer.valueOf(sum_tps[iter]));
        }

        result.put("total_tx_run_each_second",
                   formatCounts(max, sum_list));

        result.put("client_count_each_second",
                   formatCounts(max, mClientCounts));

        int median_clients = findMedian(mClientCounts);

        result.put("client_count_median", Integer.valueOf(median_clients));

        return result;
    }


    /**
     * Given list of counts, format them as comma-separated values.
     *
     * @param nEntries the number of entries that the comma-separated list
     * should have.  If necessary, this method will zero-pad the right
     * of the output string.
     */

    private String formatCounts(int nEntries, List<Integer> counts) {
        StringBuilder sb = new StringBuilder(nEntries * 4);

        for (int iter = 0; iter < nEntries; iter++) {
            int value = 0;

            if (iter < counts.size()) {
                value = counts.get(iter);
            }

            if (sb.length() > 0) {
                sb.append(',');
            }

            sb.append(value);
        }

        return sb.toString();
    }


    private int findMedian(List<Integer> list) {
        if (list == null || list.size() == 0) {
            return 0;
        }

        ArrayList<Integer> copy = new ArrayList<Integer>(list.size());
        copy.addAll(list);
        Collections.sort(copy);
    
        /* if odd, return the one in the middle */
        if (copy.size() % 2 == 1) {
            return copy.get(copy.size() / 2).intValue();
        }

        /* if even, return the median of the two middle elements */
        int a = copy.get(copy.size() / 2);
        int b = copy.get( (copy.size()/2) - 1 );
        
        return (a + b) / 2;
    }

}
