/*
 * $Id: Probability.java,v 1.3 2010/04/26 01:19:32 srevilak Exp $
 */

package edu.umb.bench.framework;

/**
 * Utilities for working with probability distributions
 */
public class Probability {

    
    /**
     * Given a (perhaps) unscaled probability distribution <tt>in</tt>,
     * scale it so that sum(in) == 1.0.
     *
     * <p> For example, <tt>[1, 2, 1]</tt> should be scaled
     * as <tt>[0.25, 0.5, 0.25]</tt>.
     */
    public static double[] scale(double in[]) {
        if (in == null || in.length == 0) {
            return in;
        }

        double sum = 0.0;
        for (int iter = 0; iter < in.length; iter++) {
            sum += in[iter];
        }

        /* already scaled? */
        if (sum == 1.0) {
            return in;
        }

        /* scale each element of in by sum */
        double out[] = new double[in.length];
        double cum_sum = 0.0;
        for (int iter = 0; iter < in.length - 1; iter++) {
            out[iter] = in[iter] / sum;
            cum_sum += out[iter];
        }

        /* Element n should be 1 - (sum of elements [1, n-1]) */
        out[out.length - 1] = 1.0 - cum_sum;

        return out;
    }
    
    
    /**
     * Given a probability distribution, select an element, returning
     * the index of the element selected.
     * 
     * @param probabilities a probability distribution
     */
    public static int pickElement(double probabilities[]) {
        assert probabilities != null; 
        double rand = Math.random();
        double cumsum = 0.0;
        
        /*
         * general strategy -- iterate over the array, calculating
         * a cumulative sum.  We pick the first element that causes
         * the cumulative sum to exceed the random number
         */
        for (int iter = 0; iter < probabilities.length; iter++) {
            cumsum += probabilities[iter];
            if (cumsum > rand) {
                return iter;
            }
        }

        return (probabilities.length - 1); 
    }


    /**
     * Given a string of the form
     *
     * <pre>
     * a,b,c,d[,..]
     * </pre>
     * parse and return a scaled probability distribution.
     *
     *<p>a,b,c,d are treated as floating point numbers.
     * Theses should be parseable by Double.parseDouble.
     *
     * @exception NumberFormatException if (a) an element of the
     * string is not parseable via Double.parseDouble or (b) if the
     * parsed string produces an empty list of elements.
     */
    public static double[] parse(String in) {
        if (in == null) {
            throw new NullPointerException("input distribution cannot be null");
        }

        String tokens[] = in.trim().split("\\s*,\\s*");
        if (tokens == null || tokens.length == 0) {
            throw new NumberFormatException("no numbers in probability string "
                                            + "'" + in + "'");
        }

        double unscaled[] = new double[tokens.length];
        for (int iter = 0; iter < unscaled.length; iter++) {
            unscaled[iter] = Double.parseDouble(tokens[iter]);
        }

        return scale(unscaled);

    }

}
