/*
 * $Id: RandomHotspot.java,v 1.4 2010/06/26 19:46:17 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;
import java.util.Collections;

/**
 * Hotspot that selects rows so that they are randomly distributed across
 * a table, rather than confined to a continuous range.
 *
 * <p>This class works as follows.  For a hotspot of <i>n</i> rows, we
 * create an array of <i>n</i> unique integers.  To choose a hot
 * value, we randomly select and return a member of this array.  To
 * choose a non-not value, we attempt to select a number that is not a
 * member of the hotspot array.
 * 
 * <p>Non hot rows are choosen as follows:
 * <pre>
 * for (1..10):
 *     select random number
 *     if number does not appear in hot array:
 *         return number
 *     endif
 * end for
 * return any old random number within the table size
 * </pre>
 *
 * <p>
 * This should be adequate if the hotspot is a small portion of he table.
 * If the hotspot is a majority of the table, then we may "give up" and
 * return hot values more frequently that the hotspot probability.
 *
 */
public class RandomHotspot implements IHotspot {

    private int mTableSize;

    private int mHotRows;

    private INumberGenerator mRandom;

    private double mProbability;

    /**
     * Random selection of rows that comprises the hotspot
     */
    private int mRows[];
    private Set<Integer> mRowsSet;

    /**
     * @param nRows the number of possible values, [1, nRows]
     *
     * @param hotRows the number of hot values
     *
     * @param probability the probability of choosing a hot row
     */
    public RandomHotspot(int nRows, int hotRows, double probability) {

        if (hotRows > nRows) {
            throw new IllegalArgumentException("must have hotRows <= nRows");
        }

        if (probability < 0.0 || probability > 1.0) {
            throw new IllegalArgumentException("invalid probability: " 
                                                + probability);
        }

        if (nRows < 0) {
            throw new IllegalArgumentException("invalid nRows: " + nRows);
        }
       
        if (hotRows < 0) {
            throw new IllegalArgumentException("invalid hotRows: " + hotRows);
        }

        mRandom = new GeneratorUnif();
        mTableSize = nRows;
        mHotRows = hotRows;
        mProbability = probability;

        setup();
    }


    /**
     * @param nRows The number of table rows. (More precisely, the N in [1,N],
     * where values are drawn from the range [1,N].
     *
     * @param hotRows a pre-determined set of hot values.  Each element in
     * hotRows must be distinct.
     *
     * @param probability the probability of choosing a hot value
     */
    public RandomHotspot(int nRows, int hotRows[], double probability) {
        if (probability < 0.0 || probability > 1.0) {
            throw new IllegalArgumentException("invalid probability: " 
                                                + probability);
        }

        if (nRows < 0) {
            throw new IllegalArgumentException("invalid nRows: " + nRows);
        }
       
        if (hotRows == null) {
            throw new IllegalArgumentException("hotRows cannot be null");
        }

        mRandom = new GeneratorUnif();
        mTableSize = nRows;
        mHotRows = hotRows.length;
        mProbability = probability;
        setup(hotRows);
    }


    private void setup() {
        mRowsSet = new HashSet<Integer>(mHotRows);
        mRows    = new int[mHotRows];

        /* choose set of rows, without repetition */
        int iter = 0;
        while (iter < mRows.length) {

            int value = (int) ( 1 + (mRandom.getNumber() * mTableSize));
            Integer valueO = Integer.valueOf(value);

            if (mRowsSet.contains(valueO)) {
                continue;
            }

            mRows[iter++] = value;
            mRowsSet.add(valueO);
        }
    }


    private void setup(int hotRows[]) {
        mRowsSet = new HashSet<Integer>(hotRows.length);
        mRows    = new int[hotRows.length];
        System.arraycopy(hotRows, 0, mRows, 0, mRows.length);

        for (int iter = 0; iter < mRows.length; iter++) {
            Integer val = Integer.valueOf(mRows[iter]);
            if (mRowsSet.contains(val)) {
                throw new IllegalArgumentException("duplicate hot value: "
                                                   + val);
            }
            mRowsSet.add(val);
        }
    }


    /**
     * Returns a read-only copy of our hotspot.
     */
    public Set<Integer> getHotspot() {
        return Collections.unmodifiableSet(mRowsSet);
    }

    /**
     * @inheritDoc
     */
    public int getValue() {
        double in_hot = mRandom.getNumber();

        /* no hot spot.  Pick anything */
        if (mRows.length == 0) {
            return (int) (1 + (mRandom.getNumber() * mTableSize));
        }

        /* choose from hotspot */
        if (in_hot < mProbability) {
            int index = (int) (mRandom.getNumber() * mRows.length);
            return mRows[index];
        }

        /* try to choose a value that's not in the hotspot */
        for (int iter = 0; iter < 10; iter++) {

            int value = (int) (1 + (mRandom.getNumber() * mTableSize));
            Integer iValue = Integer.valueOf(value);

            if (! mRowsSet.contains(iValue)) {
                return value;
            }
        }

        /* give up and choose anything */
        return (int) (1 + (mRandom.getNumber() * mTableSize));
    }
    

    /**
     * @inheritDoc
     */
    public int getTableSize() {
        return mTableSize;
    }


    /**
     * @inheritDoc
     */
    public int getHotspotSize() {
        return mRows.length;
    }


    /**
     * @inheritDoc
     */
    public double getHotspotProbability() {
        return mProbability;
    }


    public String toString() {
        String setStr = mRowsSet.toString();
        StringBuilder sb = new StringBuilder(setStr.length() + 64);
        
        sb
            .append("tableSize=")
            .append(getTableSize())
            .append(", hotRows=")
            .append(getHotspotSize())
            .append(", probability=")
            .append(getHotspotProbability())
            .append(", rows=")
            .append(setStr);

        return sb.toString();
    }

}
