/*
 * $Id: ResultContainer.java,v 1.3 2010/04/21 16:15:28 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.Map;
import java.io.File;
import java.util.Properties;
import java.io.IOException;
import java.io.Writer;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.sql.Timestamp;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Arrays;

/**
 * Test for gathering all configuration and test results,
 * and recording them in a result file.
 *
 * @version $Revision: 1.3 $ $Date: 2010/04/21 16:15:28 $
 */
public class ResultContainer {

    private Map<String,Object> mDatabaseConfig;
    private Map<String,Object> mTestConfig;
    private Map<String,Object> mTestResults;

    private Map<String,Object> mServerConfig;
    private Map<String,Object> mServerCounters;
    private Map<String,Object> mGlobals;

    private final String NL;

    public ResultContainer() {
        NL = System.getProperty("line.separator");
    }

    /**
     * Adds the JDBC database configuration properties
     *
     * @see BenchConfig#getDatabaseProperties
     */
    public void addDatabaseConfig(Properties props) {
        mDatabaseConfig = propertiesToMap(props);
    }


    /**
     * Adds the benchmark test configuration properties.
     *
     * @see BenchConfig#getTestProperties
     */
    public void addTestConfig(Properties props) {
        mTestConfig = propertiesToMap(props);
    }



    /**
     * Converts a properties table to a map of <String,Object>
     */
    private Map<String,Object> propertiesToMap(Properties props) {
        if (props == null) {
            return null;
        }

        Map<String,Object> map = new HashMap<String,Object>();

        Enumeration<?> pnames = props.propertyNames();

        while (pnames.hasMoreElements()) {

            Object oname = pnames.nextElement();

            if (! (oname instanceof String)) {
                continue;
            }

            Object value = props.getProperty((String) oname);

            map.put((String) oname, value);
        }

        return map;
    }


    /**
     * Adds the results of a benchmark test
     *
     * @see IAggregator#getResults
     */
    public void addTestResults(Map<String,Object> results) {
        mTestResults = results;
    }


    /**
     * Returns the set of test results 
     */
    public Map<String,Object> getTestResults() {
        return mTestResults;
    }


    /**
     * Adds the database server configuration, as a set of key/value pairs.
     *
     * @see ISystemAuditor#getConfiguration
     */
    public void addServerConfig(Map<String,Object> config) {
        mServerConfig = config;
    }


    /**
     * Adds the database server performance counters
     *
     * @see ISystemAuditor#getCounters
     */
    public void addServerCounters(Map<String,Object> counters) {
        mServerCounters = counters;
    }


    /**
     * Adds global configuration parameters and results.
     * This includes parameters supplied by
     * {@link BenchConfig}, and other measurements
     * taken directly by {@link BenchRunner}.
     *
     * @see BenchConfig#getBenchConfig
     */
    public void addGlobals(Map<String,Object> globals) {
        mGlobals = globals;
    }


    /**
     * Returns the set of globals.
     */
    public Map<String,Object> getGlobals() {
        return mGlobals;
    }

    /**
     * Writes the collection of results to a file.
     */
    public void writeResults(File file) throws IOException {
        if (file == null) {
            throw new NullPointerException("file");
        }

        Writer writer = null;
        boolean need_separator = file.exists();
        try {
            writer = new BufferedWriter(new FileWriter(file, true));
            if (need_separator) {
                writeSeparator(writer);
            }
            writeResults(writer);
        }
        finally {
            if (writer != null) {
                writer.close();
            }
        }
    }


    /**
     * If we are appending to a file, make a clean break
     * between different sets of results
     */
    private void writeSeparator(Writer writer) throws IOException {
        writer.write(NL);
        writer.write('\f');
        writer.write(NL);
    }    


    /**
     * Writes the collection of results to a Writer.
     *
     * <p>The caller is responsible for closing the writer.
     */
    public void writeResults(Writer writer) throws IOException {
        writeHeader(writer);

        writeMap("globals", mGlobals, writer);
        writeMap("test-config", mTestConfig, writer);
        writeMap("test-results", mTestResults, writer);
        writeMap("jdbc-config", mDatabaseConfig, writer);
        writeMap("mysqld-config", mServerConfig, writer);
        writeMap("mysqld-counter", mServerCounters, writer);

        writer.flush();
    }


    /**
     * Write a short header for the result file
     */
    private void writeHeader(Writer writer) throws IOException {
        Timestamp ts = new Timestamp(System.currentTimeMillis());
        writer.write("### Results recorded " + ts);
        writer.write(NL);
    }


    /**
     * Write a map to the result file.  Each line is pre-pended with
     * the category "category".
     */
    private void writeMap(String category, Map<String,Object> map,
                          Writer writer) throws IOException {
        if (map == null) {
            return;
        }

        StringBuilder sb = new StringBuilder(4096);

        String keys[] = map.keySet().toArray(new String[map.size()]);
        if (keys.length == 0) {
            return;
        }

        Arrays.sort(keys);
        sb.append(NL);

        for (String key : keys) {
            Object value = map.get(key);

            sb.append(category);
            sb.append('\t');
            sb.append(key);
            sb.append('\t');
            if (value != null) {
                sb.append(value);
            }
            sb.append(NL);
        }
        writer.write(sb.toString());
    }
}
