/*
 * $Id: TargetActiveForeman.java,v 1.6 2011/06/10 01:58:37 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;
import java.sql.SQLException;
import java.util.Collections;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * A foreman that tries to maintain a target number of "active"
 * transactions, where the definition of "active" does not include
 * time spent waiting for <tt>COMMIT</tt> to return.
 * 
 * @author srevilak
 * @since 6/4/2011
 */
public class TargetActiveForeman extends BaseForeman {

    /**
     * Every single worker that we know about
     */
    private List<IWorker> mAllWorkers;

    /**
     * Pool of worker + threads.  To unpool, we take from
     * the head.  To re-pool we add to the tail.
     */
    private LinkedList<Cons> mPool;

    /**
     * Number of "active" threads.  These threads are executing SQL
     * statements, but are not in the process of committing.
     */
    private AtomicInteger mActiveCount;

    /**
     * The set of in-use threads.  These threads can be active, or
     * waiting to commit.
     *
     * We should have mUsed.size() + mPool.size() == mAllWorkers.size()
     */
    private Map<String,Cons> mUsed;

    /**
     * Target number of active (but not committing) threads, aka target MPL.
     *
     * We try to keep mActiveCount == mTargetActive 
     */
    private int mTargetActive;

    /**
     * Transaction starter
     */
    private TransactionStarter mStarter;
    private Thread mStarterThread;

    private int mStartedFromPool = 0;
    private int mStartedFromScratch = 0;

    protected void childInit() {
        BenchConfig config = getConfig();

        mTargetActive = config.getMPL();

        mAllWorkers = Collections.synchronizedList
            (new ArrayList<IWorker>(mTargetActive * 2));

        /* mPool must be manually synchronized */
        mPool = new LinkedList<Cons>();

        mActiveCount = new AtomicInteger();

        mUsed = Collections.synchronizedMap(new HashMap<String,Cons>());

        mStarter = new TransactionStarter(this);
        mStarterThread = new Thread(mStarter);

        /*
         * Create our workers, and put them in the pool.
         * Technically, this understocks the pool; we'll let the 
         * pool fill in naturally.
         */
        try {
            for (int iter = 0; iter < mTargetActive; iter++) {
                IWorker worker = createWorker();
                Thread thread = new Thread(worker);
                Cons cons = new Cons(worker, thread);
                
                mAllWorkers.add(worker);
                mPool.addLast(cons);
            }
        }
        catch (SQLException e) {
            getLogger().error("exception during init: ", e);
            throw new RuntimeException(e);
        }

    }


    /**
     * Start all of the active workers.
     */
    public void start() {
        ArrayList<Cons> startingSet = new ArrayList<Cons>(mTargetActive);

        /*
         * Unpool the set of workers that we are 
         * about to start.  Mark them in-use.
         */
        for (int iter = 0; iter < mTargetActive; iter++) {
            Cons cons = mPool.removeFirst();
            mUsed.put(cons.worker.getName(), cons);
            startingSet.add(cons);
        }

        /* Must set active count before starting the starter */
        mActiveCount.set(startingSet.size());
        mStarterThread.start();

        try {
            Thread.currentThread().sleep(100);
        }
        catch (InterruptedException e) {
            /* ignored */
        }

        /* now, start the threads */
        for (Cons cons : startingSet) {
            cons.thread.start();
        }
    }


    /**
     * This method waits until all workers are finished.
     */
    public void shutdown() {

        getLogger().info("waiting for workers-in-use to finish");

        while (mUsed.size() > 0) {

            try {
                Thread.currentThread().sleep(1000);
            }
            catch (InterruptedException e) {
                /* ignore */
            }

            if (mUsed.size() == 0) {
                break;
            }

            getLogger().info("still waiting for workers in use ("
                             + mUsed.size() + ")");

            herdKittens(mUsed);
        }
        

        /* now, clean the pool.  First, let's notify everyone */
        getLogger().info("shutting down thread pool");

        for (Cons cons : mPool) {
            synchronized(cons.worker) {
                cons.worker.notifyAll();
            }
        }

        int collected = 0;
        while (mPool.size() > 0) {
            Cons cons = mPool.removeFirst();

            try {
                cons.thread.join(100);
            }
            catch (InterruptedException e) {
                /* ignore */
            }

            if (cons.thread.isAlive()) {
                /* nudge thread.  Try again laster */
                synchronized (cons.worker) {
                    cons.worker.notifyAll();
                }
                synchronized (cons.thread) {
                    cons.thread.interrupt();
                }
                mPool.addLast(cons);
            }
            else {
                collected++;
            }
        }

        if (collected != mAllWorkers.size()) {
            getLogger().warn("WARN: wrong number of threads collected? "
                             + "expected=" + mAllWorkers.size()
                             + ", got=" + collected);
        }

        getLogger().info("started from pool=" + mStartedFromPool);
        getLogger().info("started from scratch=" + mStartedFromScratch);

        /* and wait for the starter thread to finish */
        while (mStarterThread.isAlive()) {
            try {
                mStarterThread.interrupt();
                mStarterThread.join(100);
            }
            catch (InterruptedException e) {
                /* okay */
            }
        }
    }



    /**
     * "Interrupt used threads" might have been a more descriptive
     * name, but herding kittens sounds more appealing.  Either way,
     * probably not much difference.
     */
    private void herdKittens(Map<String,Cons> used) {
        for (Map.Entry<String,Cons> entry : used.entrySet()) {

            Cons cons = entry.getValue();

            IWorker worker = cons.worker;

            Thread thread = cons.thread;

            synchronized (worker) {
                worker.notifyAll();
                thread.interrupt();
            }
        }
    }


    /**
     * Returns the list of worker objects.
     */
    public List<IWorker> getWorkers() {
        return mAllWorkers;
    }



    public int getClientCount() {
        synchronized (mPool) {
            return mUsed.size();
        }
    }


    private void startTransaction() {
        try {
            startTransactionSub();
        }
        catch (SQLException e) {
            getLogger().error("Error starting transaction", e);
            throw new RuntimeException(e);
        }
    }


    private void startTransactionSub() throws SQLException {
       
        Cons cons = null;
        boolean usedPool = false;

        synchronized (mPool) {
            if (mPool.size() > 0) {
                usedPool = true;
                cons = mPool.removeFirst();
            }
        }

        if (cons == null) {
            IWorker worker = createWorker();
            mAllWorkers.add(worker);
            cons = new Cons(worker, new Thread(worker));
        }

        String name = cons.worker.getName();
        if (mUsed.containsKey(name)) {
            throw new IllegalStateException("duplicate worker name: " + name);
        }
        mUsed.put(name, cons);

        /* If the thread was already running, then trigger it.
         * Otherwise, just start it */
        if (usedPool) {
            synchronized (cons.worker) {
                mStartedFromPool++;
                cons.worker.notify();
            }
        }
        else {
            mStartedFromScratch++;
            cons.thread.start();
        }
    }

    
    private void handleEnd(IWorker worker) {
        String name = worker.getName();
        
        Cons cons = mUsed.remove(name);
        if (cons == null) {
            /* At the conclusion of a test, the "end" might come after
             * waiting for notification, instead of after executing */
            return;
        }

        synchronized (mPool) {
            mPool.addLast(cons);
        }
    }
    

    @Override protected final IWorker createWorker() throws SQLException {
        IWorker worker = super.createWorker();
        worker.setTrigger(worker);
        worker.setBeforeCommitListener(new BeforeCommit(this, mStarter));
        worker.setEndListener(new EndListener(this));
        return worker;
    }



    /*******************************************************************
     */

    private static class TransactionStarter implements Runnable {
        
        private TargetActiveForeman foreman;
        
        public TransactionStarter(TargetActiveForeman foreman) {
            this.foreman = foreman;
        }

        public void run() {
            try {
                runInternal();
            }
            catch (InterruptedException e) {
                /* we are done */
            }
            
        }
         
        private void runInternal() throws InterruptedException {
            while (true) {

                if (BaseObject.TEST_STATE == BaseObject.TestState.DONE) {
                    return;
                }

                int active = 0;
                int delta = 0;
                
                /**
                 * Figure out how many transactions to start,
                 * to maintain the target number of active transactions.
                 *
                 * Note: BeforeCommit decrements active count.
                 */
                active = foreman.mActiveCount.get();
                    
                delta = foreman.mTargetActive - active;
                
                foreman.mActiveCount.addAndGet(delta);
                
                /** now, start that many transactions */
                for (int iter = 0; iter < delta; iter++) {
                    foreman.startTransaction();
                }
                
                Thread.yield();
            }
        }
    }


    /**
     * A "before commit" listener.  This listener starts another
     * transaction running when the notifier is about to commit.
     */
    private static class BeforeCommit implements IBeforeCommitListener {
        private TargetActiveForeman foreman;
        private TransactionStarter starter;

        public BeforeCommit(TargetActiveForeman f, TransactionStarter starter) {
            this.foreman = f;
            this.starter = starter;
        }

        public void beforeCommit(Object notifier) {
            if (! (notifier instanceof IWorker)) {
                throw new IllegalArgumentException("not IWorker: " + notifier);
            }

            foreman.mActiveCount.decrementAndGet();
        }
    }


    /**
     * Listener that finishes up after a transaction.
     * The transaction has finished, so it is not part
     * of the active count.
     *
     * Take the transaction out of the used list, and put it back
     * in the pool
     */
    private static class EndListener implements IEndListener {

        TargetActiveForeman foreman;

        public EndListener(TargetActiveForeman f) {
            this.foreman = f;
        }

        public void atEnd(Object notifier) {
            if (! (notifier instanceof IWorker)) {
                throw new IllegalArgumentException("not an IWorker: " + notifier);
            }

            IWorker worker = (IWorker) notifier;
            foreman.handleEnd(worker);
        }
    }

    /**
     * A coupling of a worker to its thread
     */
    private static class Cons {
        public IWorker worker;
        public Thread thread;

        public Cons(IWorker worker, Thread thread) {
            this.worker = worker;
            this.thread = thread;
        }
    }





}
