/*
 * $Id: Util.java,v 1.9 2010/06/20 01:41:28 srevilak Exp $
 */

package edu.umb.bench.framework;

import java.util.HashMap;
import java.util.Map;
import java.util.List;
import java.util.Properties;
import java.io.File;
import java.io.IOException;
import java.io.FileInputStream;

/**
 * Assorted utility functions.
 *
 *
 * @version $Revision: 1.9 $ $Date: 2010/06/20 01:41:28 $
 */
public class Util {


    /**
     * Operator for combining map values.
     */
    public interface MapOperator {
        public int op(int x, int y);
        public long op(long x, long y);
        public double op(double x, double y);
    }

    
    /**
     * Map operator that adds numeric values.  (x + y)
     */
    public static class AddOperator implements MapOperator {
        public int op(int x, int y) {
            return x + y;
        }

        public long op(long x, long y) {
            return x + y;
        }

        public double op(double x, double y) {
            return x + y;
        }
    }


    /**
     * Map operator that subtracts numeric values (x - y)
     */
    public static class SubtractOperator implements MapOperator {
        public int op(int x, int y) {
            return x - y;
        }

        public long op(long x, long y) {
            return x - y;
        }

        public double op(double x, double y) {
            return x - y;
        }
    }



    /**
     * Equivalent to <tt>addMaps(in, out, null);</tt>
     *
     */
    public static void addMaps(Map<String,Object> out, Map<String,Object> in) {
        addMaps(out, in, null);
    }


    /**
     * Combines two maps.  The semantics are
     *
     * <pre>
     * out OP= in
     * </pre>
     *
     * The combination is done on a key by key basis.
     *
     * <p><tt>op</tt> dictates how the combination occurs.  For
     * example, +=, -=, etc.  If no operator is given, += is used.
     *
     * <ul>
     * <li> if in contains K and out does not, then add K,V to out.
     * <li> if V is a Number in both in, out then add in's V to out's v.
     * <li> in V is a String, then replace in's V with out's V
     * <li> if one V is a number and the other is not, throw
     * an IllegalArgumentException.
     *
     * @exception IllegalArgumentException as described above.
     * 
     * @exception NullPointerException if out is null.
     */
    public static void addMaps(Map<String,Object> out, Map<String,Object> in,
                               MapOperator op) {

        if (out == null) {
            throw new NullPointerException("missing `Map out'");
        }

        if (in == null) {
            return;
        }

        if (op == null) {
            op = new AddOperator();
        }
        
        for (Map.Entry<String,Object> e : in.entrySet()) {

            String key = e.getKey();
            Object inValue = e.getValue();

            if (out.get(key) == null) {
                out.put(key, inValue);
                continue;
            }

            Object outValue = out.get(key);

            if ((inValue instanceof Integer)
                && (outValue instanceof Integer)) {

                int sum = op.op(((Integer) outValue).intValue(),
                                ((Integer) inValue).intValue());
                                

                out.put(key, Integer.valueOf(sum));
            }

            else if ((inValue instanceof Double) 
                     && (outValue instanceof Double)) {

                double sum = op.op(((Double) outValue).doubleValue(),
                                   ((Double) inValue).doubleValue());

                out.put(key, Double.valueOf(sum));
            }

            else if ((inValue instanceof Long) 
                     && (outValue instanceof Long)) {

                long sum = op.op(((Long) outValue).longValue(),
                                 ((Long) inValue).longValue());

                out.put(key, Long.valueOf(sum));
            }

            else if ((inValue instanceof Number)
                     && (outValue instanceof Number)) {

                double sum = op.op(((Number) outValue).doubleValue(),
                                   ((Number) inValue).doubleValue());

                out.put(key, Double.valueOf(sum));
            }
            
            else if ((inValue instanceof Number)
                     || (outValue instanceof Number)) {
                
                /* one is number, other is not */
                throw new IllegalArgumentException
                    ("key="+ key
                     + ": cannot combine '" + inValue + "'" 
                     + " and '" + outValue + "'");
            }

            else {
                out.put(key, inValue);
            }
        }
    }


    /**
     * Combines a set of maps.  The semantics are 
     * <pre>
     * out += i
     * </pre>
     * for each i in in.
     *
     * @param out the result map.  If null, a new map is created
     *
     * @param in the list of maps to combine
     *
     * @return the out map.
     */
    public static Map<String,Object> addMaps(Map<String,Object> out, 
                                             List<Map<String,Object>> in) {
        if (out == null) {
            out = new HashMap<String,Object>();
        }

        if (in == null) {
            return out;
        }

        for (Map<String,Object> m : in) {
            if (m != null) {
                addMaps(out, m);
            }
        }

        return out;
    }


    /**
     * Loads a class by name.
     *
     * @param classname the name of the class to load.
     *
     * @param classtype The desired type.  Classname must be
     * assignable to this type
     *
     * @exception TypeNotPresentException if classname cannot be loaded,
     * or cannot cast to the desired type.
     */
    public static <T> Class<? extends T> loadClass(String classname, 
                                                   Class<T> classtype) {
        if (classname == null) {
            throw new NullPointerException("classname cannot be null");
        }

        if (classtype == null) {
            throw new NullPointerException("classtype cannot be null");
        }

        ClassLoader loader1 = Thread.currentThread().getContextClassLoader();
        ClassLoader loader2 = Util.class.getClassLoader();

        Class<? extends T> cl = loadClassSub(loader1, classname, classtype);
        if (cl != null) {
            return cl;
        }

        cl = loadClassSub(loader2, classname, classtype);
        if (cl != null) {
            return cl;
        }

        throw new TypeNotPresentException(classname,
                                          new ClassNotFoundException(classname));
    }


    /**
     * Loads a class by name, using a specific classloader.
     */
    @SuppressWarnings("unchecked")
    private static <T> Class<? extends T> loadClassSub(ClassLoader loader,
                                                       String classname, 
                                                       Class<T> classtype) {
        if (loader == null) {
            return null;
        }

        Class<?> cl = null;
        try {
            cl = loader.loadClass(classname);
        }
        catch (ClassNotFoundException e) {
            return null;
        }
        
        assert cl != null;
        if (classtype.isAssignableFrom(cl)) {
            return (Class<? extends T>) cl;
        }
        else {
            throw new TypeNotPresentException(classname, 
                                              new IllegalArgumentException
                                              ("classname " + classname
                                              + " is not assignable to "
                                               + classtype.getName()));
        }
    }


    /**
     * Given a class object, instantiate an instance (via no arg constructure).
     */
    public static <T> T makeInstance(Class<T> classObject) {
        try {
            return makeInstanceSub(classObject);
        }
        catch (Exception e) {
            String name = (classObject == null ? "null" : classObject.getName());
            throw new TypeNotPresentException(name, e);
        }
    }


    private static <T> T makeInstanceSub(Class<T> classObject) 
        throws InstantiationException, IllegalAccessException {

        assert classObject != null;

        if (classObject == null) {
            throw new NullPointerException("Class<T> classObject required");
        }

        return classObject.newInstance();
    }


    
    /*==================================================================
     */

    /**
     * Reads a set of properties from a java properties file.
     *
     * Returns the properties object.
     *
     * @exception RuntimeException if the properties file cannot be read.
     */
    public static Properties readProperties(File file) {
        if (file == null) {
            throw new NullPointerException("file cannot be null");
        }

        Properties props = new Properties();
        try {
            readProperties(props, file);
            return props;
        }
        catch (IOException e) {
            throw new RuntimeException("error loading properties file: " + file, 
                                       e);
        }
    }


    private static void readProperties(Properties props, File file) 
        throws IOException {

        FileInputStream in = null;

        try {
            in = new FileInputStream(file);
            props.load(in);
        }

        finally {
            if (in != null) {
                in.close();
            }
        }
    }


    /**
     * Sleeps for a randomized period of time, centered around
     * <tt>baseSleep</tt>.
     *
     * <p>
     * If <tt>baseSleep</tt> &le; 0, then this method returns immediately.
     *
     * Otherwise, the current thread is put to sleep for a period of
     * time in the range [0.5baseSleep, 1.5baseSleep].  Within this
     * range, sleep times are random with uniform distribution.
     */
    public static void randomSleep(double baseSleep) throws InterruptedException {
        final long NANO_PER_SEC = 1000000L;
        if (baseSleep <= 0.0) {
            return;
        }

        double value = (Math.random() * baseSleep)
            + (baseSleep / 2.0);

        long ms = (long) value;
        int ns = (int) (((long) (value * NANO_PER_SEC)) - (ms * NANO_PER_SEC));

        Thread.sleep(ms, ns);
    }


    /**
     * Like {@link #randomSleep}, but does not throw InterruptedException.
     *
     * An InterruptedException silently ends the sleep period.
     */
    public static void randomSleepQuiet(double baseSleep) {
        try {
            randomSleep(baseSleep);
        }
        catch (InterruptedException e) {
            /* Ok, wake up */
        }
    }

}
