/*
 * $Id: LightningBolt.java,v 1.3 2010/07/01 02:10:51 srevilak Exp $
 */

package edu.umb.bench.sibench;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.DriverManager;

/**
 * A one-off test that happens to use the sibench table.
 *
 * This test creates a very chain of anti-dependencies, done in such a
 * way that ordinary pruning will not remove any nodes from the CTG.
 * The general pattern is
 * <pre>
 * r1(a) w1(b)
 * r2(b) w2(c) c1
 * r3(c) w3(d) c2
 * r4(d) w4(e) c3
 *   ...
 *</pre>
 *
 * It's kind of a system test to see what happens when
 * we try to create an unbounded list of dependencies
 *
 * <p>
 * Main takes two or three arguments: (1) jdbc url, (2) username
 * (3) password (if needed).
 */
public final class LightningBolt {

    public static final String DRIVER_CLASS = "com.mysql.jdbc.Driver";
    public static final int NROWS = 349;

    String url = null;
    String user = null;
    String passwd = null;

    Connection conn[];
    Statement st[];


    public LightningBolt(String url, String user, String passwd) {
        this.url = url;
        this.user = user;
        this.passwd = passwd;
    }


    private void diag(String s) {
        System.out.println(s);
    }



    private void executeUpdate(int idx, String sql) throws SQLException {

        diag("conn " + idx + ": "  + sql);

        st[idx].execute(sql);
    }



    private void executeSelect(int idx, String sql) throws SQLException {

        diag("conn " + idx + ": "  + sql);

        st[idx].execute(sql);
        ResultSet rs = st[idx].getResultSet();
        while (rs.next()) {
            ;
        }
        rs.close();
    }


    
    private void executeCommit(int idx) throws SQLException {
        diag("conn " + idx + ": commit");

        try {
            conn[idx].commit();
        }
        catch (SQLException e) {
            if (e.getMessage().indexOf("error 147 during COMMIT") >= 0) {
                /* test passed */
                diag("conn " + idx + ": *** ROLLBACK: CTG path too long");
                conn[idx].rollback();
            }
            else {
                throw e;
            }
        }
    }



    private void loop() throws SQLException {

        int row = 1; 
        
        while (row <= NROWS) {
            String sql = "";

            sql = "select value from sibench where krandseq = " + row;
            executeSelect(0, sql);
            row++;

            sql = "update sibench set value = value + 1 where krandseq = " + row;
            executeUpdate(0, sql);

            if (row > 2) {
                executeCommit(1);
            }

            sql = "select value from sibench where krandseq = " + row;
            executeSelect(1, sql);
            row++;
            
            sql = "update sibench set value = value + 1 where krandseq = " + row;
            executeUpdate(1, sql);
            
            executeCommit(0);

        }

        executeCommit(0);
        executeCommit(1);

    }



    private void connect() throws SQLException {
        conn = new Connection[2];
        st = new Statement[conn.length];

        try {
            Class.forName(DRIVER_CLASS);
        }
        catch (ClassNotFoundException e) {
            throw new TypeNotPresentException(DRIVER_CLASS, e);
        }

        for (int iter = 0; iter < conn.length; iter++) {
            diag("opening connection to " + url);
            conn[iter] = DriverManager.getConnection(url, user, passwd);
            conn[iter].setAutoCommit(false);

            st[iter] = conn[iter].createStatement();
        }
    }

    

    private void disconnect() throws SQLException {

        for (int iter = 0; conn != null && iter < conn.length; iter++) {
            diag("close connection");
            conn[iter].close();
        }

    }



    public void run() throws SQLException {

        connect();
        loop();
        disconnect();
    }


    public static void main(String argv[]) {
        if (argv.length < 2 || argv.length > 3) {
            System.err.println("usage: LightningBolt URL USERNAME [PASSWD]");
            System.exit(1);
        }

        String url = argv[0];
        String user = argv[1];
        String passwd = null;

        if (argv.length == 3) {
            passwd = argv[2];
        }

        LightningBolt l = new LightningBolt(url, user, passwd);

        try {
            l.run();
        }
        catch (SQLException e) {
            e.printStackTrace();
            System.exit(1);
        }

        System.exit(0);
    }

}
