/*
 * $Id: SIBench1TransactionFactory.java,v 1.14 2010/07/25 03:17:20 srevilak Exp $
 */

package edu.umb.bench.sibench;

import edu.umb.bench.framework.BaseTransactionFactory;
import edu.umb.bench.framework.INumberGenerator;
import edu.umb.bench.framework.GeneratorUnif;
import edu.umb.bench.framework.Probability;
import edu.umb.bench.framework.ITransaction;
import edu.umb.bench.framework.IHotspot;
import edu.umb.bench.framework.RandomHotspot;
import java.util.Set;
import java.util.HashSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

/**
 * Transaction Factory for SIBench1.
 *
 * @version $Revision: 1.14 $ $Date: 2010/07/25 03:17:20 $
 */
public class SIBench1TransactionFactory extends BaseTransactionFactory<TxResult> {

    private INumberGenerator mGenerator;
    private SIBench1Transaction mTransaction;
    private int nrows;

    /* updates always use a unique index.   We use that as the
     * basis for choosing hot rows */
    private static volatile IHotspot mUpdateHotspot;

    /* To find hot spots for (possibly non-unique) secondary indexes,
     * we select values from hot rows */
    private static volatile IHotspot mSelectHotspots[];

    private int mSelectStatements;
    private int mUpdateStatements;
    private double mRandomSleep;

    static final double[] DEFAULT_SELECT_PROBABILITIES = {
        0.0, /* kseq */
        1.0, /* krandseq */
        0.0, /* k4 */
        0.0, /* k8 */
        0.0, /* k16 */
        0.0, /* k32 */
        0.0, /* k64 */
        0.0, /* k128 */
        0.0, /* k256 */
        0.0, /* k512 */
        0.0, /* k1024 */
        0.0, /* k2500 */
        0.0, /* k5k */
        0.0, /* k10k */
        0.0, /* k25k */
        0.0, /* k50k */
        0.0, /* k100k */
        0.0, /* k250k */
        0.0  /* k500k */
    };


    private double selectProbabilities[];

    /**
     * Initialization/
     */
    public void init() {
        if (DEFAULT_SELECT_PROBABILITIES.length != 
            Constants.COL_NAMES.length) {
            throw new IllegalStateException
                ("DEFAULT_SELECTPROBABILITIES.length != COL_NAME.length");
        }

        mGenerator = new GeneratorUnif();

        /* one transaction object per thread */
        mTransaction = new SIBench1Transaction();

        commonSetup(mTransaction);

        setupProbabilities();
        
        setupUpdateHotspot();

        setupSelectHotspots();

        setupSelectCount();

        setupUpdateCount();

        setupRandomSleep();
    }


    /**
     * Setup probability distribution for choosing the 
     * select statement column 
     */
    private void setupProbabilities() {

        nrows = (int) (getScale() * Constants.ROWS_PER_SCALE);
        
        String pdist = getConfig().getProperty(Constants.CONFIG_SELECT_DISTRIBUTION);

        if (pdist == null || pdist.length() == 0) {
            selectProbabilities = DEFAULT_SELECT_PROBABILITIES;
            return;
        }
        
        double tmp[] = Probability.parse(pdist);

        if (tmp.length == DEFAULT_SELECT_PROBABILITIES.length) {
            selectProbabilities = tmp;
            return;
        }

        getLogger().warn("bad "
                         + Constants.CONFIG_SELECT_DISTRIBUTION
                         + "="
                         + pdist
                         + ", substituting defaults");
        selectProbabilities = DEFAULT_SELECT_PROBABILITIES;
    }


    /**
     * Configure a select hotspot for column that might be used in a select.
     */
    private synchronized void setupSelectHotspots() {
        if (mSelectHotspots != null) {
            return;
        }
  
        IHotspot tmp[] = new IHotspot[DEFAULT_SELECT_PROBABILITIES.length];
        final Set<Integer> hotRows = 
            ((RandomHotspot) mUpdateHotspot).getHotspot();
       
        for (int iter = 0; iter < selectProbabilities.length; iter++) {

            /* if not using this column, then move on to the next */
            if (selectProbabilities[iter] == 0.0) {
                continue;
            }

            String column = Constants.COL_NAMES[iter];
            int card = Card.getMaxValue(column);

            /* if this is a unique index, use the same hotspot
             * for selects and updates */
            if (card == Card.UNIQUE_INDEX) {
                tmp[iter] = mUpdateHotspot;
                continue;
            }

            /* not a unique index.  Find values that appear in hot rows */
            int values[] = findValuesThatAppearIn(column, hotRows);
            tmp[iter] = new RandomHotspot(card, values, 
                                          mUpdateHotspot.getHotspotProbability());
        }

        mSelectHotspots = tmp;
    }


    /**
     * Configure hotspot for row updates (and selects that use
     * a unique index 
     */
    private synchronized void setupUpdateHotspot() {
        IHotspot tmp = null;
        if (mUpdateHotspot != null) {
            return;
        }

        String hotspotProbabilityStr = 
            getConfig().getProperty(Constants.CONFIG_HOTSPOT_PROBABILITY);

        String hotspotSizeStr = 
            getConfig().getProperty(Constants.CONFIG_HOTSPOT_SIZE);

        boolean writeBack = false;

        if (hotspotProbabilityStr == null || hotspotProbabilityStr == null) {
            tmp = defaultHotspot();
            writeBack = true;
        }
        else {
            double hSize = Double.parseDouble(hotspotSizeStr);
            double hpct= Double.parseDouble(hotspotProbabilityStr);
            int hSizeRows = (int) ( ((double) nrows) * hSize);

            tmp = new RandomHotspot(nrows, hSizeRows, hpct);
        }

        /* write back hotspot configuration, so that it appears in the
         * test output */
        if (writeBack) {
            getConfig().setProperty(Constants.CONFIG_HOTSPOT_PROBABILITY,
                                    Double.toString
                                    (tmp.getHotspotProbability()));
            double hSize = 
                ((double) tmp.getHotspotSize()) /
                ((double) tmp.getTableSize());

            getConfig().setProperty(Constants.CONFIG_HOTSPOT_SIZE,
                                    Double.toString(hSize));
        }

        mUpdateHotspot = tmp;
   }



    /**
     * Given a secondary indexed "column" and a set of hotRows
     * (which are krandseq values), find the "column" values
     * that appear in the hot krandseq rows.
     */
    private int[] findValuesThatAppearIn(String column, 
                                         Set<Integer> hotRows) {

        /* it would be more efficient to handle all of the
         * secondary indexes at once, but it's easier to handle
         * them one at a time.  The vast majority of our tests
         * will use one column for selects, so this should be
         * acceptable */

        getLogger().info("Determining set of hot " + column + " values");

        PreparedStatement st = null;
        ResultSet rs = null;
        Set<Integer> foundHot = new HashSet<Integer>(hotRows.size());
        Connection conn = null;

        try {
            conn = getConnection();
            st = conn.prepareStatement
                ("select " + column + " from bench where "
                 + Constants.COL_UPDATE + " = ?");

            for (Integer k : hotRows) {

                st.setInt(1, k.intValue());
                st.execute();
                rs = st.getResultSet();

                while (rs.next()) {
                    int v = rs.getInt(column);
                    foundHot.add(Integer.valueOf(v));
                }

                rs.close();
            }

            st.close();
            conn.commit();
        }
        catch (SQLException e) {
            try {
                if (conn != null) {
                    conn.rollback();
                }
            }
            catch (Exception e2) {  /* do nothing */ }

            /* kick back and unwind ... */
            throw new RuntimeException("problem determining "
                                       + column
                                       + " values for hotspot", e);
        }

        /* turn the set into an array */
        int rv[] = new int[foundHot.size()];
        int idx = 0;

        for (Integer v : foundHot) {
            rv[idx++] = v.intValue();
        }

        return rv;
    }



    private synchronized void setupSelectCount() {
       String countstr = getConfig()
           .getProperty(Constants.CONFIG_SELECT_STATEMENTS);
       int count = 1;
       boolean writeback = false;
       
       if (countstr != null) {
           count = Integer.parseInt(countstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_SELECT_STATEMENTS,
                                   Integer.toString(count));
       }
       mSelectStatements = count;
   }


    private synchronized void setupUpdateCount() {
       String countstr = getConfig()
           .getProperty(Constants.CONFIG_UPDATE_STATEMENTS);
       int count = 1;
       boolean writeback = false;
       
       if (countstr != null) {
           count = Integer.parseInt(countstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_UPDATE_STATEMENTS,
                                   Integer.toString(count));
       }
       mUpdateStatements = count;
   }


    private synchronized void setupRandomSleep() {
       String sleepstr = getConfig()
           .getProperty(Constants.CONFIG_RANDOM_SLEEP);
       double duration = 3.0;
       boolean writeback = false;
       
       if (sleepstr != null) {
           duration = Double.parseDouble(sleepstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_RANDOM_SLEEP,
                                   Double.toString(duration));
       }
       mRandomSleep = duration;
   }


    private IHotspot defaultHotspot() {
        /* 90% of rows come from 10% of table */
        int hsize =  (int) (Constants.DEFAULT_HOTSPOT_SIZE * nrows);
        double hpct = Constants.DEFAULT_HOTSPOT_PROBABILITY;
        return new RandomHotspot(nrows, hsize, hpct);
    }


    /**
     * Returns a configured transaction.
     */
    public ITransaction<TxResult> createTransaction() {
        /* Factories are per-thread, so we'll use the same
         * transaction profile, and just re-configure it each time 
         */

        int ordinal = Probability.pickElement(selectProbabilities);
        String colname = Constants.COL_NAMES[ordinal];
        int cardinality = Card.getMaxValue(colname);
        IHotspot selectHotspot = mSelectHotspots[ordinal];

        int updateKeys[] = new int[mUpdateStatements];
        int selectKeys[] = new int[mSelectStatements];

        Set<Integer> allValues = new HashSet<Integer>();

        for (int iter = 0; iter < selectKeys.length; iter++) {
            /* get select hotspot associated with select columns,
             * and generate a values from that */
            selectKeys[iter] = newValue(selectHotspot, allValues);
        }
        
        for (int iter = 0; iter < updateKeys.length; iter++) {
            updateKeys[iter] = newValue(mUpdateHotspot, allValues);
        }

        mTransaction.setParams(colname, 
                               selectKeys,
                               updateKeys,
                               mRandomSleep);

        return mTransaction;

    }

    /**
     * Use hotspot h to generate a random value, but try
     * to find a value that is not already in s.
     * This is an attempt to generate distinct sets of select and
     * update values.  
     *
     * <p>When the same column is used for select and
     * update, we will end up with values that are distinct.  
     *
     * <p>If different columns are used for select and update, the set
     * of select values will be distinct, and the set of update values
     * will be distinct.  However, the set of selects + updates might
     * intersect on a common row.
     */
    private int newValue(IHotspot h, Set<Integer> s) {
        int value = 0;

        /* make three attempts to generate a distinct value.
         * If we can't get a distinct value after three attempts, then
         * use the last value we picked. */

        for (int iter = 0; iter < 3; iter++) {
            value = h.getValue();

            Integer ival = Integer.valueOf(value);

            if (! s.contains(ival)) {
                s.add(ival);
                return value;
            }
        }

        return value;
    }

}
