/*
 * $Id: SIBenchMTransactionFactory.java,v 1.1 2010/07/12 02:00:27 srevilak Exp $
 */

package edu.umb.bench.sibench;

import edu.umb.bench.framework.BaseTransactionFactory;
import edu.umb.bench.framework.INumberGenerator;
import edu.umb.bench.framework.GeneratorUnif;
import edu.umb.bench.framework.Probability;
import edu.umb.bench.framework.ITransaction;
import edu.umb.bench.framework.IHotspot;
import edu.umb.bench.framework.RandomHotspot;

import java.util.Set;
import java.util.HashSet;
import java.sql.SQLException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

/**
 * Transaction Factory for sibench-m
 *
 * @version $Revision: 1.1 $ $Date: 2010/07/12 02:00:27 $
 */
public class SIBenchMTransactionFactory extends BaseTransactionFactory<TxResult> {

    private SIBenchMTransaction mTransaction;
    private INumberGenerator mGenerator;

    /** Number of rows in range scans (selects) */
    private int mSelectRange;

    /** number of single row update statements */
    private int mUpdateStatements;

    /** delay between statements */
    private double mRandomSleep;

    /** number of rows in table. */
    private int mRows;

    /** number of rows in hotspot */
    private int mHotRows;

    /** low value of the current hotspot */
    private double mPoint;

    /** amount that hotspots shifts (per transaction) */
    private double mShift;



    /**
     * Initialization/
     */
    public void init() {

        mGenerator = new GeneratorUnif();

        mTransaction = new SIBenchMTransaction();

        commonSetup(mTransaction);

        setupHotRange();

        resetPoint();
        
        setupSelectRange();

        setupUpdateCount();

        setupRandomSleep();

        setupHotspotShift();

    }


    /**
     * Setup probability distribution for choosing the 
     * select statement column 
     */
    private synchronized void setupHotRange() {

        mRows = (int) (getScale() * Constants.ROWS_PER_SCALE);
        
        double size = 0.001;

        String hotspotSizeStr = 
            getConfig().getProperty(Constants.CONFIG_HOTSPOT_SIZE);

        boolean writeback = false;


        if (hotspotSizeStr != null) {
            size = Double.parseDouble(hotspotSizeStr);
        }
        else {
            writeback = true;
        }
        
        if (writeback) {
            getConfig().setProperty(Constants.CONFIG_HOTSPOT_SIZE,
                                    Double.toString(size));
        }

        mHotRows = (int) (mRows * size);

    }


    private synchronized void setupSelectRange() {

       String rangestr = getConfig()
           .getProperty(Constants.CONFIG_SELECT_RANGE);

       int range = 10;
       boolean writeback = false;
       
       if (rangestr != null) {
           range = Integer.parseInt(rangestr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_SELECT_RANGE,
                                   Integer.toString(range));
       }
       mSelectRange = range;
   }


    private synchronized void setupUpdateCount() {
       String countstr = getConfig()
           .getProperty(Constants.CONFIG_UPDATE_STATEMENTS);
       int count = 1;
       boolean writeback = false;
       
       if (countstr != null) {
           count = Integer.parseInt(countstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_UPDATE_STATEMENTS,
                                   Integer.toString(count));
       }
       mUpdateStatements = count;
   }


    private synchronized void setupRandomSleep() {
       String sleepstr = getConfig()
           .getProperty(Constants.CONFIG_RANDOM_SLEEP);
       double duration = 3.0;
       boolean writeback = false;
       
       if (sleepstr != null) {
           duration = Double.parseDouble(sleepstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_RANDOM_SLEEP,
                                   Double.toString(duration));
       }
       mRandomSleep = duration;
   }


    private synchronized void setupHotspotShift() {
       String shiftstr = getConfig()
           .getProperty(Constants.CONFIG_HOTSPOT_SHIFT);
       double shift = 1.0;
       boolean writeback = false;
       
       if (shiftstr != null) {
           shift = Double.parseDouble(shiftstr);
       }
       else {
           writeback = true;
       }

       if (writeback) {
           getConfig().setProperty(Constants.CONFIG_HOTSPOT_SHIFT,
                                   Double.toString(shift));
       }

       mShift = shift;
   }



    /**
     * Returns a configured transaction.
     */
    public ITransaction<TxResult> createTransaction() {
        int point = (int) mPoint;
        
        /* determine range scan */
        int low = chooseFromHotspot();
        int high = low + mSelectRange - 1;

        /* determine which rows to update */
        int updates[] = new int[mUpdateStatements];
        for (int iter = 0; iter < updates.length; iter++) {
            updates[iter] = chooseFromHotspot();
        }

        mTransaction.setParams(low,
                               high,
                               updates,
                               mRandomSleep);

        /* advance point for the next iteration */
        mPoint += mShift;
        if (mPoint > (mRows - mHotRows)) {
            resetPoint();
        }

        return mTransaction;
    }


    private int chooseFromHotspot() {
        int rv =  (int) Math.round(mPoint) + 
            (int) Math.round(mGenerator.getNumber() * mHotRows);

        return rv;
    }


    /**
     * Initial value for point.
     */
    private void resetPoint() {
        mPoint = 1.0;
    }


    public String toString() {
        return new StringBuilder(512)
            .append("SiBench2mTransactionFactory: ")
            .append(" rows=")
            .append(mRows)
            .append(", hotRows=")
            .append(mHotRows)
            .append(", point=")
            .append(mPoint)
            .append(", shift=")
            .append(mShift)
            .append(", selectRange=")
            .append(mSelectRange)
            .append(", updateStatements=")
            .append(mUpdateStatements)
            .toString();
    }

}
