/*
 * $Id: UtilTest.java,v 1.3 2010/02/08 19:22:35 srevilak Exp $
 */

package edu.umb.bench.framework;

import org.junit.Test;
import org.junit.Before;
import org.junit.After;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.junit.Assert.assertNotNull;
import java.util.Map;
import java.util.HashMap;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Properties;

public class UtilTest {



    @Before public void setup() {

    }

    @After public void teardown() {

    }

    public static void main(String argv[]) {
        TestUtil.runClass(UtilTest.class);
    }


    /*==================================================================
     */

    /**
     * Verify that invalid classnames produce an exception.
     */
    @Test public void testLoadClassBadClass() {
        try {
            Class<?> foo = Util.loadClass("edu.umb.no.such.ClassName", 
                                          Object.class);
            fail("no illegal argument exception for invalid classname");
        }
        catch (TypeNotPresentException e) {
            assertTrue("OK", true);
        }
    }


    /**
     * Verify that invalid type specification produces an exception.
     */
    @Test public void testLoadClassWrongType() {
        try {
            Class<?> foo = Util.loadClass("java.lang.Integer",
                                          String.class);
            fail("no exception for wrong type");
        }
        catch (TypeNotPresentException e) {
            assertTrue("OK", true);
        }
    }

    
    /**
     * Basic Success Case.
     */
    @Test public void testLoadClassOK() {
        Class<? extends Map> foo = Util.loadClass("java.util.HashMap",
                                                  java.util.Map.class);
        assertNotNull("loadClass", foo);
        assertEquals("classname", "java.util.HashMap", foo.getName());
    }

    

    /**
     * Add maps, invalid parameters
     */
    @Test public void testAddMapNullOut() {
        Map<String,Object> out = null;
        Map<String,Object> in = new HashMap<String,Object>();
        try {
            Util.addMaps(out, in);
            fail("no exception for missing out map");
        }
        catch (NullPointerException e) {
            assertTrue("OK", true);
        }
    }

    
    /**
     * Add null input map to non-null output map
     */
    @Test public void testAddMapNullIn() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = null;
        Util.addMaps(out, in);
        assertEquals("out.size()", 0, out.size());
    }


    /**
     * Test copying a new key from in to out.
     */
    @Test public void testAddMapsTransferInToOut() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();

        in.put("key", "value");
        Util.addMaps(out, in);

        assertEquals("out.size()", 1, out.size());
        assertEquals("out[key]", "value", out.get("key"));
    }


    /**
     * Test summing two integers
     */
    @Test public void testAddMapsAddInteger() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();
        out.put("key", Integer.valueOf(2));
        in.put("key", Integer.valueOf(3));
        
        Util.addMaps(out, in);

        Number result = (Number) out.get("key");
        assertNotNull("out[key]", result);
        assertEquals("result", 5, result.intValue());
    }


    /**
     * Test summing two longs.
     */
    @Test public void testAddMapsAddLong() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();
        out.put("key", Long.valueOf(5));
        in.put("key", Long.valueOf(7));
        
        Util.addMaps(out, in);

        Number result = (Number) out.get("key");
        assertNotNull("out[key]", result);
        assertEquals("result", 12L, result.longValue());
    }


    /**
     * test summing two doubles.
     */
    @Test public void testAddMapsAddDouble() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();
        out.put("key", Double.valueOf(2.0));
        in.put("key", Double.valueOf(4.0));
        
        Util.addMaps(out, in);

        Number result = (Number) out.get("key");
        assertNotNull("out[key]", result);
        assertEquals("result", 6.0, result.doubleValue(), 0.0001);
    }


    /**
     * Test summing two numbers, of different types
     */
    @Test public void testAddMapsAddNumbers() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();
        out.put("key", Integer.valueOf(11));
        in.put("key", Long.valueOf(13));

        Util.addMaps(out, in);

        Number result = (Number) out.get("key");
        assertNotNull("out[key]", result);
        assertEquals("result", 24, result.intValue());

    }


    /**
     * Test summing of number and string.  (Raises exception)
     */
    @Test public void addNumberNotNumber() {
        Map<String,Object> out = new HashMap<String,Object>();
        Map<String,Object> in = new HashMap<String,Object>();
        out.put("key", "foo");
        in.put("key", Integer.valueOf(13));
        
        try {
            Util.addMaps(out, in);
            fail("no exception for incomparable types");
        }
        catch (IllegalArgumentException e) {
            assertTrue("OK", true);
        }
    }



    /*==================================================================
     */
    
    /**
     * Verify behavior when we try to read a properties file that
     * doesn't exist
     */
    @Test public void testReadPropertiesFile() {
        File f = new File("/this/file/does/not/exist");
        assertTrue("f.exists()", ! f.exists());

        try {
            Util.readProperties(f);
            fail("expected error reading: " + f);
        }
        catch (RuntimeException e) {
            assertTrue("OK", true);
        }
    }


    /**
     * Basic success case: write a few properties to a file
     * and then read them back.
     */
    @Test public void testBasicRead() throws IOException {
        String properties = "a=1\nb=2\n";
        File tmpProperties = File.createTempFile("UtilTest-", ".properties");
        tmpProperties.deleteOnExit();
        
        FileWriter writer = new FileWriter(tmpProperties);
        writer.write(properties);
        writer.flush();
        writer.close();

        Properties p = Util.readProperties(tmpProperties);
        assertNotNull("readProperties", p);
        assertEquals("a", "1", p.getProperty("a"));
        assertEquals("b", "2", p.getProperty("b"));
        assertEquals("size", 2, p.size());
    }


    @Test public void testMakeInstance() {
        HashMap m = Util.makeInstance(HashMap.class);
        assertNotNull("makeInstance(HashMap)", m);
    }
}
