/*
 * $Id: lock0ctg.h,v 1.24 2010/07/10 19:47:51 srevilak Exp $
 *
 * SREVILAK: Lock system interface to Cycle Testing Graph (CTG)
 */

#ifndef lock0ctg_h
#define lock0ctg_h

#include "mem0mem.h"
#include "ut0byte.h"
#include "trx0types.h"
#include "lock0types.h"

/**
 * Resets all CTG-related fields in a transaction.
 */
void trx_ctg_reset(trx_t * trx, ibool create_edgelist);

/*
 * Creates an (empty) ctg edgelist.
 *
 * There is not a corresponding free function.  The caller should
 * release memory with mem_heap_empty(), or mem_heap_free().  In practice,
 * this memory is allocated from trx->lock_heap, so that CTG structures live
 * as long as lock structures.
 *
 * @param heap heap for memory allocation
 *
 * @return new edgelist (array of dulint)
 */
ctg_edgelist_t * ctg_edgelist_create(mem_heap_t * heap);


/*
 * Returns the number of edges in the edgelist
 */
ulint ctg_edgelist_size(const ctg_edgelist_t * edgelist);


/*
 * Returns the allocated elements.  Only useful for testing.
 */
ulint ctg_edgelist_alloc_size(const ctg_edgelist_t * edgelist);


/*
 * Returns the n-th edge in the edgelist
 */
dulint ctg_edgelist_get(const ctg_edgelist_t * edgelist, ulint n);


/*
 * LOW LEVEL FUNCTION: consider ctg_add_edge instead.
 *
 * Adds an edge to the edgelist.  If the edgelist already
 * contains trxid, then no change is made.
 *
 * "heap" should be the memory heap that allocated edgelist.
 * Heap is necessary in cases where more memory has to be allocated
 *
 * Return: TRUE if edge added, FALSE if edge was already present
 */
ibool ctg_edgelist_add(ctg_edgelist_t ** edgelist, mem_heap_t * heap, dulint trxid);


/*
 * Tests whether the edgelist contains an edge to trxid.
 */
ibool ctg_edgelist_contains(const ctg_edgelist_t * edgelist, dulint trxid);


/*
 * Removes trxid from the edgelist.  Returns TRUE
 * if the edgelist was modified.  False otherwise.
 */
ibool ctg_edgelist_remove(ctg_edgelist_t * edgelist, dulint trxid);


/*
 * "zeros" all edges in the edgelist.  Does not change allocated
 * size, or change the trxid.
 */
void ctg_edgelist_clear(ctg_edgelist_t * edgelist);


/*==================================================================
 * CTG Proper
 */

/**
 * Creates a cycle testing graph.
 *
 * n_buckets - number of buckets in the hashtable of graph nodes
 *
 * the_kernel_mutex - pointer to mutex that protects lock system.
 */
ctg_t * ctg_create(ulint n_buckets, mutex_t * the_kernel_mutex);


/**
 * Sets the Dangerous structure testing flag.  The CTG's default
 * behavior is to do cycle testing.  Setting this flag to true
 * causes the CTG to do dangerous structure testing instead 
 * of cycle testing.
 *
 * In general, we assume that this flag will be set once, at startup,
 * based upon the value of srv_dangerous_structure_testing.
 * We expose a setter here for the benefit of unit tests.
 *
 * Dangerous structure tesing changes three things
 *
 *   - transactions keep in-edges as well as out edges
 *
 *   - serializability is determined by a dangerous structure
 *     test, as opposed to a cycle test
 *
 *   - pruning requirements are changed.  (We don't consider whether
 *     Ti has in edges.
 */
void ctg_set_dangerous_structure_testing(ibool on_off);

/**
 * Returns a boolean, which indicates whether or not the 
 * CTG is doing dangerous structure testing.
 */
ibool ctg_get_dangerous_structure_testing(void);


/**
 * [SREVILAK 2009-11-18]
 * Sets the CTG's debugging flag.
 *
 * Typical usage
 *
 *    static char my_debug_flag = TRUE;
 *    ctg_set_debug_flag(ctg, &my_debug_flag);
 *
 * Note: debug flag is a pointer to a single character, and not a
 * string.
 *
 * Passing NULL as debug_flag restores the default debugging
 * flag (i.e., no debugging).
 */
void ctg_set_debug_flag(ctg_t * ctg, const char * debug_flag);


/**
 * CTG Free
 *
 * Destroys the cycle testing graph, and zeros the callers
 * pointer to the cycle testing graph.
 *
 * Note: we will zero out any hashtable pointers to trx_t's in the
 * graph, but we do not deallocate them.
 */
void ctg_free(ctg_t ** ctg);


/*
 * Adds a transaction to the cycle testing graph.
 *
 * ctg: the ctg
 *
 * trx: the transaction to add
 *
 * Returns: true if added, false if trx was already in the ctg.
 */
ibool ctg_add(ctg_t * ctg, trx_t * trx);

/*
 * Tells the CTG to adjust it's internal data structures, after
 * trx has been assigned a commit timestamp.
 */
void ctg_reposition_commit_ts(ctg_t * ctg, trx_t * trx);


/*
 * Given transaction id, find transaction in the cycle testing
 * graph.
 */
trx_t * ctg_get(const ctg_t * ctg, dulint trxid);

/*
 * Adds a dependency edge from source to sink.  Returns
 * true if the edge was added, false if the edge was already present.
 * Both source an sink must exist in the graph.
 *
 * ctg: the ctg
 *
 * source: source node of edge
 *
 * sink: sink node of edge
 *
 * Return: true if edge added, false if edge not added
 */
ibool ctg_add_edge(ctg_t * ctg, trx_t * source, trx_t * sink);


/*
 * Removes trx from the cycle testing graph.
 *
 * ctg: the cycle testing graph
 *
 * trx: the transaction to remove
 *
 * Return: TRUE if trx removed, FALSE otherwise
 */
ibool ctg_remove(ctg_t * ctg, trx_t * trx);


/**
 * Returns the number of graph nodes in the ctg.
 */
ulint ctg_size(const ctg_t * ctg);


/**
 * Performs a cycle (or DS test), based on the server configuration
 * variable "dangerous_structure_testing".
 *
 * Returns DB_SUCCESS (no cycle), DB_DEADLOCK (cycle or ds found), or
 * DB_CTG_TOO_LONG (too many nodes in CTG, or CTG path too long).
 */
ulint ctg_cycle_or_ds_test(ctg_t * ctg, trx_t * trx);


/*
 * Performs a cycle test, starting at trx.  NOTE: this function exists
 * for unit tests.  Production code should use ctg_cycle_or_ds_test().
 *
 * trx: node to start cycle testing
 *
 * Returns DB_SUCCESS (no cycle), DB_DEADLOCK (cycle), or
 * DB_CTG_TOO_LONG.
 */
ulint ctg_cycle_test(ctg_t * ctg, trx_t * trx);


/**
 * Given committing transaction "trx", determine
 * whether trx creates a dangerous structure.
 *
 * "trx" creates a dangerous structure if "trx" is a pivot
 * transaction, and T_out (in the dangerous structure) commits first.
 *
 * NOTE: this function exists for unit tests.  Production code should
 * use ctg_cycle_or_ds_test().
 *
 * Returns DB_SUCCESS (no dangerous structure), DB_DEADLOCK (dangerous
 * structure), DB_CTG_TOO_LONG (too many nodes in CTG)
 */
ulint ctg_ds_test(ctg_t * ctg, trx_t * trx);


/*
 * Registers an observer.  The observer is notified when
 * transactions are pruned from the ctg.
 *
 * ctg: the ctg which which observer should be registered.
 *
 * prune_observer: function to be called when trx is pruned.
 * trx will be passed to this function.
 *
 * RETURN: true if observer registered, FALSE otherwise (i.e., too
 * many observers, or observer already registered).
 */
typedef void (* ctg_observer_t)(trx_t * trx);
ibool ctg_register_observer(ctg_t * ctg, ctg_observer_t prune_observer);


/*
 * Clears the list of registered observers.  
 *
 * ctg: the ctg whose list of observers should be cleared.
 *
 * RETURN: number of observers de-registered.
 */
ulint ctg_clear_observers(ctg_t * ctg);


/*
 * Prunes a cycle testing graph.  Returns the 
 * number of nodes pruned.
 *
 * ctg: the ctg to prune
 *
 * oldest_active_start: start timestamp (i.e. trx->id) of
 * the oldest active transaction.
 */
ulint ctg_prune(ctg_t * ctg, dulint oldest_active_start);


/*
 * Given a transaction, dump the list of edges to file
 */
void ctg_edgelist_dump(const trx_t * trx, FILE * file);


/*
 * Dump the entire CTG to file
 */
void ctg_dump(const ctg_t * ctg, FILE * file);


/*
 * Dumps the CTG's cycle length histogram to FILE.
 */
void ctg_dump_histogram(const ctg_t * ctg, FILE * file);


/*
 * Returns the number of times that the CTG has been
 * asked to prune itself.
 */
ulint ctg_prune_count(const ctg_t * ctg);


/**
 * Returns the total number of CTG nodes examined by prune calls.
 */
ulint ctg_prune_nodes_examined(const ctg_t * ctg);


/**
 * Returns the number of times that ctg_cycle_or_ds_test was called.
 */
ulint ctg_tests_done(const ctg_t * ctg);


/**
 * Returns sum(ctg_size()) for all calls to ctg_cycle_or_ds_test.
 *
 * Knowing the sum of ctg sizes and the number of tests allows
 * one to compute the average ctg size */
ulint ctg_tests_sum_size(const ctg_t * ctg);

/**
 * Returns the total number of edges traversed by cycle testing
 */
ulint ctg_sum_edges_traversed(const ctg_t * ctg);


/*
 * Returns the number of cycles (or dangerous structures) found
 * by calling ctg_cycle_or_ds_test
 */
ulint ctg_cycles_found(const ctg_t * ctg);


/*
 * Length of longest path traversed during 
 * cycle testing
 */
ulint ctg_max_ctg_path(const ctg_t * ctg);


/*
 * Returns the sum of longest paths traversed during cycle testing.
 * Thus ctg_sum_ctg_path()/ctg_tests_done() gives you the average
 * length of the longest path.
 *
 * Returns zero if we are doing dangerous structure testing.
 */
ulint ctg_sum_ctg_path(const ctg_t * ctg);


/*
 * Returns the length of the longest cycle found.
 *
 * Returns zero if we are using dangerous structure testing.
 */
ulint ctg_max_cycle_length(const ctg_t * ctg);


/*
 * Returns the total length of all cycles found.
 *
 * Thus ctg_sum_cycle_length()/ctg_tests_done() gives the 
 * average cycle length.
 *
 * Returns zero if we are doing dangerous structure testing.
 */
ulint ctg_sum_cycle_length(const ctg_t * ctg);

/*
 * Returns the number of times that ctg_add_edge
 * added a new edge to the ctg
 */
ulint ctg_add_edge_count(const ctg_t * ctg);

/*
 * Returns the sum of all trx->bolt_lengths found by
 * ctg_cycle_test
 */
ulint ctg_sum_bolt_length(const ctg_t * ctg);


/*
 * Returns the number of transactions where
 * ctg_cycle_test found trx->bolt_length > 0
 */
ulint ctg_trx_with_bolts(const ctg_t * ctg);


/*
 * Returns the maximum bolt length found by ctg_cycle_test
 */
ulint ctg_max_bolt_length(const ctg_t * ctg);



#endif
