/*
 * $Id: lock0hash.h,v 1.10 2010/05/12 02:55:09 srevilak Exp $
 *
 *
 * Replacement hash table for lock0lock.
 *
 * In a nutshell, this is a version of hash0hash where the hash chains
 * are doubly-linked lists, instead of singly linked lists.  This permits
 * bidirectional iteration over a hash chain.
 *
 * Given a known struct, we can remove it from the hashtable without
 * having to search a hash chain for it.
 */

#ifndef srevilak_lock_0_hash
#define srevilak_lock_0_hash

#include "ut0lst.h"
#include "univ.i"
#include "mem0mem.h"
#include "sync0sync.h"
#include "lock0types.h"

/**
 * lock0lock.c doesn't use hashtable mutexes.  Probably
 * a no-op here
 */
#ifdef UNIV_SYNC_DEBUG
# define LOCK_HASH_ASSERT_OWNED(TABLE, FOLD) \
ut_ad(!(TABLE)->mutexes || mutex_own(lock_hash_get_mutex(TABLE, FOLD)))
#else
# define LOCK_HASH_ASSERT_OWNED(TABLE, FOLD)
#endif


/**
 * Given structure, return the next structure in the 
 * hash chain.
 */
#define LOCK_HASH_GET_NEXT(DATA) UT_LIST_GET_NEXT(hash_chain, DATA)


/**
 * [SREVILAK 2009-11-01]
 * Given structure, return the previous structure in the 
 * hash chain.
 */
#define LOCK_HASH_GET_PREV(DATA) UT_LIST_GET_PREV(hash_chain, DATA)

/**
 * Given space and page number, get the first chained record in the
 * hash cell.
 */
#define LOCK_HASH_GET_FIRST(TABLE, HASH_VAL)                    \
  UT_LIST_GET_FIRST(*(lock_hash_get_nth_cell(TABLE, HASH_VAL)))


/**
 * [SREVILAK 2009-11-01]
 * Given hash cell number, get the last chained record in the hash cell.
 */
#define LOCK_HASH_GET_LAST(TABLE, HASH_VAL)                    \
  UT_LIST_GET_LAST(*(lock_hash_get_nth_cell(TABLE, HASH_VAL)))



/**
 * Given lock hash chain (i.e., head of a UT list)
 * return the first lock
 */
#define LOCK_HASH_GET_FIRST_FROM_CHAIN(HASH_CHAIN)\
  UT_LIST_GET_FIRST(*(HASH_CHAIN))

/**
 * Given lock hash chain (i.e., head of a UT list)
 * return the last lock
 */
#define LOCK_HASH_GET_LAST_FROM_CHAIN(HASH_CHAIN)\
  UT_LIST_GET_LAST(*(HASH_CHAIN))



/**
 * Inserts a lock into the hash table, at the right end of the hash chain
 *
 */
#define LOCK_HASH_INSERT(SPACE, PAGE_NO, DATA)                          \
  do {                                                                  \
    lock_hash_cell_t*	cell3333;                                       \
                                                                        \
    (DATA)->hash_chain.prev = NULL;                                     \
    (DATA)->hash_chain.next = NULL;                                     \
                                                                        \
    cell3333 = lock_hash_get_hash_cell(SPACE, PAGE_NO);                 \
                                                                        \
    UT_LIST_ADD_LAST(hash_chain, *cell3333, DATA);                      \
  } while (0)


/**
 * Inserts a lock into the hash table, at the left end of the hash chain
 *
 */
#define LOCK_HASH_INSERT_FIRST(SPACE, PAGE_NO, DATA)                    \
  do {                                                                  \
    lock_hash_cell_t*	cell3333;                                       \
                                                                        \
    (DATA)->hash_chain.prev = NULL;                                     \
    (DATA)->hash_chain.next = NULL;                                     \
                                                                        \
    cell3333 = lock_hash_get_hash_cell(SPACE, PAGE_NO);                 \
                                                                        \
    UT_LIST_ADD_FIRST(hash_chain, *cell3333, DATA);                     \
  } while (0)



/** 
 * Inserts LOCK1 immediately before LOCK2
 */
#define LOCK_HASH_INSERT_BEFORE(BASE, LOCK1, LOCK2)             \
  do {                                                          \
    if ((LOCK2)->hash_chain.prev == NULL) {                     \
      UT_LIST_ADD_FIRST(hash_chain, BASE, LOCK1);               \
    }                                                           \
    else {                                                      \
      lock_t * lock4444 = UT_LIST_GET_PREV(hash_chain, LOCK2);  \
      UT_LIST_INSERT_AFTER(hash_chain, BASE, lock4444, LOCK1);  \
    }                                                           \
  } while(0)


/**
 * Inserts LOCK2 immediately after LOCK1
 */
#define LOCK_HASH_INSERT_AFTER(BASE, LOCK1, LOCK2)      \
  UT_LIST_INSERT_AFTER(hash_chain, BASE, LOCK1, LOCK2)


/**
 * Removes a lock from the hash table
 *
 */
#define LOCK_HASH_DELETE(SPACE, PAGE_NO, DATA)                          \
  do {                                                                  \
    lock_hash_cell_t*	cell3333;                                       \
                                                                        \
    cell3333 = lock_hash_get_hash_cell(SPACE, PAGE_NO);                 \
                                                                        \
    UT_LIST_REMOVE(hash_chain, *cell3333, DATA);                        \
    DATA->hash_chain.prev = NULL;                                       \
    DATA->hash_chain.next = NULL;                                       \
  } while (0)



/**
 * creates a hash table with n cells
 *
 * Note: here is no lock_hash_free().  lock0lock doesn't do that.
 */
lock_hash_table_t * lock_hash_create(ulint n);


/**
 * Gets hash table header cell n (which is a ut0lst doubly-linked list
 * header).
 */
lock_hash_cell_t * 
lock_hash_get_nth_cell(lock_hash_table_t *h, ulint n);

/**
 * Return the number of hash cells in the lock table.
 */
ulint
lock_hash_get_n_cells(lock_hash_table_t * h);

/**
 * 
 * Given hashcode "fold", determine table cell
 */
ulint
lock_hash_calc_hash(ulint		fold,	/* in: folded value */
                    lock_hash_table_t*	table);	/* in: hash table */


/**
 * [SREVILAK 2009-11-27]. Given space and page number, 
 * return the left-to-right linked list for the lock hash
 * cell chain
 */
lock_hash_cell_t *
lock_hash_get_hash_cell(ulint space, ulint page_no);


/* The hash table structure */
struct lock_hash_table_struct {
  /* number of cells in the hash table */
  ulint n_cells;

  /* table headers -- array of doubly linked lists */
  lock_hash_cell_t * array;	

  /* if mutexes != NULL, then the number of
     mutexes, must be a power of 2 */
  ulint n_mutexes;

  /* NULL, or an array of mutexes used to
     protect segments of the hash table */
  mutex_t * mutexes;

  /* if this is non-NULL, hash chain nodes for
     external chaining can be allocated from these
     memory heaps; there are then n_mutexes many of
     these heaps */
  mem_heap_t ** heaps;
  
  mem_heap_t * heap;

  ulint magic_n;
};

#define LOCK_HASH_TABLE_MAGIC_N	(1 + 0xDEADF00)


#endif
