/*
 * $Id: lock0ctg.c,v 1.54 2010/07/10 19:47:51 srevilak Exp $
 *
 * SREVILAK: Cycle Testing Graph (CTG) portion of lock system
 */

#include "lock0ctg.h"
#include "hash0hash.h"
#include "trx0trx.h"
#include "dyn0dyn.h"

/**
 * If UMB_PRINT_CYCLES is defined, we print the length of each
 * cycle found to stderr.  This is a cheap and easy way to 
 * build a histogram of cycle lengths
 */
/* #define UMB_PRINT_CYCLES */

/*
 * Index 0 contains size information.  The .high field
 * gives the number of dulints allocated, which includes
 * elements that hold metadata.  The .low field gives the number
 * of elements stored.
 */
#define EDGELIST_SIZE_INDEX 0

/*
 * Edges start at element 1
 */
#define EDGELIST_TRX_LIST 1
#define EDGELIST_N_HEADERS EDGELIST_TRX_LIST

#define CTG_MAGIG_N 0x11115862

/**
 * Maximum length of a path that travels backwards in time.
 * 
 * Also the maximum depth to go when pruning
 */
#define CTG_MAX_RECURSION 200

/*
 * Start with 7 edges.
 * (7 * 8) + (1 * 8) = 64 bytes, which should align nicely
 * From earlier experiments, we've found that transactions typically
 * have only a small number of edges
 */
#define EDGELIST_DEFAULT_SIZE 7
#define EDGELIST_DEFAULT_ALLOC_SIZE (EDGELIST_DEFAULT_SIZE + EDGELIST_N_HEADERS)

#define EDGELIST_ALLOC_SIZE(edgelist)           \
  ((edgelist)[EDGELIST_SIZE_INDEX].high)

#define EDGELIST_USED(edgelist)                 \
  ((edgelist)[EDGELIST_SIZE_INDEX].low)

#define EDGELIST_IDX(n) (n) + EDGELIST_N_HEADERS

#define CHECK_SIZE(edgelist)                                            \
  ut_ad(EDGELIST_USED(edgelist)                                         \
        + EDGELIST_N_HEADERS <= EDGELIST_ALLOC_SIZE(edgelist))

#define CTG_MAX_OBSERVERS 8

/**
 * For storing a histogram of cycle lengths.  This is the
 * number of array elements in the historgram.  The first
 * array element contains the number of two cycles.
 */
#define CTG_HISTOGRAM_SIZE 64
#define CTG_HISTOGRAM_OFFSET 2

/**
 * Conversion of histogram index to cycle length
 */
#define CTG_INDEX_TO_LENGTH(x) ((x) + CTG_HISTOGRAM_OFFSET)


/**
 * [SREVILAK 2009-11-18] Default debugging flag.
 */
static char CTG_DEFAULT_DEBUG = FALSE;


/**
 * [SREVILAK 2010-01-09].  Are we doing dangerous structure testing.
 */
static char dangerous_structure_testing = FALSE;

/*
 * Cycle testing graph.
 *
 * This is a very standard graph implementation.  Nodes are
 * trx_t's, which are stored in a hashtable (hashed by trx_t.id).
 *
 * Each trx_t has (1) a ctg_edgelist_t, with a list of out-edges
 * and (2) a count of the number of in edges.  The out edges
 * are dulint trx_t.id's, not pointers to trx_t.
 *
 * Each time we do a cycle test, we increment ctg_visited, and use
 * that value as the `visited' marker.  This saves us the trouble
 * of "unmarking" transactions before each cycle test.
 *
 * Strictly speaking, the ctg is a singleton.  It lives inside of
 * the lock_sys_t.
 */
struct ctg_struct {
  ulint magic_n;           /* magic number */
  hash_table_t * nodes;      /* hashtable of graph nodes */
  mutex_t * kernel_mutex;  /* pointer to kernel mutex.  For assertions */
  ulint ctg_visited;       /* visit counter.  Used for cycle testing */
  ulint ctg_size;          /* number of graph nodes */
  ctg_observer_t prune_observers[CTG_MAX_OBSERVERS];
                           /* Observer: who to tell when a trx is
                            * pruned.  Shouldn't need more than
                            * CTG_MAX_OBSERVERS. */
  ulint num_observers;     /* number of registered observers */
  UT_LIST_BASE_NODE_T(trx_t) commit_ts_order;
                           /* list of transactions, ordered by commit ts.
                            * Prune in this order, stopping at oldest active
                            * start */
  const char * debug_flag; /* pointer to global debugging variable,
                            * like srv_debug_lock_manager.  (Note:
                            * this is a pointer to a single char, and NOT
                            * a string */
  ulint prune_count;       /* Number of times prune called */
  ulint prune_nodes_examined; /* total number of nodes examined by prune */
  ulint tests_done;        /* total number of cycle (or ds) tests */
  ulint tests_sum_size;    /* sum(ctg_size) for all cycle (or ds) tests */
  ulint cycles_found;      /* number of cycles (or dangerous structures) found */
  ulint sum_ctg_path;      /* sum of longest paths over all cycle tests */
  ulint max_ctg_path;      /* longest path examined */
  ulint sum_cycle_length;  /* sum of all cycle lengths */
  ulint max_cycle_length;  /* length of longest cycle */
  ulint add_edge_count;    /* number of times add_edge adds a new edge */
  ulint sum_edges_traversed; /* number of edges traversed during cycle tests */
  dulint prev_oldest_active_start; /* oldest active start from previous pruning */
  ulint sum_bolt_length;   /* sum of (trx->bolt_length) */
  ulint trx_with_bolts;    /* number of trx having bolt_length > 0 */
  ulint max_bolt_length;   /* maximum bolt length */
  ulint histogram[CTG_HISTOGRAM_SIZE]; /* histogram of cycle lengths */
};


typedef struct cycle_test_notes {
  ulint longest_path;   /* length of longest path examined */
  ulint cycle_length;   /* if cycle, length of cycle */
  ulint edges_traversed; /* number of edges traversed */
} cycle_test_notes_t;

/**
 * [SREVILAK 2010-01-10]
 * Bookkeeping information for dangerous structure testing.
 *
 * The members correspond to [T_in -> T_pivot -> T_out]
 */
typedef struct {
  const trx_t * in;
  const trx_t * out;
  const trx_t * pivot;
} ds_bookkeeping_t;


void trx_ctg_reset(trx_t * trx, ibool create_edgelist) {
  ut_ad(trx);
  trx->ctg_hash = NULL;
  trx->out_edges = NULL;
  trx->in_edges = NULL;
  trx->in_edge_count = 0;
  trx->ctg_visited = 0;
  trx->ctg_eligible = FALSE;
  trx->has_fcw = FALSE;
  trx->ctg_prune_starting_set = FALSE;
  trx->bolt_length = 0;

  if (create_edgelist) {

    ut_ad(trx->lock_heap);
    ut_ad(UT_LIST_GET_LEN(trx->trx_locks) == 0);

    trx->out_edges = ctg_edgelist_create(trx->lock_heap);

    if (UNIV_UNLIKELY(dangerous_structure_testing)) {
      trx->in_edges = ctg_edgelist_create(trx->lock_heap);
    }

  }
}



/*==================================================================
 * Functions for managing sets of edges
 */

/*
 * Creates an (empty) ctg edgelist.
 *
 * There is not a corresponding free function.  The caller should
 * release memory with mem_heap_empty(), or mem_heap_free().  In practice,
 * this memory is allocated from trx->lock_heap, so that CTG structures live
 * as long as lock structures.
 *
 * @param heap heap for memory allocation
 *
 * @param trxid the transaction id associated with
 * the list of edges
 *
 * @return new edgelist (array of dulint)
 */
ctg_edgelist_t * ctg_edgelist_create(mem_heap_t * heap) {
  ut_ad(heap);

  ctg_edgelist_t * list = mem_heap_alloc(heap,
                                         sizeof(dulint) *
                                         EDGELIST_DEFAULT_ALLOC_SIZE);
                                         
  if (list == NULL) {
    return list;
  }
  
  memset(list, 0, sizeof(dulint) * EDGELIST_DEFAULT_ALLOC_SIZE);

  EDGELIST_ALLOC_SIZE(list) = EDGELIST_DEFAULT_ALLOC_SIZE;
  EDGELIST_USED(list) = 0;
  
  return list;
}



/*
 * Returns the number of edges in the edgelist
 */
ulint ctg_edgelist_size(const ctg_edgelist_t * edgelist) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);
  return EDGELIST_USED(edgelist);
}


/*
 * Returns the allocated elements.  Only useful for testing.
 */
ulint ctg_edgelist_alloc_size(const ctg_edgelist_t * edgelist) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);
  return EDGELIST_ALLOC_SIZE(edgelist);
}


/*
 * Returns the n-th edge in the edgelist
 */
dulint ctg_edgelist_get(const ctg_edgelist_t * edgelist, ulint n) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);
  ut_ad(n < EDGELIST_USED(edgelist));
  return edgelist[EDGELIST_IDX(n)];
}


/**
 * Attempts to locate trxid in the txlist.
 *
 * returns TRUE if found, FALSE otherwise.
 *
 */
static ibool ctg_edgelist_find(const ctg_edgelist_t * edgelist, 
                               dulint trxid, 
                               ulint * out) /* out: virtual index 
                                             * or ULINT_UNDEFINED */
{
  if (out) {
    *out = ULINT_UNDEFINED;
  }

  ulint idx = 0;

  for (; idx < EDGELIST_USED(edgelist); idx++) {

    if (UT_DULINT_EQ(trxid, edgelist[EDGELIST_IDX(idx)])) {
      if (out) {
        *out = idx;
      }
      return TRUE;
    }
  }

  return FALSE;
}


/*
 * Adds an edge to the edgelist.  If the edgelist already
 * contains trxid, then no change is made.
 *
 * Returns true if the edgelist was modified, false otherwise.
 */
ibool ctg_edgelist_add(ctg_edgelist_t ** edgelist, mem_heap_t * heap, dulint trxid) {
  ut_ad(edgelist);
  ut_ad(*edgelist);
  ut_ad(heap);
  CHECK_SIZE(*edgelist);

  /* already in edgelist ? */
  ulint found = 0;
  if (ctg_edgelist_find(*edgelist, trxid, &found)) {
    return FALSE;
  }

  /* got room ? */
  if (EDGELIST_ALLOC_SIZE(*edgelist) > 
      EDGELIST_USED(*edgelist) + EDGELIST_N_HEADERS) {

    (*edgelist)[EDGELIST_IDX(EDGELIST_USED(*edgelist))] = trxid;
    EDGELIST_USED(*edgelist)++;

    return TRUE;

  }

  /* must make bigger ... */
  ulint nbytes = sizeof(dulint) * 2 * EDGELIST_ALLOC_SIZE(*edgelist);
  ctg_edgelist_t * newlist = mem_heap_alloc(heap, nbytes);

  memset(newlist, 0, nbytes);
  memcpy(newlist, *edgelist, sizeof(dulint) * EDGELIST_ALLOC_SIZE(*edgelist)); 

  EDGELIST_ALLOC_SIZE(newlist) = 2 * EDGELIST_ALLOC_SIZE(*edgelist);

  newlist[EDGELIST_IDX(EDGELIST_USED(newlist))] = trxid;
  EDGELIST_USED(newlist)++;
  
  *edgelist = newlist;

  return TRUE;
}


/*
 * Tests whether the edgelist contains a given edge.
 */
ibool ctg_edgelist_contains(const ctg_edgelist_t * edgelist, dulint trxid) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);
  return ctg_edgelist_find(edgelist, trxid, NULL);
}


/*
 * Removes trxid from the edgelist.  Returns TRUE
 * if the edgelist was modified.  False otherwise.
 */
ibool ctg_edgelist_remove(ctg_edgelist_t * edgelist, dulint trxid) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);

  ulint found = 0;
  if (! ctg_edgelist_find(edgelist, trxid, &found)) {
    return FALSE;
  }
  
  /* shuffle left */
  ulint idx = found;
  for (; idx < EDGELIST_USED(edgelist) - 1; idx++) {
    edgelist[EDGELIST_IDX(idx)] = edgelist[EDGELIST_IDX(idx + 1)];
  }
  
  /* clobber the end */
  edgelist[EDGELIST_IDX(EDGELIST_USED(edgelist) - 1)] = ut_dulint_zero;
  EDGELIST_USED(edgelist)--;

  return TRUE;
}


void ctg_edgelist_clear(ctg_edgelist_t * edgelist) {
  ut_ad(edgelist);
  CHECK_SIZE(edgelist);

  ulint idx;
  for (idx = 0; idx < EDGELIST_USED(edgelist); idx++) {
    edgelist[EDGELIST_IDX(idx)] = ut_dulint_zero;
  }
  EDGELIST_USED(edgelist) = 0;
}




/*==================================================================
 * CTG proper
 */

#ifdef UNIV_DEBUG
#define CHECK_CTG(x) do {                                               \
    ut_ad((x) != NULL && (x)->magic_n == CTG_MAGIG_N);                  \
    ut_ad((x)->debug_flag != NULL);                                     \
    ut_ad((x)->kernel_mutex == NULL || mutex_own((x)->kernel_mutex));   \
  } while(0);
#else
#define CHECK_CTG(x)
#endif

#define CHECK_TRX(x) ut_ad((x) != NULL && (x)->magic_n == TRX_MAGIC_N)


UNIV_INLINE 
void ctg_record_cycle_length(ctg_t * ctg, ulint len) {
  ulint idx;

  CHECK_CTG(ctg);

  ut_ad(len >= 2);

  if (len < 2) {
    return;
  }

  idx = len - CTG_HISTOGRAM_OFFSET;

  if (idx >= CTG_HISTOGRAM_SIZE) {
    idx = CTG_HISTOGRAM_SIZE - 1;
  }

  ctg->histogram[idx]++;
}


/**
 * Fold a transaction id.  These are effectively timestamps.  
 * Since start and commit timestamps come from the same counter,
 * we should get a pretty good distribution by dividing the 
 * transaction id by two.
 */
UNIV_INLINE
ulint ctg_fold_id(dulint trxid) {
  return (ut_dulint_get_low(trxid) >> 1);
}

UNIV_INLINE
ulint ctg_fold(trx_t * trx) {
  ut_ad(trx);
  return ctg_fold_id(trx_get_trx_id(trx));
}




ctg_t * ctg_create(ulint n_buckets, mutex_t * the_kernel_mutex) {
  ut_ad(n_buckets > 0);

  ctg_t * ctg = mem_alloc(sizeof(ctg_t));

  if (ctg == NULL) {
    return ctg;
  }

  memset(ctg, 0, sizeof(ctg_t));

  ctg->magic_n = CTG_MAGIG_N;
  ctg->ctg_visited = 1;
  ctg->kernel_mutex = the_kernel_mutex; /* null in unit tests */
  ctg->nodes = hash_create(n_buckets);
  ctg->ctg_size = 0;
  ctg->debug_flag = &CTG_DEFAULT_DEBUG;
  ctg->prev_oldest_active_start = ut_dulint_create(0, 0);

  UT_LIST_INIT(ctg->commit_ts_order);

  return ctg;
}


void ctg_set_dangerous_structure_testing(ibool on_off) {
  dangerous_structure_testing = on_off;
}


ibool ctg_get_dangerous_structure_testing(void) {
  return dangerous_structure_testing;
}



ulint ctg_prune_count(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->prune_count;
}



ulint ctg_prune_nodes_examined(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->prune_nodes_examined;
}



ulint ctg_tests_done(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->tests_done;
}



ulint ctg_tests_sum_size(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->tests_sum_size;
}



ulint ctg_cycles_found(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->cycles_found;
}



ulint ctg_max_ctg_path(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->max_ctg_path;
}



ulint ctg_sum_ctg_path(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->sum_ctg_path;
}



ulint ctg_max_cycle_length(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->max_cycle_length;
}



ulint ctg_sum_cycle_length(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->sum_cycle_length;
}


ulint ctg_sum_edges_traversed(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->sum_edges_traversed;
}


ulint ctg_add_edge_count(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->add_edge_count;
}


ulint ctg_sum_bolt_length(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->sum_bolt_length;
}


ulint ctg_trx_with_bolts(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->trx_with_bolts;
}


ulint ctg_max_bolt_length(const ctg_t * ctg) {
  ut_ad(ctg);
  return ctg->max_bolt_length;
}


void ctg_set_debug_flag(ctg_t * ctg, const char * debug_flag) {
  ut_ad(ctg != NULL && ctg->magic_n == CTG_MAGIG_N);
  if (debug_flag) {
    ctg->debug_flag = debug_flag;
  }
  else {
    ctg->debug_flag = &CTG_DEFAULT_DEBUG;
  }
}


void ctg_free(ctg_t ** ctg) {
  CHECK_CTG(*ctg);
  
  ctg_t * tmp = *ctg;

  ulint n_cells = hash_get_n_cells(tmp->nodes);
  ulint iter = 0;
  trx_t * trx = NULL;

  /* walk through the hash table.  Zero out everybody's pointers */
  for (; iter < n_cells; iter++) {

    hash_cell_t * cell = hash_get_nth_cell(tmp->nodes, iter);

    while (cell->node != NULL) {
      trx = cell->node;
      cell->node = trx->ctg_hash;               
      trx->ctg_hash = NULL;

      ctg_edgelist_clear(trx->out_edges);
      trx->in_edge_count = 0;
    }
  }

  /* walk through commit_ts_order.  Unlink everyone */
  trx = NULL;
  while ((trx = UT_LIST_GET_FIRST(tmp->commit_ts_order)) != NULL) {
    UT_LIST_REMOVE(commit_ts_order, tmp->commit_ts_order, trx);
  }

  hash_table_free(tmp->nodes);
  tmp->nodes = NULL;
  tmp->magic_n = 0;
  mem_free(tmp);
  *ctg = NULL;
}


/**
 * Assertion to ensure that the commit-timestamp ordered
 * list really is in commit timestamp order
 */
#ifdef UNIV_DEBUG
static ibool list_in_timestamp_order(const ctg_t * ctg) {
  ut_ad(ctg);

  if (UT_LIST_GET_LEN(ctg->commit_ts_order) == 0) {
    return TRUE;
  }

  dulint prior_ts = ut_dulint_zero;
  trx_t * trx = UT_LIST_GET_FIRST(ctg->commit_ts_order);

  while (trx != NULL) {

    dulint current_ts = trx_get_commit_ts(trx);

    /* should have: ut_dulint_zero < commit_ts <= ut_dulint_max */
    ut_ad(ut_dulint_cmp(ut_dulint_zero, current_ts) < 0);
    ut_ad(ut_dulint_cmp(ut_dulint_max, current_ts) >= 0);
    
    if (ut_dulint_cmp(prior_ts, current_ts) > 0) {
      fprintf(stderr, "commit_ts_order out of order: %u.%u before %u.%u\n",
              (unsigned) ut_dulint_get_high(prior_ts),
              (unsigned) ut_dulint_get_low(prior_ts),
              (unsigned) ut_dulint_get_high(current_ts),
              (unsigned) ut_dulint_get_low(current_ts));
      return FALSE;
    }

    prior_ts = current_ts;
    trx = UT_LIST_GET_NEXT(commit_ts_order, trx);
  }
    
  return TRUE;
}
#endif


/* Adds trx to our list of commit-timestamp ordered transactions.
 * When Ti is added, it hasn't been assigned a commit timestamp.
 * Therefore the commit_ts == ut_dulint_max, and we can always append
 * it to the end.
 *
 * After the final commit_ts is assigned, commit code will
 * have to ask for trx to be repositioned.
 */
static void append_commit_ts_order(ctg_t * ctg, trx_t * trx) {
  ut_ad(ctg);
  ut_ad(trx);

  UT_LIST_ADD_LAST(commit_ts_order, ctg->commit_ts_order, trx);
}


void ctg_reposition_commit_ts(ctg_t * ctg, trx_t * trx) {
  CHECK_CTG(ctg);
  CHECK_TRX(trx);

  ut_ad(trx_assigned_commit_ts(trx));
  ut_ad(ctg_get(ctg, trx_get_trx_id(trx)) == trx);

  UT_LIST_REMOVE(commit_ts_order, ctg->commit_ts_order, trx);

  if (UT_LIST_GET_LEN(ctg->commit_ts_order) == 0) {
    UT_LIST_ADD_FIRST(commit_ts_order, ctg->commit_ts_order, trx);
    goto finish_reposition;
  }

  trx_t * point = UT_LIST_GET_LAST(ctg->commit_ts_order);
  ut_ad(point != NULL);

  while (point != NULL) {

    if (ut_dulint_cmp(trx_get_commit_ts(point),
                      trx_get_commit_ts(trx)) < 0) {

      UT_LIST_INSERT_AFTER(commit_ts_order, ctg->commit_ts_order,
                           point, trx);

      goto finish_reposition;
    }

    point = UT_LIST_GET_PREV(commit_ts_order, point);

  }
    
  UT_LIST_ADD_FIRST(commit_ts_order, ctg->commit_ts_order, trx);


 finish_reposition:
  ut_ad(list_in_timestamp_order(ctg));
  return;
}



ibool ctg_add(ctg_t * ctg, trx_t * trx) {
  CHECK_CTG(ctg);
  CHECK_TRX(trx);
  
  trx_t * found = ctg_get(ctg, trx_get_trx_id(trx));

  if (found != NULL) {
    /* already there */
    return FALSE;
  }

  ulint fold = ctg_fold(trx);
  hash_cell_t * nth_cell = hash_get_nth_cell(ctg->nodes, 
                                             hash_calc_hash(fold, ctg->nodes));
  
  /* stick in front */
  trx_t * first = nth_cell->node;
  nth_cell->node = trx;
  trx->ctg_hash = first;
  trx->in_edge_count = 0;
  ctg->ctg_size++;

  append_commit_ts_order(ctg, trx);

  ut_ad(UT_LIST_GET_LEN(ctg->commit_ts_order) == ctg_size(ctg));

  return TRUE;
}



trx_t * ctg_get(const ctg_t * ctg, dulint trxid) {
  CHECK_CTG(ctg);
  ulint fold = ctg_fold_id(trxid);

  trx_t * found = NULL;
  HASH_SEARCH(ctg_hash, ctg->nodes, fold, found, 
              UT_DULINT_EQ(trxid, trx_get_trx_id(found)));

  return found;
}


ibool ctg_add_edge(ctg_t * ctg, trx_t * source, trx_t * sink) {
  CHECK_CTG(ctg);
  ut_ad(source);
  ut_ad(sink);

  ut_ad(ctg_get(ctg, trx_get_trx_id(source)) != NULL);
  ut_ad(ctg_get(ctg, trx_get_trx_id(sink)) != NULL);
  
  ibool was_added = ctg_edgelist_add(&(source->out_edges), 
                                     source->lock_heap,
                                     trx_get_trx_id(sink));

  if (was_added) {
    sink->in_edge_count++;

    ctg->add_edge_count++;

    /* [SREVILAK 2010-01-09].  Double adjacency list for
     * DS testing, if warranted */
    if (UNIV_UNLIKELY(dangerous_structure_testing)) {
      was_added = ctg_edgelist_add(&(sink->in_edges),
                                    sink->lock_heap,
                                    trx_get_trx_id(source));
      ut_ad(was_added);
      
      /* [SREVILAK 2010-01-09].  Note: in_edge_count is the true
       * number of in edges.  trx->in_edges is allowed to have
       * trx id's of transactions that are no longer in the graph */
    }

    if (UNIV_UNLIKELY(*ctg->debug_flag)) {
        fprintf(stderr, "ADD_EDGE: %u.%u -> %u.%u\n",
                (unsigned) source->id.high, 
                (unsigned) source->id.low,
                (unsigned) sink->id.high, 
                (unsigned) sink->id.low);
      }
  }

  return was_added;
}



ibool ctg_remove(ctg_t * ctg, trx_t * trx) {
  CHECK_CTG(ctg);
  ut_ad(trx);

  /* make sure our transaction is there */
  trx_t * found_trx = ctg_get(ctg, trx_get_trx_id(trx));
  if (found_trx == NULL) {
    ut_ad(ctg_size(ctg) == UT_LIST_GET_LEN(ctg->commit_ts_order));
    return FALSE;
  }
  ut_ad(found_trx == trx);

  /* pull trx out of the hash */
  ulint fold = ctg_fold(trx);
  HASH_DELETE(trx_t, ctg_hash, ctg->nodes, fold, trx);
  
  /* Decrement in-edge count for each sink */
  ulint iter = 0;

  for (; iter < EDGELIST_USED(trx->out_edges); iter++) {
    dulint sink_id = ctg_edgelist_get(trx->out_edges, iter);

    trx_t * sink = ctg_get(ctg, sink_id);
    if (sink == NULL) {
      /* no longer in graph */
      continue;
    }

    ut_ad(sink->in_edge_count > 0);
    sink->in_edge_count--;

    if (UNIV_UNLIKELY(dangerous_structure_testing)) {
      /* [SREVILAK 2010-01-09].  Note: removing edges is expensive,
       * since it (potentially) involves pulling an element out of the
       * middle of an array.  Accordingly, we don't clean up sink's
       * in-edges.  Instead, we just prepare ourselves for the
       * possibility that a trx_id in in-edges is not in the graph.
       * Of course, we already make the same assumption for
       * out_edges */
    }
  }

  /* zero out the edgelist */
  ctg_edgelist_clear(trx->out_edges);

  if (UNIV_UNLIKELY(dangerous_structure_testing)) {
    ctg_edgelist_clear(trx->in_edges);
  }

  trx->in_edge_count = 0;
  trx->ctg_hash = NULL;

  ut_ad(ctg->ctg_size > 0);
  ctg->ctg_size--;

  /* remove from commit_ts_order */
  UT_LIST_REMOVE(commit_ts_order, ctg->commit_ts_order, trx);
  ut_ad(ctg_size(ctg) == UT_LIST_GET_LEN(ctg->commit_ts_order));

  return TRUE;
}


static void dulint_print(FILE * file, dulint d, ibool newline) {
  fprintf(file, "(%u.%u)%s",
          (unsigned) ut_dulint_get_high(d),
          (unsigned) ut_dulint_get_low(d),
          (newline ? "\n" : ""));
}


void ctg_edgelist_dump(const trx_t * trx, FILE * file) {
  if (file == NULL) {
    file = stderr;
  }
  if (trx == NULL) {
    fprintf(file, "[null trx_t]\n");
    return;
  }

  ulint iter = 0;
  dulint trx_id = trx_get_trx_id(trx);

  if (UNIV_UNLIKELY(dangerous_structure_testing)) {
    fprintf(file, "  OUT edges:\n");
  }

  for (; iter < ctg_edgelist_size(trx->out_edges); iter++) {
    dulint sink = ctg_edgelist_get(trx->out_edges, iter);
    fprintf(file, "  ");
    dulint_print(file, trx_id, FALSE);
    fprintf(file, " -> ");
    dulint_print(file, sink, TRUE);
  }

  if (UNIV_LIKELY(! dangerous_structure_testing)) {
    return;
  }

  ut_ad(trx->in_edges != NULL);

  /* [SREVILAK 2010-01-09].  If using double adjacency lists,
   * print the other adjacency list */
  fprintf(file, "  IN edges:\n");
  for (iter= 0; iter < ctg_edgelist_size(trx->in_edges); iter++) {
    dulint source = ctg_edgelist_get(trx->in_edges, iter);
    fprintf(file, "  ");
    dulint_print(file, source, FALSE);
    fprintf(file, " -> ");
    dulint_print(file, trx_id, TRUE);
  }
}


/*
 * Dump the entire CTG to file
 */
void ctg_dump(const ctg_t * ctg, FILE * file) {

  if (file == NULL) {
    file = stderr;
  }

  if (ctg == NULL) {
    fprintf(file, "[null ctg_t]\n");
    return;
  }

  ulint n_cells = hash_get_n_cells(ctg->nodes);
  ulint iter = 0;

  fprintf(file, "CTG: size=%u\n", (unsigned) ctg->ctg_size);

  for (; iter < n_cells; iter++) {

    hash_cell_t * cell = hash_get_nth_cell(ctg->nodes, iter);
    trx_t * trx = cell->node;

    while (trx != NULL) {
      fprintf(file, "cell=%u, trx_id=", (unsigned) iter);
      dulint_print(file, trx_get_trx_id(trx), FALSE);
      fprintf(file, ", in_edges=%u\n", (unsigned) trx->in_edge_count);
      ctg_edgelist_dump(trx, file);
      trx = HASH_GET_NEXT(ctg_hash, trx);
    }
    
  }

}


void ctg_dump_histogram(const ctg_t * ctg, FILE * file) {
  ulint idx;
  ulint last_idx = 0;

  ut_ad(ctg);
  ut_ad(file);

  
  /* find highest numbered index where there is something to show */
  for (idx = 0; idx < CTG_HISTOGRAM_SIZE; idx++) {
    if (ctg->histogram[idx] > 0) {
      last_idx = idx + 1;
    }
  }

  fprintf(file, "cycle_length\tcount\n");
  for (idx = 0; idx < last_idx; idx++) {
    fprintf(file, "%lu\t%lu\n", 
            CTG_INDEX_TO_LENGTH(idx),
            ctg->histogram[idx]);
  }
  fprintf(file, "\n");
}


ulint ctg_size(const ctg_t * ctg) {
  ut_ad(ctg);
  if (ctg) {
    return ctg->ctg_size;
  }
  else {
    return ULINT_UNDEFINED;
  }
}



static ulint ctg_cycle_test_r(ctg_t * ctg, dulint start, 
                              trx_t * current, 
                              ulint depth,
                              cycle_test_notes_t * ct_notes) {
  ut_ad(ctg);
  ut_ad(current);
  ut_ad(ct_notes);

  /* [SREVILAK 2010-01-25].  Keep track of longest path */
  if (depth > ct_notes->longest_path) {
    ct_notes->longest_path = depth;
  }

  if (depth >= CTG_MAX_RECURSION) {
    return DB_CTG_TOO_LONG;
  }

  /* been there, done that? */
  if (current->ctg_visited == ctg->ctg_visited) {
    return DB_SUCCESS;
  }

  current->ctg_visited = ctg->ctg_visited;

  /* recursively check out edges */
  ulint n_edges = ctg_edgelist_size(current->out_edges);
  ulint iter = 0;
  
  for (; iter < n_edges; iter++) {
    dulint to_check = ctg_edgelist_get(current->out_edges, iter);
    ulint err = DB_SUCCESS;

    /* cycle ? */
    if (UT_DULINT_EQ(to_check, start)) {

      /* note cycle and path length */
      ct_notes->cycle_length = depth + 1;
      if (ct_notes->longest_path < ct_notes->cycle_length) {
        /* adjust, since we short circuit the recursive call */
        ct_notes->longest_path = ct_notes->cycle_length;
      }

      ctg_record_cycle_length(ctg, ct_notes->cycle_length);

      return DB_DEADLOCK;
    }

    trx_t * sink = ctg_get(ctg, to_check);
    if (sink == NULL) {
      /* no longer in graph */
      continue;
    }

    /* recursion means that we traverse an edge */
    ct_notes->edges_traversed++;

    /* if recursion finds a cycle, propagate up the chain */    
    err = ctg_cycle_test_r(ctg, start, sink, depth + 1, ct_notes);

    if (err != DB_SUCCESS) {
      return err;
    }
  }

  return DB_SUCCESS;
}


/**
 * Compute the "bolt_length" of trx, which is the number of (1 +
 * number of edges) that trx travels backwards in time.  In other
 * words, bolt_length reflects the number of transactions in the bolt,
 * rather than the number of edges.
 */
static ulint compute_bolt_length(const ctg_t * ctg, 
                                 const trx_t * trx) {
  ut_ad(ctg);
  ut_ad(trx);

  ulint n_out_edges = ctg_edgelist_size(trx->out_edges);
  ulint iter = 0;
  ulint max_bolt = 0;
  ibool have_sink = FALSE;
  
  /* if no out edges, then we have no bolt */
  if (n_out_edges == 0) {
    return 0;
  }

  /* Examine trx out-edges.  Ti is committing.  Therefore each Ti->Tj
   * is an anti-dependency edge where commit(Tj) < commit(Ti).  We
   * have time travel because the commit order does not match the
   * direction of dependencies. */

  /* Find the max bolt length in { Tj | Ti -> Tj } */
  for (; iter < n_out_edges; iter++) {
    dulint sink_id = ctg_edgelist_get(trx->out_edges, iter);
    trx_t * sink = ctg_get(ctg, sink_id);
    
    if (sink == NULL) {
      /* no longer in graph */
      continue;
    }

    have_sink = TRUE;

    if (sink->bolt_length > max_bolt) {
      max_bolt = sink->bolt_length;
    }
  }

  if (! have_sink) {
    /* nothing should have had a chance to prune any Tj yet,
     * so assert that a sink exists */
    ut_ad(have_sink);
    return 0;
  }

  /* we want to measure bolt length as the number of transactions, not
   * the number of edges.  If max_bolt == 0, then we have a bolt
   * length of 2 (and one edge).  Otherwise, we have a bolt_length of
   * (1 + max_bolt) */
  if (max_bolt == 0) {
    return 2;
  }
  else {
    ut_ad(max_bolt != 1);
    return 1 + max_bolt;
  }
}


static void record_bolt_length(ctg_t * ctg, const trx_t * trx) {
  ut_ad(ctg);
  ut_ad(trx);

  ctg->sum_bolt_length += trx->bolt_length;

  if (trx->bolt_length > ctg->max_bolt_length) {
    ctg->max_bolt_length = trx->bolt_length;
  }

  if (trx->bolt_length > 0) {
    ctg->trx_with_bolts++;
  }
}



ulint ctg_cycle_test(ctg_t * ctg, trx_t * trx) {
  CHECK_CTG(ctg);
  ut_ad(trx);

  cycle_test_notes_t ct_notes;

  dulint start_id = trx_get_trx_id(trx);

  trx_t * found = ctg_get(ctg, start_id);
  if (! found) {
    return DB_SUCCESS;
  }
  ut_ad(found == trx);

  ut_ad(trx->bolt_length == 0);
  trx->bolt_length = compute_bolt_length(ctg, trx);
  record_bolt_length(ctg, trx);

  if (UNIV_UNLIKELY(trx->bolt_length > CTG_MAX_RECURSION)) {
    return DB_CTG_TOO_LONG;
  }

  /* if trx has no in-edge, then no work to do */
  if (trx->in_edge_count == 0) {
    return DB_SUCCESS;
  }

  /* bump the visit number */
  ctg->ctg_visited++;

  memset(&ct_notes, 0, sizeof(ct_notes));

  ulint err = ctg_cycle_test_r(ctg, start_id, trx, 0, &ct_notes);

  /* sanity check on longest_path */
  ut_ad(ct_notes.longest_path <= ctg->ctg_size);
  
  /* keep track of longest paths traversed (avg. max path length) */
  ctg->sum_ctg_path += ct_notes.longest_path;

  /* keep track of sum of edges traversed */
  ctg->sum_edges_traversed += ct_notes.edges_traversed;

  /* was this the longest path? */
  if (ct_notes.longest_path >= ctg->max_ctg_path) {
    ctg->max_ctg_path = ct_notes.longest_path;
  }

  /* extra bookkeeping for cycles, if warranted */
  if (err == DB_DEADLOCK) {
    ut_ad(ct_notes.cycle_length >= 2);

    /* [SREVILAK 2010-01-25] add to total lengths of cycles, so we can
     * report average cycle length */
    ctg->sum_cycle_length += ct_notes.cycle_length;

    /* [SREVILAK 2010-01-25] is this the longest cycle we've found */
    if (ct_notes.cycle_length > ctg->max_cycle_length) {
      ctg->max_cycle_length = ct_notes.cycle_length;
    }

#ifdef UMB_PRINT_CYCLES
    fprintf(stderr, "CYCLE: Length %lu\n", (ulint) ct_notes.cycle_length);
#endif

  }

  ut_ad(err == DB_SUCCESS
        || err == DB_DEADLOCK
        || err == DB_CTG_TOO_LONG);

  return err;
}



ulint ctg_cycle_or_ds_test(ctg_t * ctg, trx_t * trx) {

  ut_ad(ctg);

  ctg->tests_done++;
  ctg->tests_sum_size += ctg->ctg_size;

  ulint result = DB_SUCCESS;

  if (UNIV_UNLIKELY(dangerous_structure_testing)) {
    result = ctg_ds_test(ctg, trx);
  }
  else {
    result = ctg_cycle_test(ctg, trx);
  }

  if (UNIV_UNLIKELY(result == DB_DEADLOCK)) {
    ctg->cycles_found++;
  }

  ut_ad(result == DB_SUCCESS
        || result == DB_DEADLOCK
        || result == DB_CTG_TOO_LONG);

  return result;
}



/*******************************************************************
 * pruning.
 */

static void notify_observers(ctg_t * ctg, trx_t * trx) {
  ut_ad(trx);
  ut_ad(ctg);
  ulint iter = 0;
  for (; iter < ctg->num_observers; iter++) {
    ctg->prune_observers[iter](trx);
  }
}


/**
 * [SREVILAK 2010-01-09].  Test of in-edge criteria for pruning.
 * For cycle testing, we require in_edge_count == 0.  For dangerous
 * structure testing, we ignore in_edge_count .
 */
UNIV_INLINE
ibool trx_satisfies_in_edge_criteria(const trx_t * trx) {
 
  if (UNIV_UNLIKELY(dangerous_structure_testing)) {
     return TRUE;
   }

   return (trx->in_edge_count == 0);
}


UNIV_INLINE
ibool trx_can_be_pruned(const trx_t * trx, dulint oldest_active_start) {

  ut_ad(trx);

  /* step 1: trx must be committed. */
  if (! trx_has_committed(trx)) {
    return FALSE;
  }

  /* step 1a: if trx does't have state CTG_RESIDENT, then it's
   * still in the process of committing.  Don't prune yet */
  if (trx->conc_state != TRX_CTG_RESIDENT) {
    ut_ad(trx->conc_state == TRX_COMMITTED_IN_MEMORY
          || trx->conc_state == TRX_ACTIVE
          || trx->conc_state == TRX_PREPARED);
    return FALSE;
  }

  /* step 2: in-edge requirement */
  if (! trx_satisfies_in_edge_criteria(trx)) {
    return FALSE;
  }

  /* step 3: trx must have committed before the oldest active
   * transaction started */
  if (ut_dulint_cmp(trx_get_commit_ts(trx),
                    oldest_active_start) >= 0) {
    return FALSE;
  }
  
  return TRUE;
}



/* Build the starting set for CTG pruning.  These are transactions that
 * (a) committed before the oldest active transaction started, and 
 * (b) have no in edges.
 *
 * ctg: cycle testing graph
 *
 * oldest_active_start: trx->id of oldest active transaction
 *
 * start_set: OUT: buffer where we place transaction ids in
 * the start set.
 *
 * RETURN: number of transactions in the starting set
 */
static ulint ctg_prune_starting_set(ctg_t * ctg, 
                                    dulint oldest_active_start, 
                                    dyn_array_t * start_set) {
  ut_ad(ctg);
  ut_ad(start_set);
  ulint count = 0;
  trx_t * trx;

  if (UT_LIST_GET_LEN(ctg->commit_ts_order) == 0) {
    ut_ad(ctg_size(ctg) == 0);
    return 0;
  }

  /* [SREVILAK 2010-01-25].  Note: if a txn is not part of the 
   * starting set, then we count it as examined.  We had to examine
   * it in order to determine that it should not be part of the
   * starting set.
   *
   * If a txn is part of the starting set, then we will
   * increment prune_nodes_examined in prune_sub_r.
   *
   * We don't want to double count examinations, so we only
   * increment prune_nodes_examined when we reject nodes from
   * the starting set.
   */

  for (trx = UT_LIST_GET_FIRST(ctg->commit_ts_order);
       trx;
       trx = UT_LIST_GET_NEXT(commit_ts_order, trx)) {

    ctg->prune_nodes_examined++;

    if (trx_can_be_pruned(trx, oldest_active_start)) {
      /* add Ti to starting set */
      ut_ad(trx_assigned_commit_ts(trx));
      
      dulint * ptr = dyn_array_push(start_set, sizeof(dulint));
      *ptr = trx_get_trx_id(trx);
      ut_ad(ctg_get(ctg, *ptr) != NULL);

      trx->ctg_prune_starting_set = TRUE;
      count++;
    }

    else {
      
      dulint commit_ts = trx_get_commit_ts(trx);

      /* if Ti commits after oldest active start, then
       * we are done */
      if (ut_dulint_cmp(commit_ts, oldest_active_start) >= 0) {
        /* note: trx may be awaiting a commit timestamp */
        break; 
      }
    }
  }

  return count;
}


/*
 * Copies trx's out edges to out_edges
 *
 * Returns: number of edges copied 
 */
static ulint copy_out_edge_list(const trx_t * trx, 
                                dyn_array_t * out_edges) {
  ut_ad(trx);
  ut_ad(out_edges);

  ulint edge_count = ctg_edgelist_size(trx->out_edges);
  ulint iter;

  for (iter = 0; iter < edge_count; iter++) {
    dulint trx_id = ctg_edgelist_get(trx->out_edges, iter);
    dulint * trx_id_ptr = dyn_array_push(out_edges, sizeof(dulint));
    *trx_id_ptr = trx_id;
  }

  return edge_count;
}



/**
 * Recursive pruning.
 *
 * (1) see if trx_id can be pruned.  If not, return
 * 
 * (2) if trx can be pruned, assemble the set of out
 * edges.  Prune trx, and recurively prune every sink of trx
 *
 * ctg: cycle testing graph
 *
 * oldest_active_start: trx->id of oldest active transaction
 *
 * trx_id: transaction to (maybe) prune
 *
 * hit_bottom: in/out.  Did recursion hit CTG_MAX_RECURSION?
 *
 * depth: current depth of recursion
 *
 * Return: number of transactions pruned.
 */
static ulint ctg_prune_sub_r(ctg_t * ctg, 
                             dulint oldest_active_start, 
                             dulint trx_id,
                             ulint * hit_bottom,
                             ulint depth) {

  CHECK_CTG(ctg);
  ut_ad(! UT_DULINT_EQ(trx_id, ut_dulint_max));
  ut_ad(hit_bottom != NULL);

  ulint prune_count = 0;
  dyn_array_t out_edges;
  ulint n_edges = 0;
  ulint iter;

  /* if we've hit the maximum depth of recursion, then back off.
   * ctg_prune will try again */
  if (depth >= CTG_MAX_RECURSION) {
    *hit_bottom = TRUE;
  }
  if (*hit_bottom) {
    return 0;
  }

  trx_t * trx = ctg_get(ctg, trx_id);
  if (trx == NULL) {
    /* already pruned.  Shouldn't happen, but be defensive */
    return 0;
  }

  /* If trx not in starting set and not prunable, then 
   * we're done with this branch */
  if (trx->ctg_prune_starting_set == FALSE
      && trx_can_be_pruned(trx, oldest_active_start) == FALSE) {
    return 0;
  }

  if (! trx->ctg_prune_starting_set) {
    /* Count the node we just examined */
    ctg->prune_nodes_examined++;
  }

  /* we can prune trx */
  
  dyn_array_create(&out_edges);

  /* copy the set of out edges */
  n_edges = copy_out_edge_list(trx, &out_edges); 
  
  /* remove trx */
  ibool did_remove = ctg_remove(ctg, trx);
  did_remove = did_remove;
  ut_ad(did_remove);
  prune_count++;

  /* notify prune observers */
  notify_observers(ctg, trx);
      
  /* recursively (try_to) remove all sink nodes */
  for (iter = 0; iter < n_edges; iter++) {
    dulint * sink_id = dyn_array_get_element(&out_edges, 
                                             iter * sizeof(dulint));
    prune_count += ctg_prune_sub_r(ctg, 
                                   oldest_active_start, 
                                   *sink_id,
                                   hit_bottom, depth + 1);
  }

  dyn_array_free(&out_edges);

  return prune_count;
}


/*
 * Lower level entry point for ctg pruning.
 *
 * ctg: cycle testing graph
 *
 * oldest_active_start: trx->id of oldest active transaction
 *
 * hit_bottom: in/out.  Did recursion reach a depth of CTG_MAX_RECURSION?
 *
 * Return: number of nodes removed
 */
static ulint ctg_prune_sub(ctg_t * ctg, 
                           dulint oldest_active_start,
                           ibool * hit_bottom) {

  CHECK_CTG(ctg);

  dyn_array_t start_set;
  ulint prune_count = 0;
  
  dyn_array_create(&start_set);

  /* collect starting list of transaction ids */
  ulint start_size = ctg_prune_starting_set(ctg, oldest_active_start, &start_set);
  ut_ad((start_size * sizeof(dulint)) == dyn_array_get_data_size(&start_set));

  if (start_size == 0) {
    dyn_array_free(&start_set);
    return 0;
  }

  /* Recursively prune each transaction in the starting set */
  ulint iter = 0;
  for (; iter < start_size; iter++) {
    dulint * trx_id_ptr = dyn_array_get_element(&start_set, iter * sizeof(dulint));

    ulint pruned = ctg_prune_sub_r(ctg, 
                                   oldest_active_start, 
                                   *trx_id_ptr,
                                   hit_bottom, 0);
    prune_count += pruned;
  }
  
  dyn_array_free(&start_set);

  return prune_count;
}


ulint ctg_prune(ctg_t * ctg, dulint oldest_active_start) {
  CHECK_CTG(ctg);
  
  ibool hit_bottom;
  ulint changed = 0;
  int counter = 0;

  ctg->prune_count++;

  if (ctg_size(ctg) == 0) {
    return 0;
  }

  /**
   * [SREVILAK 2010-06-17].  If the start timestamp of the oldest
   * active transaction has not changed, then we won't find anything
   * to prune.  (Exception: if oldest active start is ut_dulint_max,
   * then there are no active transactions, and we definitely need to
   * prune).
   */
  if (ut_dulint_cmp(oldest_active_start, 
                    ctg->prev_oldest_active_start) == 0
      && ut_dulint_cmp(oldest_active_start, ut_dulint_max) != 0) {
    return 0;
  }
  
  ctg->prev_oldest_active_start = oldest_active_start;

  /* pthreads seems to run out of stack space at ~ 350 recusive calls.
   * To stay within this constraint, be prepared to prune in chunks */
  hit_bottom = TRUE;

  while (hit_bottom == TRUE) {

    hit_bottom = FALSE;

    changed += ctg_prune_sub(ctg, oldest_active_start, &hit_bottom);    

    counter++;

    /* I don't believe this if test will ever be true, but let's add a
     * safeguard to ensure that the loop always terminates */
    if (counter >= 5) {
      break;
    }
  }

  return changed;
}


ibool ctg_register_observer(ctg_t * ctg, ctg_observer_t prune_observer) {
  ut_ad(ctg);
  ut_ad(prune_observer);

  if (ctg->num_observers >= CTG_MAX_OBSERVERS) {
    return FALSE;
  }

  ulint iter = 0;
  for (; iter < CTG_MAX_OBSERVERS; iter++) {

    if (ctg->prune_observers[iter] == prune_observer) {
      /* already registered */
      return FALSE;
    }

    if (ctg->prune_observers[iter] == NULL) {
      ut_ad(iter == ctg->num_observers);
      ctg->prune_observers[iter] = prune_observer;
      ctg->num_observers++;
      return TRUE;
    }
    
  }

  /* really shouldn't get here */
  ut_ad(0);
  return FALSE;
}


/*
 * Clears the list of registered observers.  
 *
 * ctg: the ctg whose list of observers should be cleared.
 *
 * RETURN: number of observers de-registered.
 */
ulint ctg_clear_observers(ctg_t * ctg) {
  ut_ad(ctg);
  ulint iter = 0;
  ulint count = 0;
  for (; iter < ctg->num_observers; iter++) {
    ut_ad(ctg->prune_observers[iter] != NULL);
    count++;
    ctg->prune_observers[iter] = NULL;
  }
  ctg->num_observers = 0;
  return count;
}



/*==================================================================
 * Dangerous structure testing
 */

static void ds_bookkeeping_clear(ds_bookkeeping_t * b) {
  ut_ad(b);
  memset(b, 0, sizeof(ds_bookkeeping_t));
}


/**
 * [SREVILAK 2010-01-10]
 *
 * Look for a T_out.  Given several candidate T_out's, choose the
 * one with the earliest commit timestamp 
 */
static void pivot_find_t_out(ctg_t * ctg, 
                             ds_bookkeeping_t * b) {
  ulint iter = 0;

  const trx_t * trx = b->pivot;
  ulint n_edges = ctg_edgelist_size(trx->out_edges);  

  ut_ad(b->out == NULL);

  for (; iter < n_edges; iter++) {

    dulint out_trx_id = ctg_edgelist_get(trx->out_edges, iter);

    trx_t * t_out = ctg_get(ctg, out_trx_id);

    if (t_out == NULL) {
      /* no longer in graph, continue */
      continue;
    }
    
    if (! trx_are_concurrent(trx, t_out)) {
      continue;
    }

    /* definitely an unsafe anti-dependency edge */

    /* First T_out candidate.  Remember and move on */
    if (b->out == NULL) {
      b->out = t_out;
      continue;
    }

    if (ut_dulint_cmp(trx_get_commit_ts(t_out),
                      trx_get_commit_ts(b->out)) < 0) {
      /* T_out committed earlier.  New candidate */
      b->out = t_out;
    }
  }
}


/**
 * [SREVILAK 2010-01-10]
 *
 * Look for a Tin.  Given several candidate T_in's, choose the
 * one with the latest commit timestamp 
 */
static void pivot_find_t_in(ctg_t * ctg,
                            ds_bookkeeping_t * b) {
  ulint iter = 0;

  const trx_t * trx = b->pivot;
  ulint n_edges = ctg_edgelist_size(trx->in_edges);  

  ut_ad(b->in == NULL);

  for (; iter < n_edges; iter++) {

    dulint in_trx_id = ctg_edgelist_get(trx->in_edges, iter);

    trx_t * t_in = ctg_get(ctg, in_trx_id);

    if (t_in == NULL) {
      /* pruned, continue */
      continue;
    }
    
    if (! trx_are_concurrent(trx, t_in)) {
      continue;
    }

    /* definitely an unsafe anti-dependency edge */

    /* First T_in candidate.  Remember and move on */
    if (b->in == NULL) {
      b->in = t_in;
      continue;
    }

    if (ut_dulint_cmp(trx_get_commit_ts(t_in),
                      trx_get_commit_ts(b->in)) > 0) {
      /* T_in committed later.  New candidate */
      b->in = t_in;
    }
  }
}


/**
 * Determine whether trx is a pivot transaction that must 
 * be aborted
 */
static ibool is_pivot_for_abort(ctg_t * ctg, ds_bookkeeping_t * b) {

  ut_ad(b->pivot != NULL);
  ut_ad(b->pivot != b->in);
  ut_ad(b->pivot != b->out);

  /* if either Tin, Tout are missing, then not a dangerous structure. */
  if (b->in == NULL || b->out == NULL) {
    return FALSE;
  }

  /* If T_in == T_out, we have a cycle of length two.  Abort. */
  if (b->in == b->out) {
    ctg_record_cycle_length(ctg, 2);
    return TRUE;
  }

  /* Pivot is committing now, so T_out must have committed earlier. 
   *
   * Some of our C unit tests get a little obnoxious here, so
   * test rather than assert */
  if (ut_dulint_cmp(trx_get_commit_ts(b->out),
                    trx_get_commit_ts(b->pivot)) > 0) {
    return FALSE;
  }

  /* if out commits before in, we have a dangerous structure */
  if (ut_dulint_cmp(trx_get_commit_ts(b->out),
                    trx_get_commit_ts(b->in)) < 0) {
    /* T_out commits before T_in */
    return TRUE;
  }

  /* not a pivot in a dangerous structure */

  return FALSE;
}



/**
 * Determine whether trx is Tpivot in a dangerous structure.
 */
static ibool is_t_pivot_in_ds(ctg_t * ctg, const trx_t * trx) {

  ds_bookkeeping_t book;

  CHECK_CTG(ctg);
  ut_ad(trx);

  ds_bookkeeping_clear(&book);

  book.pivot = trx;

  pivot_find_t_out(ctg, &book);

  pivot_find_t_in(ctg, &book);

  return is_pivot_for_abort(ctg, &book);
}


/**
 * Determine whether trx is Tin in a dangerous structure.
 */
static ibool is_t_in_in_ds(ctg_t * ctg, const trx_t * trx) {
  CHECK_CTG(ctg);
  ut_ad(trx);
  
  /* find candidate pivots, which are { Tj | Ti -> Tj }.
   *
   * If any Tj as bolt_length > 0, then trx is T_in in a
   * dangerous structure, and Tj is a pivot */

  ulint n_edges = ctg_edgelist_size(trx->out_edges);  
  ulint iter = 0;

  for (; iter < n_edges; iter++) {

    dulint sink_id = ctg_edgelist_get(trx->out_edges, iter);

    trx_t * sink_trx = ctg_get(ctg, sink_id);

    if (sink_trx == NULL) {
      /* pruned, continue */
      continue;
    }
    
    if (! trx_are_concurrent(trx, sink_trx)) {
      continue;
    }

    /* definitely an unsafe anti-dependency edge */

    if (sink_trx->bolt_length > 0) {
      /* sink is Tpivot, and bolt_length implies the existence
       * of Tout.  trx is Tin, and we have a dangerous structure */
      return TRUE;
    }
  }

  return FALSE;
}


/**
 * Given committing transaction "trx", determine
 * whether trx creates a dangerous structure.
 *
 * "trx" creates a dangerous structure if "trx" is a pivot transaction,
 * and T_out commits first.
 */
ulint ctg_ds_test(ctg_t * ctg, trx_t * trx) {

  CHECK_CTG(ctg);
  ut_ad(trx);

  ut_ad(trx->in_edges != NULL);
  ut_ad(trx->out_edges != NULL);
  ut_ad(trx->bolt_length == 0);

  trx->bolt_length = compute_bolt_length(ctg, trx);

  record_bolt_length(ctg, trx);

  if (trx->bolt_length == 0) {
    /* cannot be part of dangerous structure */
    return DB_SUCCESS;
  }

  if (is_t_pivot_in_ds(ctg, trx)) {
    return DB_DEADLOCK;
  }
  
  else if (is_t_in_in_ds(ctg, trx)) {
    return DB_DEADLOCK;
  }

  return DB_SUCCESS;
}
