/*
 * $Id: umb-test-ctg-basic.c,v 1.12 2010/07/05 19:15:52 srevilak Exp $
 */

#ifdef UNIV_DEBUG

#include "umb-test.h"
#include "lock0ctg.h"
#include "trx0trx.h"
#include "srv0srv.h"

static ctg_t * ctg;


static void setup_n_transactions(ulint n, trx_t ** array) {

  ctg = ctg_create(128, &kernel_mutex);

  mutex_enter(&kernel_mutex);

  ulint iter = 0;

  dulint counter = ut_dulint_create(0, 0);

  for (; iter < n; iter++) {
    counter = ut_dulint_add(counter, 1);
    array[iter] = trx_create(NULL);
    array[iter]->id = counter;
    counter = ut_dulint_add(counter, 1);
    array[iter]->no = counter;
  }

  mutex_exit(&kernel_mutex);
}


static void teardown(ulint n, trx_t ** array) {

  ulint iter;

  mutex_enter(&kernel_mutex);

  ctg_free(&ctg);

  for (iter = 0; iter < n; iter++) {
    ut_ad(array[iter]->ctg_hash == NULL);
    ut_ad(array[iter]->in_edge_count == 0);
    trx_free(array[iter]);
    array[iter] = NULL;
  }

  mutex_exit(&kernel_mutex);
}


/**
 * Put a transaction in.
 * Get the transaction out
 */
static void test_add_get(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_add_get");

  trx_t * trx_list[1];
  setup_n_transactions(1, trx_list);

  mutex_enter(&kernel_mutex);

  ibool added = ctg_add(ctg, trx_list[0]);
  ut_ad(added);

  added = ctg_add(ctg, trx_list[0]);
  ut_ad(! added);

  trx_t * retrieve = ctg_get(ctg, trx_get_trx_id(trx_list[0]));
  ut_ad(retrieve);
  
  ut_ad(retrieve == trx_list[0]);
  ut_ad(UT_DULINT_EQ(retrieve->id, trx_list[0]->id));

  mutex_exit(&kernel_mutex);

  teardown(1, trx_list);
}


static void test_add_remove_2(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_add_remove_2");
  trx_t * trx_list[2];
  const ulint n_trx = 2;
  ibool rv;
  trx_t * trx;
  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);
  
  /* put the transactions in the graph */
  rv = ctg_add(ctg, trx_list[0]);
  ut_ad(rv);

  rv = ctg_add(ctg, trx_list[1]);
  ut_ad(rv);

  /* ctg_dump(ctg, stderr); */

  /* verify that they are in the graph */
  trx = ctg_get(ctg, trx_get_trx_id(trx_list[0]));
  ut_ad(trx == trx_list[0]);

  trx = ctg_get(ctg, trx_get_trx_id(trx_list[1]));
  ut_ad(trx == trx_list[1]);

  /* remove one transaction */
  rv = ctg_remove(ctg, trx_list[0]);
  ut_ad(trx_list[0]->ctg_hash == NULL);
  ut_ad(rv);

  /* second remove should not do anything */
  rv = ctg_remove(ctg, trx_list[0]);
  ut_ad(! rv);

  /* ctg_dump(ctg, stderr); */

  /* [0] should not be in the graph, but [1] should be */
  trx = ctg_get(ctg, trx_get_trx_id(trx_list[0]));
  ut_ad(trx == NULL);

  trx = ctg_get(ctg, trx_get_trx_id(trx_list[1]));
  ut_ad(trx == trx_list[1]);

  /* now remove [1] */
  rv = ctg_remove(ctg, trx_list[1]);
  ut_ad(rv);

  /* neither should be in the graph */
  trx = ctg_get(ctg, trx_get_trx_id(trx_list[0]));
  ut_ad(trx == NULL);

  trx = ctg_get(ctg, trx_get_trx_id(trx_list[1]));
  ut_ad(trx == NULL);

  /* ctg_dump(ctg, stderr); */

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


/*
 *   1 -> 0
 *   1 -> 2
 *   2 -> 0
 *      
 * Remove 1
 */
static void test_remove_with_edges(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_remove_with_edges");
  trx_t * trx_list[3];
  const ulint n_trx = 3;
  ibool rv;
  trx_t * trx;
  ulint iter;
  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* put transactions in graph */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* create edges */
  rv = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[2], trx_list[0]);
  ut_ad(rv);

  /* ctg_dump(ctg, stderr); */
  
  /* verify number of in-edges */
  ut_ad(trx_list[0]->in_edge_count == 2);
  ut_ad(trx_list[1]->in_edge_count == 0);
  ut_ad(trx_list[2]->in_edge_count == 1);

  /* verify number of out edges */
  ut_ad(ctg_edgelist_size(trx_list[0]->out_edges) == 0);
  ut_ad(ctg_edgelist_size(trx_list[1]->out_edges) == 2);
  ut_ad(ctg_edgelist_size(trx_list[2]->out_edges) == 1);

  /* remove node [1] */
  rv = ctg_remove(ctg, trx_list[1]);
  ut_ad(rv);

  /* ctg_dump(ctg, stderr); */

  /* verify [1] no longer in graph */
  trx = ctg_get(ctg, trx_get_trx_id(trx_list[1]));
  ut_ad(trx == NULL);

  /* verify [0,2] in graph */
  trx = ctg_get(ctg, trx_get_trx_id(trx_list[0]));
  ut_ad(trx == trx_list[0]);

  trx = ctg_get(ctg, trx_get_trx_id(trx_list[2]));
  ut_ad(trx == trx_list[2]);

  /* verify in edges counts after remove */
  ut_ad(trx_list[0]->in_edge_count == 1);
  ut_ad(trx_list[2]->in_edge_count == 0);

  /* verify out_edge count */
  ut_ad(ctg_edgelist_size(trx_list[0]->out_edges) == 0);
  ut_ad(ctg_edgelist_size(trx_list[2]->out_edges) == 1);

  /* verify that 2->0 is our only edge */
  rv = ctg_edgelist_contains(trx_list[2]->out_edges, trx_get_trx_id(trx_list[0]));
  ut_ad(rv);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


/*
 * test with bad input.  txn not in ctg, so no cycle
 */
static void cycle_test_bad_txn(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_bad_txn");
  trx_t * trx_list[1];
  const ulint n_trx = 1;
  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  ulint result = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(result == DB_SUCCESS);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


/*
 * 0 -> 1 is not a cycle
 */
static void cycle_test_one_edge(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_one_edge");
  trx_t * trx_list[2];
  const ulint n_trx = 2;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* add edge */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  /* no cycle, starting from either transaction */
  rv = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(rv == DB_SUCCESS);

  rv = ctg_cycle_test(ctg, trx_list[1]);
  ut_ad(rv == DB_SUCCESS);

  /* in 0 -> 1, t0 has bolt length 2 */
  ut_ad(trx_list[0]->bolt_length == 2);
  ut_ad(trx_list[1]->bolt_length == 0);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


/*
 * 0 -> 1 -> 0 *is* is a cycle
 */
static void cycle_test_two_nodes(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_two_nodes");
  trx_t * trx_list[2];
  const ulint n_trx = 2;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* add edges */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(rv);

  /* there is a cycle, no matter where we start */
  rv = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(rv == DB_DEADLOCK);

  rv = ctg_cycle_test(ctg, trx_list[1]);
  ut_ad(rv == DB_DEADLOCK);

  /* given this order of testing, 0 has bolt length 2,
   * 1 has bolt length 3 */
  ut_ad(trx_list[0]->bolt_length == 2);
  ut_ad(trx_list[1]->bolt_length == 3);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}



/*
 * Graph is 0 -> 1 -> 2 -> 3 -> 1
 *
 * Starting cycle test at {1, 2, 3} should find a cycle.
 *
 * Starting test at zero should not find a cycle.  (Zero is not on
 * the cycle, and we never get back to zero
 */
static void cycle_test_spoon(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_spoon");
  trx_t * trx_list[4];
  const ulint n_trx = 4;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* add edges */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[2], trx_list[3]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[3], trx_list[1]);
  ut_ad(rv);


  /* zero does not lie on a cycle. */
  rv = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(rv == DB_SUCCESS);

  /* everything else lies on a cycle */
  for (iter = 1; iter < n_trx; iter++) {
    rv = ctg_cycle_test(ctg, trx_list[iter]);
    ut_ad(rv == DB_DEADLOCK);
  }

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);

}



/*
 * 0 -> 1 -> 2
 * 0 -> 3 -> 4 -> 2
 *
 * No cycle from any node
 */
static void cycle_test_splay(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_splay");
  trx_t * trx_list[5];
  const ulint n_trx = 5;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* 0 -> 1 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  /* 1 -> 2 */
  rv = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(rv);

  /* 0 -> 3 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[3]);
  ut_ad(rv);

  /* 3 -> 4 */
  rv = ctg_add_edge(ctg, trx_list[3], trx_list[4]);
  ut_ad(rv);

  /* 4 -> 2 */
  rv = ctg_add_edge(ctg, trx_list[4], trx_list[2]);
  ut_ad(rv);


  /* nothing lies on a cycle */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_cycle_test(ctg, trx_list[iter]);
    ut_ad(rv == DB_SUCCESS);
  }

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);

}



/*
 * 0 -> 1 -> 2 -> 3
 *
 * 0 -> 1 -> 4 -> 5 -> 0
 *
 * {2,3} do not lie on a cycle; {0,1,4,5} do.
 *
 * Note that some cases will require the recursion to backtrack.
 */
static void cycle_test_branch_loop(void) {
  UMB_TEST_FUNC_ANNOUNCE("cycle_test_branch_loop");
  trx_t * trx_list[6];
  const ulint n_trx = 6;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }


  /* first branch: 0 -> 1 -> 2 -> 3 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[2], trx_list[3]);
  ut_ad(rv);


  /* second branch: 0 -> 1 -> 4 -> 5 -> 0 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv == FALSE); /* already have this edge */

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[4]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[4], trx_list[5]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[5], trx_list[0]);
  ut_ad(rv);


  /* {2,3} not on cycle: everything else is */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_cycle_test(ctg, trx_list[iter]);
    if (iter == 2 || iter == 3) {
      ut_ad(rv == DB_SUCCESS);
    }
    else {
      ut_ad(rv == DB_DEADLOCK);
    }
  }

  /* ctg_dump(ctg, stderr); */

  /* all nodes have one in edge */
  for (iter = 0; iter < n_trx; iter++) {
    ut_ad(trx_list[iter]->in_edge_count == 1);
  }

  /* [1] has two out edges
   * [3] has no out edges
   *
   * everything else has one out edge */
  for (iter = 0; iter < n_trx; iter++) {
    ulint edge_count = ctg_edgelist_size(trx_list[iter]->out_edges);
    switch (iter) {
    case 1:
      ut_ad(edge_count == 2);
      break;
    case 3:
      ut_ad(edge_count == 0);
      break;
    default:
      ut_ad(edge_count == 1);
    }
  }

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);

}


static void test_reposition_commit_ts(void) {
  UMB_TEST_FUNC_ANNOUNCE(__func__);
  trx_t * trx_list[6];
  const ulint n_trx = 6;
  dulint counter = ut_dulint_create(0, 1);
  int iter;
  ibool rv;

  /* we need a slightly different setup here */
  setup_n_transactions(0, trx_list);
  
  mutex_enter(&kernel_mutex);

  for (iter = 0; iter < (int) n_trx; iter++) {
    trx_list[iter] = trx_create(NULL);
    counter = ut_dulint_add(counter, 1);
    trx_list[iter]->id = counter;
  }
  
  /* add transactions.  Note - no commit ts yet */
  for (iter = 0; iter < (int) n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* next, assign commit_ts in opposite order
   * and reposition */
  for (iter = n_trx - 1; iter >= 0; iter--) {
    counter = ut_dulint_add(counter, 1);
    trx_list[iter]->no = counter;
    trx_list[iter]->commit_ts = counter;
    ctg_reposition_commit_ts(ctg, trx_list[iter]);
  }


  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


/**
 * Add edges 0->1, 0->2.
 * Give 1 bolt_length = 2
 * Give 2 bolt_length = 4
 * 0 should get bolt length = 5
 */
static void test_bolt_length_branch(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_bolt_length_branch");
  trx_t * trx_list[3];
  const ulint n_trx = 3;
  ulint iter;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* add trx */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx_list[iter]);
    ut_ad(rv);
  }

  /* 0 -> 1 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  /* 0 -> 2 */
  rv = ctg_add_edge(ctg, trx_list[0], trx_list[2]);
  ut_ad(rv);

  trx_list[1]->bolt_length = 2;
  trx_list[2]->bolt_length = 4;

  rv = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(rv == DB_SUCCESS);

  ut_ad(trx_list[0]->bolt_length == 5);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);

}


/*
 * 0 -> 1 -> 2 , with commit order T2, T1, T0.
 *
 * Test bolt length after each commit.
 */
static void test_bolt_length_chain(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_bolt_length_chain");
  trx_t * trx_list[3];
  const ulint n_trx = 3;
  ulint rv;

  setup_n_transactions(n_trx, trx_list);

  mutex_enter(&kernel_mutex);

  /* Add t2, cycle test */
  rv = ctg_add(ctg, trx_list[2]);
  ut_ad(rv);
  
  rv = ctg_cycle_test(ctg, trx_list[2]);
  ut_ad(rv == DB_SUCCESS);

  ut_ad(trx_list[2]->bolt_length == 0);


  /* Add T1, with T1 -> T2.  Cycle test */
  rv = ctg_add(ctg, trx_list[1]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(rv);

  rv = ctg_cycle_test(ctg, trx_list[1]);
  ut_ad(rv == DB_SUCCESS);

  ut_ad(trx_list[2]->bolt_length == 0);
  ut_ad(trx_list[1]->bolt_length == 2);


  /* add T0, with T0 -> T1.  Test again */
  rv = ctg_add(ctg, trx_list[0]);
  ut_ad(rv);

  rv = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(rv);

  rv = ctg_cycle_test(ctg, trx_list[0]);
  ut_ad(rv == DB_SUCCESS);

  ut_ad(trx_list[2]->bolt_length == 0);
  ut_ad(trx_list[1]->bolt_length == 2);
  ut_ad(trx_list[0]->bolt_length == 3);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx_list);
}


UMB_TEST_ENTRY void umb_test_ctg_basic(void) {
  UMB_TEST_ANNOUNCE;

  test_add_get();
  test_add_remove_2();
  test_remove_with_edges();
  cycle_test_bad_txn();
  cycle_test_one_edge();
  cycle_test_two_nodes();
  cycle_test_spoon();
  cycle_test_splay();
  cycle_test_branch_loop();
  test_reposition_commit_ts();
  test_bolt_length_branch();
  test_bolt_length_chain();

}



#endif /* UNIV_DEBUG */
