/*
 * $Id: umb-test-ctg-ds.c,v 1.4 2010/07/08 02:31:59 srevilak Exp $
 */

#ifdef UNIV_DEBUG

#include "umb-test.h"
#include "lock0ctg.h"
#include "trx0trx.h"
#include "srv0srv.h"

static ctg_t * ctg;

static ibool ds_state;

static void setup_n_transactions(ulint n, 
                                 trx_t ** array,
                                 dulint * start_ts,
                                 dulint * commit_ts) {

  ds_state = ctg_get_dangerous_structure_testing();

  ctg_set_dangerous_structure_testing(TRUE);

  ctg = ctg_create(128, &kernel_mutex);

  mutex_enter(&kernel_mutex);

  ulint iter = 0;

  for (; iter < n; iter++) {
    array[iter] = trx_create(NULL);
    array[iter]->id = start_ts[iter];
    array[iter]->commit_ts = commit_ts[iter];
  }

  mutex_exit(&kernel_mutex);
}



static void teardown(ulint n, trx_t ** array) {

  ulint iter;

  mutex_enter(&kernel_mutex);

  ctg_free(&ctg);

  for (iter = 0; iter < n; iter++) {
    ut_ad(array[iter]->ctg_hash == NULL);
    ut_ad(array[iter]->in_edge_count == 0);
    trx_free(array[iter]);
    array[iter] = NULL;
  }

  mutex_exit(&kernel_mutex);

  ctg_set_dangerous_structure_testing(ds_state);
}


static void test_one_edge_adjacency(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_one_edge_adjacency");

  trx_t * trx_list[2];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    { 0, 1 },
    { 0, 2 }
  };

  dulint commit_ts[] = {
    { 0, 11 },
    { 0, 12 }
  };

  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  result = ctg_add(ctg, trx_list[0]);
  ut_ad(result);

  result = ctg_add(ctg, trx_list[1]);
  ut_ad(result);

  result = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(result);
  
  /* t0 should have an out edge to t1 */
  result = ctg_edgelist_contains(trx_list[0]->out_edges, 
                                 trx_get_trx_id(trx_list[1]));
  ut_ad(result);

  /* t1 should have an in-edge from t0 */
  result = ctg_edgelist_contains(trx_list[1]->in_edges,
                                 trx_get_trx_id(trx_list[0]));
  ut_ad(result);
  ut_ad(trx_list[1]->in_edge_count == 1);
  ut_ad(trx_list[0]->in_edge_count == 0);


  /* remove t0 */
  result = ctg_remove(ctg, trx_list[0]);
  ut_ad(result);

  /* t1 has no in edges */
  ut_ad(trx_list[1]->in_edge_count == 0);

  /* t0's edge lists have been cleared */
  ut_ad(ctg_edgelist_size(trx_list[0]->out_edges) == 0);
  ut_ad(ctg_edgelist_size(trx_list[0]->in_edges) == 0);

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);
}



/**
 * T2 -> T1 -> T0 where
 *
 *   -------------- T0
 *           ---------------- T1
 *                      ------------- T2
 *
 * This is a dangerous structure.  Regardless of which transaction
 * we test from, we should realize that this is a dangerous structure.
 */
static void test_ds_positive(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_positive");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
    {0, 30}
  };
  
  dulint commit_ts[] = {
    {0, 25},
    {0, 35},
    {0, 45}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  int iter;
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_add(ctg, trx_list[iter]);
    ut_ad(result == TRUE);
  }

  /* T2 -> T1 */
  result = ctg_add_edge(ctg, trx_list[2], trx_list[1]);
  ut_ad(result == TRUE);

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* We test as each transaction commit, and we rely on some properties
   * of temporal commit order.  For example a commiting Ti can never
   * be Tout in a dangerous structure (the other two transactions 
   * have not committed yet.  But we can still test Tpivot, and Tin */
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_ds_test(ctg, trx_list[iter]);
    if (iter == 0) {
      ut_ad(result == DB_SUCCESS);
    }
    else {
      ut_ad(result == DB_DEADLOCK);
    }
  }

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  
}



/**
 * T0 -> T1 -> T0 where
 *
 *   -------------- T0
 *           ---------------- T1
 *
 * This is a cycle of length two.  We should see a dangerous structure,
 * regardless of which transaction we test from.
 */
static void test_ds_cycle_2(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_cycle_2");

  trx_t * trx_list[2];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
  };
  
  dulint commit_ts[] = {
    {0, 25},
    {0, 35},
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  int iter;
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_add(ctg, trx_list[iter]);
    ut_ad(result == TRUE);
  }

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* T0 -> T1 */
  result = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(result == TRUE);

  /* Should find ds from any transaction */
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_ds_test(ctg, trx_list[iter]);
    ut_ad(result == DB_DEADLOCK);
  }

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  
}



/**
 * T0 -> T1 -> T where
 *
 *   ----T0
 *           --- T1
 *                   -- T2
 *
 * Transactions execute serially.  No dangerous structures.
 */
static void test_ds_serial_safe_dependencies(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_serial_safe_dependencies");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
    {0, 30}
  };
  
  dulint commit_ts[] = {
    {0, 15},
    {0, 25},
    {0, 35}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  int iter;
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_add(ctg, trx_list[iter]);
    ut_ad(result == TRUE);
  }

  /* T0 -> T1 */
  result = ctg_add_edge(ctg, trx_list[0], trx_list[1]);
  ut_ad(result == TRUE);

  /* T1 -> T2 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(result == TRUE);

  /* No DS from any transaction */
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_ds_test(ctg, trx_list[iter]);
    ut_ad(result == DB_SUCCESS);
  }

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  
}


/**
 * Test T2->T1->T0, where
 *
 *   ---------------- T0
 *         ------ T1
 *            -------------- T2
 *
 * Not a dangerous structure, since T_out = T0 and T0 does not
 * commit first
 */
static void test_ds_t_pivot_commits_first(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_t_pivot_commits_first");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
    {0, 30}
  };
  
  dulint commit_ts[] = {
    {0, 45},
    {0, 35},
    {0, 55}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* T1 commits first, not a dangerous structure */
  result = ctg_add(ctg, trx_list[1]);
  ut_ad(result == TRUE);

  result = ctg_ds_test(ctg, trx_list[1]);
  ut_ad(result == DB_SUCCESS);

  /* Add T0 */
  result = ctg_add(ctg, trx_list[0]);
  ut_ad(result == TRUE);

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* T0 commits next.  Stil no dangerous structure */
  result = ctg_ds_test(ctg, trx_list[0]);
  ut_ad(result == DB_SUCCESS);

  /* finally add T2 */
  result = ctg_add(ctg, trx_list[2]);
  ut_ad(result == TRUE);

  /* T2 -> T1 */
  result = ctg_add_edge(ctg, trx_list[2], trx_list[1]);
  ut_ad(result == TRUE);

  /* no dangerous structure for T2 */
  result = ctg_ds_test(ctg, trx_list[2]);
  ut_ad(result == DB_SUCCESS);

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  

}



/**
 * Test T2->T1->T0, where
 *
 *   --------------------------- T0
 *         -------------- T1
 *            ------- T2
 *
 * Not a dangerous structure, since T_out = T0 and T0 does not
 * commit first
 */
static void test_ds_t_in_commits_first(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_t_in_commits_first");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
    {0, 30}
  };
  
  dulint commit_ts[] = {
    {0, 55},
    {0, 45},
    {0, 35}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* T2 commits first */
  result = ctg_add(ctg, trx_list[2]);
  ut_ad(result == TRUE);

  /* not dangerous structure */
  result = ctg_ds_test(ctg, trx_list[2]);
  ut_ad(result == DB_SUCCESS);


  /* T1 commits next */
  result = ctg_add(ctg, trx_list[1]);
  ut_ad(result == TRUE);

  /* T2 -> T1 */
  result = ctg_add_edge(ctg, trx_list[2], trx_list[1]);
  ut_ad(result == TRUE);

  /* no DS at T1 */
  result = ctg_ds_test(ctg, trx_list[1]);
  ut_ad(result == DB_SUCCESS);


  /* T0 commits last */
  result = ctg_add(ctg, trx_list[0]);
  ut_ad(result == TRUE);

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* no ds from T0 */
  result = ctg_ds_test(ctg, trx_list[0]);
  ut_ad(result == DB_SUCCESS);


  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  

}


/*
 * Test T1 -> T0, T1 -> T2, where
 *
 *   --------- T0
 *        -------- T1
 *              -------- T2
 *
 * Not DS, since T1 has out edges, but no in-edges
 */
static void test_ds_no_pivot(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_no_pivot");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 20},
    {0, 30}
  };
  
  dulint commit_ts[] = {
    {0, 25},
    {0, 35},
    {0, 45}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  int iter;
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_add(ctg, trx_list[iter]);
    ut_ad(result == TRUE);
  }

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* T1 -> T2 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[2]);
  ut_ad(result == TRUE);

  /* No DS from any transaction */
  for (iter = 0; iter < n_trx; iter++) {
    result = ctg_ds_test(ctg, trx_list[iter]);
    ut_ad(result == DB_SUCCESS);
  }

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  

}


/**
 * Test T2->T1->T0
 *
 *   ------------ T0
 *        -------------- T1
 *          -------------------T2
 */
static void test_ds_in_last(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_in_last");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 20},
    {0, 30},
    {0, 40}
  };
  
  dulint commit_ts[] = {
    {0, 50},
    {0, 60},
    {0, 70}
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* T0 first */
  result = ctg_add(ctg, trx_list[0]);
  ut_ad(result == TRUE);
  
  result = ctg_ds_test(ctg, trx_list[0]);
  ut_ad(result == DB_SUCCESS);
  ut_ad(trx_list[0]->bolt_length == 0);


  /* T1 next */
  result = ctg_add(ctg, trx_list[1]);
  ut_ad(result == TRUE);

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* no dangerous structure yet */
  result = ctg_ds_test(ctg, trx_list[1]);
  ut_ad(result == DB_SUCCESS);
  ut_ad(trx_list[1]->bolt_length == 2);


  /* T2 next */
  result = ctg_add(ctg, trx_list[2]);
  ut_ad(result == TRUE);

  /* T2 -> T1 */
  result = ctg_add_edge(ctg, trx_list[2], trx_list[1]);
  ut_ad(result == TRUE);

  /* dangerous structure */
  result = ctg_ds_test(ctg, trx_list[2]);
  ut_ad(result == DB_DEADLOCK);
  ut_ad(trx_list[2]->bolt_length == 3);

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  

}



/**
 * Test T2->T1->T0
 *
 *   -------------- T0
 *             -------------- T1
 *     -------------------T2
 *
 */
static void test_ds_pivot_last(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_ds_t_pivot_last");

  trx_t * trx_list[3];
  const int n_trx = UT_ARR_SIZE(trx_list);
  ulint result;

  dulint start_ts[] = {
    {0, 10},
    {0, 30},
    {0, 20},
  };
  
  dulint commit_ts[] = {
    {0, 60},
    {0, 80},
    {0, 70},
  };
  
  setup_n_transactions(n_trx, trx_list, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* T0 first */
  result = ctg_add(ctg, trx_list[0]);
  ut_ad(result == TRUE);

  /* no ds from T0 */
  result = ctg_ds_test(ctg, trx_list[0]);
  ut_ad(result == DB_SUCCESS);
  ut_ad(trx_list[0]->bolt_length == 0);

  
  /* T2 next */
  result = ctg_add(ctg, trx_list[2]);
  ut_ad(result == TRUE);
  
  /* no ds yet */
  result = ctg_ds_test(ctg, trx_list[2]);
  ut_ad(result == DB_SUCCESS);
  ut_ad(trx_list[2]->bolt_length == 0);


  /* finally, T1 */
  result = ctg_add(ctg, trx_list[1]);
  ut_ad(result == TRUE);

  /* T1 -> T0 */
  result = ctg_add_edge(ctg, trx_list[1], trx_list[0]);
  ut_ad(result == TRUE);

  /* T2 -> T1 */
  result = ctg_add_edge(ctg, trx_list[2], trx_list[1]);
  ut_ad(result == TRUE);

  /* T1 is pivot in ds */
  result = ctg_ds_test(ctg, trx_list[1]);
  ut_ad(result == DB_DEADLOCK);
  ut_ad(trx_list[1]->bolt_length == 2);


  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx_list);  

}


UMB_TEST_ENTRY void umb_test_ctg_ds(void) {
  UMB_TEST_ANNOUNCE;

  test_one_edge_adjacency();
  test_ds_positive();
  test_ds_cycle_2();
  test_ds_serial_safe_dependencies();
  test_ds_t_pivot_commits_first();
  test_ds_t_in_commits_first();
  test_ds_no_pivot();
  test_ds_in_last();
  test_ds_pivot_last();

}



#endif /* UNIV_DEBUG */
