/*
 * $Id: umb-test-ctg-edgelist.c,v 1.4 2010/05/20 02:46:23 srevilak Exp $
 *
 * SREVILAK: unit test for ctg edgelist
 */

#ifdef UNIV_DEBUG

#include "umb-test.h"
#include "lock0ctg.h"
#include "mem0mem.h"

static mem_heap_t * heap = NULL;

#define THREE_EDGES { {1,1}, {2,2}, {3,3} }

static void one_time_setup(void) {
  heap = mem_heap_create(1024);
}


static void teardown(void) {
  if (heap) {
    mem_heap_empty(heap);
  }
}

static void one_time_teardown(void) {
  teardown();
  if (heap) {
    mem_heap_free(heap);
    heap = NULL;
  }
}


static void test_empty_list(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_empty_list");

  ctg_edgelist_t * list = ctg_edgelist_create(heap);
  
  /* 
   * list should have zero elements, and be owned by the correct transaction.
   */
  ut_ad(ctg_edgelist_size(list) == 0);
  
  teardown();
}


/**
 * Add one element, verify it's there, then fish it out.
 */
static void test_add_retrive_1(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_add_retrive_1");
  dulint trxid = { 101, 102 };
  ctg_edgelist_t * list = ctg_edgelist_create(heap);

  dulint trxid2 = { 200, 201 };

  /* starts empty */
  ut_ad(ctg_edgelist_size(list) == 0);

  /* first add succeeds. */
  ibool rv = ctg_edgelist_add(&list, heap, trxid2);
  ut_ad(rv == TRUE);

  /* should contain (200, 201), but not (101, 102) */
  rv = ctg_edgelist_contains(list, trxid2);
  ut_ad(rv == TRUE);

  rv = ctg_edgelist_contains(list, trxid);
  ut_ad(rv == FALSE);

  /* size should be one */
  ulint n_size = ctg_edgelist_size(list);
  ut_ad(n_size == 1);

  /* get the element out */
  dulint edge = ctg_edgelist_get(list, 0);
  ut_ad(UT_DULINT_EQ(edge, trxid2));

  /* add the element again.  Should produce no change -- already in list */
  rv = ctg_edgelist_add(&list, heap, trxid2);
  ut_ad(rv == FALSE);

  /* size is still one */
  n_size = ctg_edgelist_size(list);
  ut_ad(n_size == 1);

  teardown();
}


/**
 * Returns index of "target" in "array", or -1.
 */
static int find_dulint(dulint target, const dulint * array, int n_elements) {
  int idx = 0;
  
  for (; idx < n_elements; idx++) {
    if (UT_DULINT_EQ(target, array[idx])) {
      return idx;
    }
  }
  return -1;
}


/**
 * Add three elements, and verify.
 */
static void test_add_retrive_3(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_add_retrive_3");

  dulint edges[] = THREE_EDGES;
  int mark[] = { 0, 0, 0 };

  const ulint n_edges = 3;

  ctg_edgelist_t * edgelist = ctg_edgelist_create(heap);
  ulint iter = 0;
  ibool rv = FALSE;

  /**
   * Add three edges 
   */
  for (; iter < n_edges; iter++) {
    rv = ctg_edgelist_add(&edgelist, heap, edges[iter]);
    ut_ad(rv == TRUE);
  }

  /*
   * Make sure there are three edges
   */
  ulint edge_count = ctg_edgelist_size(edgelist);
  ut_ad(edge_count == n_edges);

  /**
   * Pull them out, and mark them off
   */
  for (iter = 0; iter < n_edges; iter++) {
    dulint fetch = ctg_edgelist_get(edgelist, iter);
    int to_mark = find_dulint(fetch, edges, n_edges);
    ut_ad(to_mark >= 0);
    mark[to_mark] = 1;
  }

  /* make sure that all edges are marked */
  for (iter = 0; iter < n_edges; iter++) {
    ut_ad(mark[iter] != 0);
  }

  teardown();
}


/**
 * Test add and retrive with lots of elements.  Enough
 * to cause reallocation. */
static void test_add_retrieve_lots(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_add_lots");
  const ulint n_edges = 100;

  dulint tmp;
  ulint iter;
  ibool rv;

  ctg_edgelist_t * edgelist = ctg_edgelist_create(heap);

  /* stick 100 edges in the list */
  for (iter = 0; iter < n_edges; iter++) {
    tmp.high = 1000 + iter;
    tmp.low = 1000 + iter;
    rv = ctg_edgelist_add(&edgelist, heap, tmp);
    ut_ad(rv == TRUE);
  }

  ulint list_size = ctg_edgelist_size(edgelist);
  ulint alloc_size = ctg_edgelist_alloc_size(edgelist);
  ut_ad(list_size == n_edges);
  ut_ad(alloc_size > list_size);
  
  /* find them all */
  for (iter = 0; iter < n_edges; iter++) {
    tmp.high = 1000 + iter;
    tmp.low = 1000 + iter;
    rv = ctg_edgelist_contains(edgelist, tmp);
    ut_ad(rv == TRUE);
  }
  
  teardown();
}


/**
 * Common prep for removal tests
 */
static void prep_for_remove(ctg_edgelist_t ** edgelist,
                            dulint * to_add,
                            ulint n_add) {
  ibool rv = FALSE;
  ulint idx = 0;
  for (; idx < n_add; idx++) {
    rv = ctg_edgelist_add(edgelist, heap, to_add[idx]);
    ut_ad(rv == TRUE);
  }

  for (idx = 0; idx < n_add; idx++) {
    rv = ctg_edgelist_contains(*edgelist, to_add[idx]);
    ut_ad(rv == TRUE);
  }
  
  ulint n_edges = ctg_edgelist_size(*edgelist);
  ut_ad(n_edges == n_add);
}



static void test_remove_helper(ulint to_remove, int expect[3]) {

  dulint edges[] = THREE_EDGES;
  const ulint n_edges = 3;
  int mark[] = { 0, 0, 0 };
  ctg_edgelist_t * edgelist = ctg_edgelist_create(heap);

  /* populate the edgelist with a set of three edges */
  ut_ad(to_remove < n_edges);
  prep_for_remove(&edgelist, edges, n_edges);

  /* remove an element */
  ibool rv = ctg_edgelist_remove(edgelist, edges[to_remove]);
  ut_ad(rv == TRUE);
  ut_ad(ctg_edgelist_size(edgelist) == 2);
  
  /* mark off the remaining elements */
  ulint idx;
  for (idx = 0; idx < ctg_edgelist_size(edgelist); idx++) {
    dulint fetch = ctg_edgelist_get(edgelist, idx);
    int mark_idx = find_dulint(fetch , edges, n_edges);
    ut_ad(mark_idx >= 0);
    mark[mark_idx] = 1;
  }

  /* verify that the remaining elements are what the caller expects */
  ut_ad(mark[0] == expect[0]);
  ut_ad(mark[1] == expect[1]);
  ut_ad(mark[2] == expect[2]);

}


/**
 * From a list of three elements, 
 * remove the first
 */
static void test_remove_first(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_remove_first");

  int expect[] = { 0, 1, 1 };
  test_remove_helper(0, expect);

  teardown();
}


/**
 * From a list of three elements, 
 * remove the second
 */
static void test_remove_second(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_remove_second");

  int expect[] = { 1, 0, 1 };
  test_remove_helper(1, expect);

  teardown();
}


/**
 * From a list of three elements, 
 * remove the third
 */
static void test_remove_third(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_remove_third");

  int expect[] = { 1, 1, 0 };
  test_remove_helper(2, expect);

  teardown();
}


/**
 * Remove an element that's not in the list
 */
static void test_remove_not_there(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_remove_not_there");

  dulint edge = { 0, 1 };
  dulint not_edge = { 3, 4 };
  ctg_edgelist_t * edgelist = ctg_edgelist_create(heap);

  ibool rv = ctg_edgelist_add(&edgelist, heap, edge);
  ut_ad(rv == TRUE);
  ut_ad(ctg_edgelist_size(edgelist) == 1);

  rv = ctg_edgelist_remove(edgelist, not_edge);
  ut_ad(rv == FALSE);
  ut_ad(ctg_edgelist_size(edgelist) == 1);
  ut_ad(ctg_edgelist_contains(edgelist, edge));

  teardown();
}


/**
 * "main"
 */
UMB_TEST_ENTRY void umb_test_ctg_edgelist(void) {
  UMB_TEST_ANNOUNCE;

  one_time_setup();

  test_empty_list();
  test_add_retrive_1();
  test_add_retrive_3();
  test_add_retrieve_lots();
  test_remove_first();
  test_remove_second();
  test_remove_third();
  test_remove_not_there();

  one_time_teardown();
}


#endif /* UNIV_DEBUG */
