/*
 * $Id: umb-test-ctg-prune.c,v 1.4 2010/03/30 01:28:00 srevilak Exp $
 */

#ifdef UNIV_DEBUG

#include "umb-test.h"
#include "lock0ctg.h"
#include "trx0trx.h"
#include "srv0srv.h"

static ctg_t * ctg;

static dulint last_pruned;

static void my_prune_observer(trx_t * trx) {
  ut_ad(trx);
  last_pruned = trx->id;

  /* fprintf(stderr, "PRUNE:%u.%u\n", 
   *         ut_dulint_get_high(last_pruned),
   *         ut_dulint_get_low(last_pruned));
   */
}


static void setup_transactions(ulint n,
                               trx_t ** array,
                               dulint * start_ts,
                               dulint * commit_ts) {

  ctg = ctg_create(128, &kernel_mutex);

  mutex_enter(&kernel_mutex);

  ulint iter = 0;

  for (; iter < n; iter++) {
    array[iter] = trx_create(NULL);
    array[iter]->id = start_ts[iter];
    array[iter]->commit_ts = commit_ts[iter];
    array[iter]->conc_state = TRX_CTG_RESIDENT; /* so that prune works */
  }

  ctg_register_observer(ctg, my_prune_observer);
  last_pruned = ut_dulint_zero;

  mutex_exit(&kernel_mutex);
}


static void teardown(ulint n, trx_t ** array) {

  ulint iter;

  mutex_enter(&kernel_mutex);

  ctg_free(&ctg);

  for (iter = 0; iter < n; iter++) {
    ut_ad(array[iter]->ctg_hash == NULL);
    ut_ad(array[iter]->in_edge_count == 0);
    array[iter]->conc_state = TRX_NOT_STARTED; /* avoid assertion in trx_free */
    trx_free(array[iter]);
    array[iter] = NULL;
  }

  mutex_exit(&kernel_mutex);
}


/**
 * Test pruning with one transaction
 */
static void test_prune_1(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_prune_1");

  dulint oldest_active_ts = { 1, 1 };
  trx_t * trx[1];
  dulint start_ts[] = { {100, 100 } };
  dulint commit_ts[] = { {100, 200 } };
  const ulint n_trx = 1;

  setup_transactions(1, trx, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* add T1 to ctg */
  ctg_add(ctg, trx[0]);

  /* prune.  oldest_active_ts is really old, so nothing should
   * be pruned */
  ulint pruned = ctg_prune(ctg, oldest_active_ts);
  ut_ad(pruned == 0);
  ut_ad(ctg_get(ctg, trx_get_trx_id(trx[0])) == trx[0]);

  /* change oldest active, so that it starts during
   * T[0]'s lifetime.  Still no change */
  oldest_active_ts = ut_dulint_create(100, 150);
  pruned = ctg_prune(ctg, oldest_active_ts);
  ut_ad(pruned == 0);
  ut_ad(ctg_get(ctg, trx_get_trx_id(trx[0])) == trx[0]);

  /* now oldest active starts after T[0] commits.  Gets pruned */
  oldest_active_ts = ut_dulint_create(100, 201);
  pruned = ctg_prune(ctg, oldest_active_ts);
  ut_ad(pruned == 1);
  ut_ad(ctg_get(ctg, trx_get_trx_id(trx[0])) == NULL);
  ut_ad(UT_DULINT_EQ(last_pruned, trx_get_trx_id(trx[0])));

  mutex_exit(&kernel_mutex);

  teardown(n_trx, trx);
}



/**
 * Test with three transactions
 *
 *    T0---------> T1 --------
 *
 *  T2----------------------
 *
 * Prune before T0 start, before T1 start, before T1 commit, and after
 * all commit.
 */
static void test_prune_3a(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_prune_3a");

  trx_t * trx[3];

  dulint start_ts[] = {
    { 100, 110 },
    { 100, 150 },
    { 100, 100 }
  };

  dulint commit_ts[] = {
    { 100, 140 }, /* T[0] commits before T[1] starts */
    { 100, 210 },
    { 100, 200 }
  };
  const ulint n_trx = 3;
  ulint iter;
  ibool rv;

  setup_transactions(n_trx, trx, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* put transactions in the graph */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx[iter]);
    ut_ad(rv);
  }

  /* add edge T0 -> T1 */
  rv = ctg_add_edge(ctg, trx[0], trx[1]);
  ut_ad(rv);

  /* prune from prior T[2] start.  No change */
  dulint oldest_active = ut_dulint_create(100, 0);
  ulint prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == 3);

  /* prune with oldest active starting during T1's lifetime.
   * should produce no change */
  oldest_active = ut_dulint_create(100, 115);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == 3);

  /* prune from after T[0] commit.  Should prune T[0] */
  oldest_active = ut_dulint_create(100, 145);
  ut_ad(trx[1]->in_edge_count == 1);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 1);
  ut_ad(ctg_size(ctg) == 2);
  ut_ad(UT_DULINT_EQ(last_pruned, trx[0]->id));
  ut_ad(trx[1]->in_edge_count == 0);

  /* prune after all trx commit.  Should zap everything */
  oldest_active = ut_dulint_create(100, 300);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 2);
  ut_ad(ctg_size(ctg) == 0);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx);
}


/**
 * Test with three transactions.  Like test_prune_3a(),
 * but the edge is T1->T0 instead of T0->T1
 *
 *    T0---------<
 *           T1 --------
 *
 *  T2----------------------
 *
 */
static void test_prune_3b(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_prune_3b");

  trx_t * trx[3];

  dulint start_ts[] = {
    { 100, 110 },
    { 100, 140 },
    { 100, 100 }
  };

  dulint commit_ts[] = {
    { 100, 150 },
    { 100, 200 },
    { 100, 210 }
  };
  const ulint n_trx = 3;
  ulint iter;
  ibool rv;

  setup_transactions(n_trx, trx, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* put transactions in the graph */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx[iter]);
    ut_ad(rv);
  }

  /* add edge T1 -> T0 */
  rv = ctg_add_edge(ctg, trx[1], trx[0]);
  ut_ad(rv);

  /* prune from prior T[2] start.  No change */
  dulint oldest_active = ut_dulint_create(100, 0);
  ulint prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == 3);

  /* prune after T[1] start, but before T[0] commit.  No change */
  oldest_active = ut_dulint_create(100, 145);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == 3);

  /* prune from after T[0] commit.  T[0] has in-edge, so no change */
  oldest_active = ut_dulint_create(100, 155);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == 3);

  /* prune after T[1] commit, but before T[2] commit.
   * Should prune T[1] then T[0] */
  oldest_active = ut_dulint_create(100, 205);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 2);
  ut_ad(ctg_size(ctg) == 1);
  ut_ad(ctg_get(ctg, trx_get_trx_id(trx[2])) == trx[2]);
  ut_ad(UT_DULINT_EQ(trx[0]->id, last_pruned));

  /* prune after commit T[2].  Should produce empty graph */
  oldest_active = ut_dulint_create(100, 300);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 1);
  ut_ad(ctg_size(ctg) == 0);
  ut_ad(UT_DULINT_EQ(trx[2]->id, last_pruned));

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx);
}


/*
 * Here, our CTG is a small binary tree
 *
 * 0 ---> 1 ----> 3
 *    |       |
 *    |       +-> 4
 *    |
 *    +-> 2 ----> 5
 *            |
 *            +-> 6
 *
 * Verify that one prune knocks down the entire tree.
 */
static void test_prune_fan_out(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_prune_fan_out");

  const ulint n_trx = 7;
  trx_t * trx[7];

  /* we're not too concerned about timestamps.  Here, 
   * we care more about how prune follows edges recursively
   */
  dulint start_ts[] = {
    {1, 1}, 
    {1, 2},
    {1, 3},
    {1, 4},
    {1, 5}, 
    {1, 6}, 
    {1, 7}
  };

  dulint commit_ts[] = {
    {1, 11}, 
    {1, 12},
    {1, 13},
    {1, 14},
    {1, 15}, 
    {1, 16}, 
    {1, 17}
  };
  
  dulint oldest_active = { 0, 100 };

  /* edges -- each dulint is a pair of trx[] indexes */
  dulint edges[] = {
    { 0, 1 },
    { 0, 2 },
    { 1, 3 },
    { 1, 4 },
    { 2, 5 },
    { 2, 6 }
  };

  ibool rv;
  ulint prune_count;
  ulint iter;

  setup_transactions(n_trx, trx, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* put transactions in the graph */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx[iter]);
    ut_ad(rv);
  }
  ut_ad(ctg_size(ctg) == n_trx);

  /* put the edges in the graph */
  for (iter = 0; iter < sizeof(edges)/sizeof(edges[0]); iter++) {
    ulint source = ut_dulint_get_high(edges[iter]);
    ulint sink = ut_dulint_get_low(edges[iter]);
    
    rv = ctg_add_edge(ctg, trx[source], trx[sink]);
    ut_ad(rv);
  }

  /* prune from before T[0] start.  Changes nothing */
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == n_trx);

  /* prune from after all commit.  Clears the graph */
  oldest_active = ut_dulint_create(500, 1);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == n_trx);
  ut_ad(ctg_size(ctg) == 0);

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx);
}



/*
 * like test_prune_fan_out, but the edges go in opposite directions.
 *
 * 0 <---< 1 <---- 3
 *    |        |
 *    |        +-< 4
 *    |
 *    +--< 2 <---- 5
 *             |
 *             +-< 6
 *
 * Verify that one prune knocks down the entire tree.
 * But this time, we have four elements in the starting set,
 * and node zero must be the last one to go.
 */
static void test_prune_fan_in(void) {
  UMB_TEST_FUNC_ANNOUNCE("test_prune_fan_in");

  const ulint n_trx = 7;
  trx_t * trx[7];

  /* again, nothing special about our choice of timestamps */
  dulint start_ts[] = {
    {1, 1}, 
    {1, 2},
    {1, 3},
    {1, 4},
    {1, 5}, 
    {1, 6}, 
    {1, 7}
  };

  dulint commit_ts[] = {
    {1, 11}, 
    {1, 12},
    {1, 13},
    {1, 14},
    {1, 15}, 
    {1, 16}, 
    {1, 17}
  };
  
  dulint oldest_active = { 0, 100 };

  /* edges -- each dulint is a pair of trx[] indexes */
  dulint edges[] = {
    { 3, 1 },
    { 4, 1 },
    { 1, 0 },
    { 5, 2 },
    { 6, 2 },
    { 2, 0 }
  };

  ibool rv;
  ulint prune_count;
  ulint iter;

  setup_transactions(n_trx, trx, start_ts, commit_ts);

  mutex_enter(&kernel_mutex);

  /* put transactions in the graph */
  for (iter = 0; iter < n_trx; iter++) {
    rv = ctg_add(ctg, trx[iter]);
    ut_ad(rv);
  }
  ut_ad(ctg_size(ctg) == n_trx);

  /* put the edges in the graph */
  for (iter = 0; iter < sizeof(edges)/sizeof(edges[0]); iter++) {
    ulint source = ut_dulint_get_high(edges[iter]);
    ulint sink = ut_dulint_get_low(edges[iter]);
    
    rv = ctg_add_edge(ctg, trx[source], trx[sink]);
    ut_ad(rv);
  }

  /* prune from before T[0] start.  Changes nothing */
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == 0);
  ut_ad(ctg_size(ctg) == n_trx);

  /* prune from after all commit.  Clears the graph */
  oldest_active = ut_dulint_create(500, 1);
  prune_count = ctg_prune(ctg, oldest_active);
  ut_ad(prune_count == n_trx);
  ut_ad(ctg_size(ctg) == 0);

  /* T[0] must be the last transaction pruned */
  ut_ad(UT_DULINT_EQ(last_pruned, trx[0]->id));

  mutex_exit(&kernel_mutex);
  teardown(n_trx, trx);
}


UMB_TEST_ENTRY void umb_test_ctg_prune(void) {
  UMB_TEST_ANNOUNCE;

  test_prune_1();
  test_prune_3a();
  test_prune_3b();
  test_prune_fan_out();
  test_prune_fan_in();

}



#endif /* UNIV_DEBUG */
